/* complx.h - definitions to support FORTRAN translations to C 
 *        made by FOR_C (TM) v3.1 from COBALT BLUE 
 *         copyright Cobalt Blue, Inc., 1988-1992
 *               ALL RIGHTS RESERVED
 */

#ifndef COMPLEX_H
#define COMPLEX_H	/* only define these objects once per file! */

/* Define the macro switch:
		INLN_COMPLEX
	if you wish to use macros expanding to inline code (instead of
	functions) for the complex conversion functions.
	Beware of duplicate names!
 */

/* NOTE: fcrt.h should be included in front of this file so that any
	configuration macro switches are properly defined! */

/* redefine complex/complexf to avoid possible conflict with compilers */
#define complex complx
#define complexf complxf

#ifdef cabs
#undef cabs
#endif

#define cabs	_c_abs
	/* avoids conflict w/cabs() defined in some C libraries */

typedef struct {
	float re;
	float im;
	}	complexf;
#define COMPLEX complexf

typedef struct {
	double re;
	double im;
	}	complex;
#define COMPLEX16 complex

/* access to the real & imaginary components of the complex nos. */
#define ZRE(z)	(z).re
#define ZIM(z)	(z).im

	/* real & imag. components for I/O purposes */
#define ZADR(z)		&(z).re, &(z).img		/* used for input */
#define ZPAIR(z)	(z).re, (z).im			/* used for output */

#ifdef PROTOTYPES

	/* complex function prototypes */

	complex cneg( complex );	/* negation of a dp complex no. */
	complex cadd( complex, complex );	/* sum of two complex nos. */
	complex csub( complex, complex );	/* difference of two complex nos. */
	complex cmul( complex, complex );	/* product of two complex nos. */
	complex cdiv( complex, complex );	/* quotient of two complex nos. */
	complex cpow( complex, complex );	/* general complex power */
	complex cpowf( complex, double );	/* complex raised to float power*/
	int ceq( complex, complex );	/* return TRUE if complex nos. equal */
	complex cftoc( complexf );		/* complexf conversion to complex */
	/* complex Library Functions */
	double ctof( complex ); 	/* dp complex (real component) to float */
	double dimag( complex );	/* imaginary component of dp complex no. */
	complex ftoc(double,double), /* convert floats to dp complex */
		dconjg( complex );	/* dp complex conjugate */
	double arg_dcpx(complex);	/* direct angle (in radians) of dp complex */
	complex csqrt( complex ),	/* (positive) sq root of dp complex no. */
		clog( complex ),	/* dp complex natural log */
		cexp( complex ),	/* dp complex exponential */
		csin( complex ),	/* sine of dp complex no. */
		ccos( complex );	/* cosine of dp complex no. */
	double cabs( complex );	/* dp complex absolute value */
	/*pass by address wrappers*/
	complex csqrt_( complex* ), clog_( complex* ), cexp_( complex* ),
	         csin_( complex* ), ccos_( complex* ), dconjg_(complex*);
	double cabs_( complex* ), dimag_( complex* );

	/* complexf function prototypes */

	complexf cfneg( complexf );	/*negation of a complex no.*/
	complexf cfadd( complexf, complexf ); /*sum of two complex nos.*/
	complexf cfsub( complexf, complexf ); /*difference of two complex nos.*/
	complexf cfmul( complexf, complexf ); /*product of two complex nos.*/
	complexf cfdiv( complexf, complexf ); /*quotient of two complex nos.*/
	complexf cfpow( complexf, complexf ); /*general complex power*/
	complexf cfpowf( complexf, double ); /*gen complex raised to float power*/
	int cfeq( complexf, complexf ); /*return TRUE if complex nos. equal*/
	complexf ctocf( complex );      /*complex conversion to complexf*/
	/* complexf Library Functions*/
	double cftof( complexf ); /*complex (real component) to float conversion*/
	double aimag( complexf ); /*imaginary component of complex no.*/
	complexf ftocf(double, double), /*convert floats to complex*/
		conjg( complexf ); /*complex conjugate*/
	double arg_cpx( complexf ); /*direct angle (in radians) of complex no*/
	complexf cfsqrt( complexf ), /*(positive) sq root of complex no.*/
		cflog( complexf ), /*complex natural log*/
		cfexp( complexf ), /*complex exponential*/
		cfsin( complexf ), /*sine of complex no.*/
		cfcos( complexf ); /*cosine of complex no.*/
	double cfabs( complexf ); /*complex absolute value*/
	/*pass by address wrappers*/
	complexf cfsqrt_(complexf*), cflog_(complexf*), cfexp_(complexf*), 
	          cfsin_(complexf*), cfcos_(complexf*), conjg_(complexf*);
	double cfabs_(complexf*), aimag_(complexf*);

#else	/* only declare the functions */

	complex cneg(), cadd(), csub(), cmul(), cdiv(), cpow(), cpowf();
	int ceq();
	complex cftoc();
	double ctof();
	double dimag(), arg_dcpx();
	complex ftoc(), dconjg(), csqrt(), clog(), cexp(), csin(), ccos();
	double cabs();
	/*pass by address wrappers*/
	complex csqrt_(), clog_(), cexp_(),
	         csin_(), ccos_(), dconjg_();
	double cabs_(), dimag_();

	complexf cfneg(), cfadd(), cfsub(), cfmul(), cfdiv(), cfpow(), cfpowf();
	int cfeq();
	complexf ctocf();
	double cftof();
	double aimag(), arg_cpx();
	complexf ftocf(), conjg(), cfsqrt(), cflog(), cfexp(), cfsin(), cfcos();
	double cfabs();
	/*pass by address wrappers*/
	complexf cfsqrt_(), cflog_(), cfexp_(), 
	          cfsin_(), cfcos_(), conjg_();
	double cfabs_(), aimag_();

#endif	/* on PROTOTYPES */

			/* * * psuedo functions * * */

#define itocf(i,j)	ftocf((double)i, (double)j)
#define sitocf(i,j)	ftocf((double)i, (double)j)
#define cftoi( z )	(long)cftof(z)
#define cftosi( z )	(short)cftof(z)

#define itodc(i,j)	ftoc((double)i, (double)j)
#define sitodc(i,j)	ftoc((double)i, (double)j)
#define ctoi( z )	(long)ctof(z)
#define ctosi( z )	(short)ctof(z)

#ifdef INLN_COMPLEX	/* then use the following macro versions */
#ifndef 	EXTERNL
#define 	EXTERNL extern
#endif
	/* NOTE: EXTERNL is defined as nothing in f77ini.c to define the
	 *  the complex EXTERNLs in that file.  This K&R approach avoids 
	 *   problems with certain older Librarians in use. */

EXTERNL complexf _z_;	/* temp for complex macros, (to avoid side effects) */
#define 	aimag(z)	(_z_=(z), _z_.im)	/* return float */
#define 	cftof(z) 	(_z_=(z), _z_.re)	/* return float */
#define 	ftocf(r,i)	((_z_.re=(r), _z_.im=(i)), _z_)	/* return complex */
#define 	conjg(z)	((_z_=(z), _z_.im=-_z_.im), _z_)	/* return complex */
#define 	cfneg(z) 	((_z_=(z), (_z_.re=-_z_.re,_z_.im=-_z_.im)), _z_)

EXTERNL complex _dz_;	/* temp for complex macros, (to avoid side effects) */
#define 	dimag(z)	(_dz_=(z), _dz_.im)	/* return float */
#define 	ctof(z) 	(_dz_=(z), _dz_.re)	/* return float */
#define 	ftoc(r,i)	((_dz_.re=(r), _dz_.im=(i)), _dz_) /* return complex */
#define 	dconjg(z)	((_dz_=(z), _dz_.im=-_dz_.im), _dz_) /* return complex */
#define 	cneg(z)	((_dz_=(z), (_dz_.re=-_dz_.re,_dz_.im=-_dz_.im)), _dz_)
#endif	/* of INLN_COMPLEX */

#endif	/* on COMPLEX_H */
