/* bitmanip.c - support for MILSPEC Bit Manipulation Functions translated
 *             by FOR_C (TM) from COBALT BLUE
 *    copyright Cobalt Blue, Inc., 1988 - 1993
 *               ALL RIGHTS RESERVED
 */

/* NOTE: While these bit functions are intended to be portable, there
	may be some machine dependances, such as byte ordering, which
	should be accounted for.  Define TEST_BITS and compile this file
	separately to test these functions for your target machine.
#define TEST_BITS
 */

	/* NOTE: The rightmost bit position is defined as 0!  A bit
		position that is <0 or >wordsize is an error.  */

#include <fcrt.h>
#define BITS_PER_BYTE	8

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
short sshft( short m, short k )	/* short bit shift */
#else
short sshft( m, k )	/* short bit shift */
short m,
	k;	/* k>0: shift left k places, k<0: shift right k places */
#endif
{
	if( k < 0 )
		return( m >> (-k) );
	else if( k > 0 )
		return( m << k );
	return( m );	/* k == 0, no shift */
}

long ishft( m, k )	/* long int bit shift */
long m,
	k;	/* k>0: shift left k places, k<0: shift right k places */
{
	if( k < 0L )
		return( m >> (-k) );
	else if( k > 0L )
		return( m << k );
	return( m );	/* k == 0, no shift */
}

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
short sshftc( short m, short k, short ic )	/* circular shift -- no bits are lost */
#else
short sshftc( m, k, ic )	/* circular shift -- no bits are lost */
short m,
	k,	/* shift k places, k>0: left shift, k<0: right shift */
	ic;	/* rightmost ic bits of m to shift */
#endif
{
	short direction, tmp, high, mask;

	/* check & save shift direction */
	if( k == 0 )
		return( m );
	if( k > 0 )
		direction = 1;
	else
		direction = 0;
	k = _ABS_(k);

	/* error checks */
	if( (ic < 1 || ic > BITS_PER_BYTE*sizeof(short) ) ||  k > ic )
		return( 0 );

	/* init. */
	high = ic - 1;
	tmp = 1;
	mask = tmp << high;

	/* circular shift the range of bits desired */
	while( k-- ){
		tmp = m;
		high = tmp & mask;
		if(direction)
			tmp = tmp << 1;
		else
			tmp = tmp >> 1;
		if( high )
			tmp = tmp | 1;
		}
	/* build the bit range mask */
	high = ic - 1;
	mask = 1;
	while( high-- ){
		mask = mask << 1;
		mask = mask | 1;
		}
	/* create OR and AND masks for shifted and non-shifted bits */
	tmp = tmp & mask;
	m = m & ~mask;
	return( m | tmp );
}

long ishftc( m, k, ic )	/* circular shift -- no bits are lost */
long m,
	k,	/* shift k places, k>0: left shift, k<0: right shift */
	ic;	/* rightmost ic bits of m to shift */
{
	long direction, tmp, high, mask;

	/* check & save shift direction */
	if( k == 0L )
		return( m );
	if( k > 0L )
		direction = 1L;
	else
		direction = 0L;
	k = _ABS_(k);

	/* error checks */
	if( (ic < 1L || ic > BITS_PER_BYTE*sizeof(long) ) ||  k > ic )
		return( 0L );

	/* init. */
	high = ic - 1L;
	tmp = 1L;
	mask = tmp << high;

	/* circular shift the range of bits desired */
	while( k-- ){
		tmp = m;
		high = tmp & mask;
		if(direction)
			tmp = tmp << 1L;
		else
			tmp = tmp >> 1L;
		if( high )
			tmp = tmp | 1L;
		}
	/* build the bit range mask */
	high = ic - 1L;
	mask = 1L;
	while( high-- ){
		mask = mask << 1L;
		mask = mask | 1L;
		}
	/* create OR and AND masks for shifted and non-shifted bits */
	tmp = tmp & mask;
	m = m & ~mask;
	return( m | tmp );
}

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
short sbits( short m, short i, short len )	/* bit extraction */
#else
short sbits( m, i, len )	/* bit extraction */
short m,	/* source int */
	i,	/* starting bit position (far right is 0) */
	len;	/* length of subfield to extract */
#endif
{
	short tmp, high, mask;

	if( i + len > sizeof(short)*BITS_PER_BYTE )
		return( 0 );	/* error */

	/* build the bit range mask */
	mask = 0;
	tmp = 1 << i;
	high = len - 1;
	/* build extraction mask */
	while( high-- ){
		mask = mask | tmp;	/* turn ON bits in mask */
		tmp = tmp << 1;	/* shift tmp left 1 bit */
		}
	/* extract and return desired bits (right justified) */
	return( (m & tmp) >> i );
}

long ibits( m, i, len )	/* bit extraction */
long m,	/* source int */
	i,	/* starting bit position (far right is 0) */
	len;	/* length of subfield to extract */
{
	long tmp, high, mask;

	if( i + len > sizeof(long)*BITS_PER_BYTE )
		return( 0L );	/* error */

	/* build the bit range mask */
	mask = 0L;
	tmp = 1L << i;
	high = len - 1L;
	/* build extraction mask */
	while( high-- ){
		mask = mask | tmp;	/* turn ON bits in mask */
		tmp = tmp << 1L;	/* shift tmp left 1 bit */
		}
	/* extract and return desired bits (right justified) */
	return( (m & tmp) >> i );
}

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
int sbtest( short n, short i )	/* short bit test--returns 1 if bit is set*/
#else
int sbtest( n, i )	/* short bit test -- returns 1 if bit is set */
short n,
	i;	/* bit position to test */
#endif
{
	short mask;
	/* determine bit position and return 0 or 1 */
	mask = 1 << i;
	if( n & mask )
		return( 1 );
	return( 0 );
}

int btest( n, i )	/* bit test -- returns 1 if bit is set */
long n,
	i;	/* bit position to test */
{
	long mask;
	/* determine bit position and return 0 or 1 */
	mask = 1L << i;
	if( n & mask )
		return( 1 );
	return( 0 );
}

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
short sbset( short n, short i )		/* short bit set */
#else
short sbset( n, i )		/* short bit set */
short n,
	i;	/* bit position to set to 1 */
#endif
{
	short mask;
	mask = 1 << i;
	return( n | mask );
}

long ibset( n, i )		/* bit set */
long n,
	i;	/* bit position to set to 1 */
{
	long mask;
	mask = 1L << i;
	return( n | mask );
}

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
short sbclr( short n, short i )		/* short bit clear */
#else
short sbclr( n, i )		/* short bit clear */
short n,
	i;	/* bit position to clear (to 0) */
#endif
{
	short mask;
	mask = 1 << i;
	return( n & ~mask );
}

long ibclr( n, i )		/* bit clear */
long n,
	i;	/* bit position to clear (to 0) */
{
	long mask;
	mask = 1L << i;
	return( n & ~mask );
}

	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
void mvsbits(short s,short so,short len,short *t,short to) /*short bit move*/
#else
void mvsbits( s, so, len, t, to )	/* short bit move */
short s, 	/* source */
	so,	/* starting bit position in source */
	len, 	/* bit field length to be moved */
	*t, /* target */
	to;	/* starting bit position in target to place the bit field */
#endif
{
	/* number of bits to move */
	while( len-- > 0 ){
		if( sbtest(s,so) )
			*t = sbset(*t,to);	/* YES, then set target */
		else
			*t = sbclr(*t,to);	/* NO, then clear target */
		so++; to++;	/* bump indexes */
		}
	return;
}

void mvbits( s, so, len, t, to )	/* long bit move */
long s, so, len, *t, to;
{
	/* number of bits to move */
	while( len-- > 0L ){
		if( btest(s,so) )
			*t = ibset(*t,to);	/* YES, then set target */
		else
			*t = ibclr(*t,to);	/* NO, then clear target */
		so++; to++;	/* bump indexes */
		}
	return;
}

	/*	*	*	*	*/

#ifdef TEST_BITS
	main()
	{
		printf( "test of MILSPEC bit functions\n\n" );
		short_btsts();
		long_btsts();
	}

	short_btsts()		/* int versions */
	{
		int m, k, i, ic, n, len, s, so, t, to;

		m = 0x00ff;
		k = 2;
		printf( "%4.4x = sshft(%4.4x, %d)\n", sshft(m,k), m, k );
		m = 0x5500;
		k = -4;
		printf( "%4.4x = sshft(%4.4x, %d)\n", sshft(m,k), m, k );
		m = 0x0055;
		k = 4;
		printf( "%4.4x = sshft(%4.4x, %d)\n\n", sshft(m,k), m, k );

	}

	long_btsts()		/* long versions */
	{
		long m, k, i, ic, n, len, s, so, t, to;

		m = 0x55555555;
		k = 1;
		printf( "%8.8lx = ishft(%8.8lx, %ld)\n\n", ishft(m,k), m, k );

		m = 0x55555555;
		k = 1;
		ic = 16;
		printf( "%8.8lx = ishftc(%8.8lx, %ld, %ld)\n\n", ishftc(m,k,ic), m, k, ic );

		m = 0x12345678;
		k = 0;
		len = 16;
		printf("%8.8lx = ibits(%8.8lx, %ld, %ld)\n", ibits(m,k,len),m,k,len);
		k = 8;
		printf("%8.8lx = ibits(%8.8lx, %ld, %ld)\n", ibits(m,k,len),m,k,len);
		k = 16;
		len = 8;
		printf("%8.8lx = ibits(%8.8lx, %ld, %ld)\n\n", ibits(m,k,len),m,k,len);

		n = 0x55555555;
		i = 8;
		printf( "%8.8lx = btest(%8.8lx, %ld)\n", btest(n,i), n, i );
		i = 7;
		printf( "%8.8lx = btest(%8.8lx, %ld)\n\n", btest(n,i), n, i );

		n = 0x55555555;
		i = 1;
		printf( "%8.8lx = lbclr(%8.8lx, %ld)\n", lbclr(n,i), n, i );
		i = 6;
		printf( "%8.8lx = lbclr(%8.8lx, %ld)\n\n", lbclr(n,i), n, i );

		s = 0x11223344;
		so = 16;
		len = 16;
		t = 0;
		to = 16;
		printf( "mvbits( %8.8lx, %ld, %ld, t=%8.8lx, %ld);\n", s, so, len, t, to );
		mvbits( s, so, len, &t, to );
		printf( "  after: t = %8.8lx\n", t );
		t = 0;
		to = 8;
		printf( "mvbits( %8.8lx, %ld, %ld, t=%8.8lx, %ld);\n", s, so, len, t, to );
		mvbits( s, so, len, &t, to );
		printf( "  after: t = %8.8lx\n", t );
		t = 0;
		to = 0;
		printf( "mvbits( %8.8lx, %ld, %ld, t=%8.8lx, %ld);\n", s, so, len, t, to );
		mvbits( s, so, len, &t, to );
		printf( "  after: t = %8.8lx\n", t );
	}

/* the following results from running the TEST_BIT version of bitmanip,

test of MILSPEC bit functions

03fc = sshft(00ff, 2)
0550 = sshft(5500, -4)
0550 = sshft(0055, 4)

aaaaaaaa = ishft(55555555, 1)

5555aaaa = ishftc(55555555, 1, 16)

00005678 = ibits(12345678, 0, 16)
00003456 = ibits(12345678, 8, 16)
00000034 = ibits(12345678, 16, 8)

00000001 = btest(55555555, 8)
00000000 = btest(55555555, 7)

55555555 = lbclr(55555555, 1)
55555515 = lbclr(55555555, 6)

mvbits( 11223344, 16, 16, t=00000000, 16);
  after: t = 11220000
mvbits( 11223344, 16, 16, t=00000000, 8);
  after: t = 00112200
mvbits( 11223344, 16, 16, t=00000000, 0);
  after: t = 00001122

*/
#endif
