/* bksp.c -- support for translated F77 BACKSPACE stmts
 *         by FOR_C (TM) from COBALT BLUE
 *    copyright Cobalt Blue, Inc., 1988 - 1992
 *               ALL RIGHTS RESERVED
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>
#include <fcio.h>

	/*	*	*	*	*/

void bkspunit( u ) /* backspace 1 record in the file */
int u;
{
	if( unit_s[u].ateof )
		unit_s[u].ateof = NO;
		/*NOTE: backspace at EOF, merely turns off the ateof flag,
		 * subsequent backspace actually backspaces
		 */
	else if( unit_s[u].form == FORMATTED ){
		bkspf( unit_s[u].fp );
		}
	else{ /* UNformatted */
		bkspunf( unit_s[u].fp );
		}
	return; 
}

	/*	*	*	*	*/

void bkspace( fp, unf )	/* backspace file 1 line if fmt'd */
FILE *fp;					/* or 1 record if unf'd */
int unf;	/* 0 is fmt'd, 1 is unformatted */
{

	if( unf )
		bkspunf( fp );
	else
		bkspf( fp );

	return;
}

	/*	*	*	*	*/

#if defined(unix) || defined(vms)

#define MAXBKSP 1024

int bkspf( fp )	/* backspace 1 line in a formatted file */
FILE *fp;			/* returns 0 if successful, else 1 */
{
	/* NOTE: DON'T use this version on MSDOS!     OK for UNIX!
	 *   Formatted files are opened in text mode on DOS (by default),
	 *   this in turn prevents ftell() & fseek() from functioning correctly.
	 */
char buf[MAXBKSP+1],
	*last;	/* ptr to buf[] */
long n,
	ros,	/* relative offset */
	pos;	/* file position */
buf[MAXBKSP] = '\0';	/* terminate the buffer w/the last char */

	if( (pos=ftell(fp)) == -1 )	/* get the current position */
		return( 1 );	/* error */

	/* determine how much to re-position */
	ros = pos < MAXBKSP ? pos : MAXBKSP;

	if( fseek(fp,pos-ros,SEEK_SET) )	/* re-position to 'pos - ros' */
		;	/* seek error */
	else if( (n=fread(buf,sizeof(char),(size_t)ros,fp)) > ros )
		;	/* read error */
	else{	/* successfully read in last 'ros' chars */
		/* ck out last char & 'skip' if '\n' */
		if( *(last= &buf[n-1]) == '\n' ){
			last--; pos--;
			}
		/* find last \n (noting above exception) */
		while( *last != '\n' && last >= buf ){
			last--; pos--;
			}
		if( fseek(fp,pos,SEEK_SET) == 0L )
			return( 0 );	/* backspace'd successfully! */
		}

return( 1 );	/* backspace error */
}
#else	/* MSDOS */

int bkspf( fp )	/* backspace 1 line in a formatted file */
FILE *fp;			/* returns 0 if successful, else 1 */
{
	/* NOTE: formatted files are opened in text mode on DOS (by default),
	 *   hence only absolute fseek/ftell combinations may be used due to
	 *   CR/LF mapping to \n.
	 */
long n,
	ros,	/* relative offset */
	pos;	/* file position */

	if( (pos=ftell(fp)) == -1 )	/* get the current position */
		return( 1 );	/* error */

	/* NOTE: assumed to be at beginning of next record! ! ! */
	pos -= 3;	/* move back 3 spaces pass CRLF pair & last char in record */
	pos = ( pos < 0 ? 0 : pos );
	if( fseek(fp,pos,SEEK_SET) )
		return( 1 );	/* error */
	if( pos == 0 )	/* at BOF */
		return( 0 );

	while( fgetc(fp) != '\n' ){
		pos -= 2;	/*reposition back 1 char (plus 1 for the char read in)*/
		if( fseek(fp,pos,SEEK_SET) )
			return( 1 );	/* error */
		if( pos == 0 )
			return( 0 );	/* at BOF */
		}

return( 0 );	/* backspaced to prior record */
}
#endif

	/*	*	*	*	*/

int bkspunf( fp )	/* backspace an unformatted file 1 record */
FILE *fp;	/* (records are marked w/header & trailer counts) */
{
	/* UNformatted seq. record:	 N_ITEMS record_contents N_ITEMS  */

	long ros,	/* relative offset */
		recsiz;

	if( ftell(fp) == 0L )
		return( 1 ); /* at BOF - can't backspace */

	/* read the trailing N_ITEMS */
	if( fseek(fp,-(long)sizeof(long),SEEK_CUR) )
		;	/* seek error */
	else if( fread(&recsiz,sizeof(long),1,fp) != 1 )
		;	/* EOF or error (use ferror() to discern) */
	else{
		/* NOTE: the current offset is back at the current record */

		/* trailer record size read, NOW do the actual backspace */
		ros = recsiz + 2*sizeof(long);	/* offset difference */
		if( fseek(fp,-ros,SEEK_CUR) == 0L )
			return( 0 );	/* sucessfull */
		}
	return( 1 );	/* backspace error */
}
