/* complx.c - COMPLEX16 functions supporting translations by
 *        FOR_C (TM) from COBALT BLUE
 *     copyright Cobalt Blue, Inc., 1988-1996
 *              ALL RIGHTS RESERVED
 */

#include <math.h>
#include <fcrt.h>
#include <fcio.h>
#include <complx.h>

		/*		*		*		*		*/

#ifdef cneg
#undef cneg
#endif

	complex cneg( dz )	/* dp complex unary minus operation */
	complex dz;
	{
		dz.re = -dz.re;	dz.im = -dz.im;
		return( dz );
	}

		/*		*		*		*		*/


complex cadd( l, r )	/* retn the sum of two dp complex nos. */
complex l, r;
{
	l.re += r.re;
	l.im += r.im;
	return( l );
}


complex csub( l, r )	/* retn the difference of two dp complex nos. */
complex l, r;
{
	l.re -= r.re;
	l.im -= r.im;
	return( l );
}


complex cmul( l, r )	/* retn the product of two dp complex nos. */
complex l, r;
{
	complex dz;
	dz.re = l.re*r.re - l.im*r.im;
	dz.im = l.re*r.im + l.im*r.re;
	return( dz );
}


complex cdiv( l, r )	/* retn the quotient of two dp complex nos. */
complex l, r;
{
	complex dz;
	double den;

	if( r.re == 0. && r.im == 0. )
		f_err( "complex division by 0." );
	den = r.re*r.re + r.im*r.im;

	dz.re = ( l.re*r.re + l.im*r.im )/den;
	dz.im = ( r.re*l.im - l.re*r.im )/den;
	return( dz );
}


complex cpow( l, r )	/* general dp complex power */
complex l, r;
{
	if( l.re == 0. && l.im == 0. )
		return( l );	/* undefined */

	return( cexp( cmul(r,clog(l)) ) );
}


complex cpowf( l, r )	/* general dp complex raised to float power */
complex l;
double r;
{

	if( l.re == 0. && l.im == 0. )
		return( l );	/* undefined */

	l = clog( l );
	l.re = r*l.re;
	l.im = r*l.im;
	return( cexp(l) );
}


int ceq( l, r )	/* return TRUE if dp complex nos. equal */
complex l, r;
{
	if( l.re == r.re  &&  l.im == r.im )
		return( 1 );
	return( 0 );
}



complex cftoc( z )	/* complexf to complex conversion */
complexf z;
{
complex dz;

	dz.re = z.re;
	dz.im = z.im;
	return( dz );
}
		/*		*		*		*		*/

#ifdef ftoc
#undef ftoc
#endif

	complex ftoc( r, i )	/* convert floats to dp complex */
	double r, i;
	{
		complex dz;
		dz.re = r;	dz.im = i;
		return( dz );
	}

#ifdef ctof
#undef ctof
#endif

	double ctof( dz )	/* dp complex (real component) to float conversion */
	complex dz;
	{
		return( dz.re );
	}

#ifdef dconjg
#undef dconjg
#endif

	complex dconjg( dz )	/* return the dp complex conjugate */
	complex dz;			/* F77 extension */
	{
		dz.im = -dz.im;
		return( dz );
	}
		complex dconjg_( dz )
		complex *dz;
		{ return dconjg(*dz); }

#ifdef dimag
#undef dimag
#endif

	double dimag( dz )	/* return imaginary component of a dp complex no. */
	complex dz;			/* F77 extension */
	{
		return( dz.im );
	}
		double dimag_( dz )
		complex *dz;
		{ return dz->im; }

		/*		*		*		*		*/


complex csqrt( dz )	/* retn (positive) sq root of dp complex no. */
complex dz;
{
	double sqrtm,	/* sq rt of complex magnitude */
		theta;	/* the complex direct angle (in radians) */

	sqrtm = sqrt( cabs(dz) );
	theta = arg_dcpx( dz );
	dz.re = sqrtm*cos( theta/2.0 );
	dz.im = sqrtm*sin( theta/2.0 );
	if( dz.re < 0. ) {
		dz.re = -dz.re;
		dz.im = -dz.im;
		}
	else if( IS_FPZERO(dz.re) && dz.im < 0. )
		dz.im = -dz.im;
	return( dz );
}
	complex csqrt_( dz )
	complex *dz;
	{ return csqrt(*dz); }

complex clog( dz )	/* dp complex natural log */
complex dz;
{
	complex l;

	l.re = log( cabs(dz) );
	l.im = arg_dcpx(dz);	/* the direct angle */
	return( l );
}
	complex clog_( dz )
	complex *dz;
	{ return clog(*dz); }


complex cexp( dz )	/* dp complex exponential */
complex dz;
{
	double exp_re;

	if( dz.re == 0. )
		exp_re = 1.;
	else
		exp_re = exp(dz.re);

	if( dz.im == 0.0 ){	/* real only complex no. */
		dz.re = exp_re;
		return( dz );
		}
	dz.re = exp_re*cos(dz.im);
	dz.im = exp_re*sin(dz.im);
	return( dz );
}
	complex cexp_( dz )
	complex *dz;
	{ return cexp(*dz); }


complex csin( dz )	/* sine of dp complex no. */
complex dz;
{
	complex dzs;

	dzs.re = sin( dz.re )*cosh( dz.im );
	dzs.im = cos( dz.re )*sinh( dz.im );
	return( dzs );
}
	complex csin_( dz )
	complex *dz;
	{ return csin(*dz); }


complex ccos( dz )	/* cosine of dp complex no. */
complex dz;
{
	complex dzc;

	dzc.re = cos(dz.re)*cosh(dz.im);
	dzc.im = -sin(dz.re)*sinh(dz.im);
	return( dzc );
}
	complex ccos_( dz )
	complex *dz;
	{ return ccos(*dz); }


double cabs( dz )	/* dp complex absolute value */
complex dz;
{
	double s;

	s = dz.re*dz.re + dz.im*dz.im;
	return( sqrt(s) );
}
	double cabs_( dz )
	complex *dz;
	{ return cabs(*dz); }

	/* support for complex library functions */

double arg_dcpx( dz )	/* direct angle (in radians) of a complex no */
complex dz;		/* must be in range -PI < angle <= PI */
{
	double argc;
	if ( IS_FPZERO(dz.im) )
		argc = acos( dz.re/cabs(dz) );
	else if ( IS_FPZERO(dz.re) )
		argc = asin( dz.im/cabs(dz) );
	else {
		argc = atan( dz.im/dz.re );
		if( dz.re < 0. ) {
			if( dz.im < 0. ) 
				argc -= _PI_;	/* move from I to III quadrant */
			else
				argc += _PI_;	/* move from IV to II quadrant */
		}
	}
#ifdef DEBUG
	printf( "arg_dcpx( (%f,%f) ) = %f)\n", dz.re,dz.im, argc );
#endif
	return( argc );
}
