/* complxf.c - COMPLEX functions supporting translations by
 *         FOR_C (TM) from COBALT BLUE
 *    copyright Cobalt Blue, Inc., 1988 - 1992
 *               ALL RIGHTS RESERVED
 */

#include <math.h>
#include <fcrt.h>
#include <fcio.h>
#include <complx.h>

		/*		*		*		*		*/

#ifdef cfneg
#undef cfneg
#endif

	complexf cfneg( z )	/* complex unary minus operation */
	complexf z;
	{
		z.re = -z.re;	z.im = -z.im;
		return( z );
	}

		/*		*		*		*		*/

complexf cfadd( l, r )	/* retn the sum of two complex nos. */
complexf l, r;
{
	l.re += r.re;
	l.im += r.im;
	return( l );
}


complexf cfsub( l, r )	/* retn the difference of two complex nos. */
complexf l, r;
{
	l.re -= r.re;
	l.im -= r.im;
	return( l );
}


complexf cfmul( l, r )	/* retn the product of two complex nos. */
complexf l, r;
{
	complexf z;
	float f, g;

	f = l.re*r.re;
	g = l.im*r.im;
	z.re = f - g;

	f = l.re*r.im;
	g = l.im*r.re;
	z.im = f + g;

	return( z );
}


complexf cfdiv( l, r )	/* retn the quotient of two complex nos. */
complexf l, r;
{
	complexf z;
	float den;

	if( r.re == 0. && r.im == 0. )
		f_err( "complex division by 0." );
	den = r.re*r.re + r.im*r.im;

	z.re = ( l.re*r.re + l.im*r.im )/den;
	z.im = ( r.re*l.im - l.re*r.im )/den;
	return( z );
}


complexf cfpow( l, r )	/* general complex power */
complexf l, r;
{
	if( l.re == 0. && l.im == 0. )
		return( l );	/* undefined */

	return( cfexp( cfmul(r,cflog(l)) ) );
}


complexf cfpowf( l, r )	/* general complex raised to float power */
complexf l;
double r;
{

	if( l.re == 0. && l.im == 0. )
		return( l );	/* undefined */

	l = cflog( l );
	l.re = r*l.re;
	l.im = r*l.im;
	return( cfexp(l) );
}


int cfeq( l, r )	/* return TRUE if complex nos. equal */
complexf l, r;
{
	if( l.re == r.re  &&  l.im == r.im )
		return( 1 );
	return( 0 );
}


complexf ctocf( dz )	/* complex conversion to complexf */
complex dz;
{
	complexf z;

	z.re = dz.re;
	z.im = dz.im;
	return( z );
}

		/*		*		*		*		*/

#ifdef ftocf
#undef ftocf
#endif

	complexf ftocf( r, i )	/* convert floats to complexf */
	double r, i;
	{
		complexf z;
		z.re = r;	z.im = i;
		return( z );
	}

#ifdef cftof
#undef cftof
#endif

	double cftof( z )	/* complexf (real component) to float conversion */
	complexf z;
	{
		return( z.re );
	}

#ifdef conjg
#undef conjg
#endif

	complexf conjg( z )	/* return the complex conjugate */
	complexf z;
	{
		z.im = -z.im;
		return( z );
	}

#ifdef aimag
#undef aimag
#endif

	double aimag( z ) /* return the imaginary component of a complex no. */
	complexf z;
	{
		return( z.im );
	}

		/*		*		*		*		*/


complexf cfsqrt( z ) /* retn (positive) sq root of complex no. */
complexf z;
{
	double sqrtm, /* sq rt of complex magnitude */
		theta; /* the complex direct angle (in radians) */

	sqrtm = sqrt( cfabs(z) );
	theta = arg_cpx( z );
	z.re = sqrtm*cos( theta/2.0 );
	z.im = sqrtm*sin( theta/2.0 );
	if( z.re < 0. ) {
		z.re = -z.re;
		z.im = -z.im;
		}
	else if( IS_FPZERO(z.re) && z.im < 0. )
		z.im = -z.im;
	return( z );
}

	complexf cfsqrt_( z )
	complexf *z;
	{ return( cfsqrt(*z) ); }


complexf cflog( z )	/* complex natural log */
complexf z;
{
	complexf l;
	l.re = log( cfabs(z) );
	l.im = arg_cpx(z);	/* the direct angle */
	return( l );
}

	complexf cflog_( z )
	complexf *z;
	{ return( cflog(*z) ); }


complexf cfexp( z )	/* complex exponential */
complexf z;
{
	double exp_re;

	if( z.re == 0. )
		exp_re = 1.;
	else
		exp_re = exp(z.re);

	if( z.im == 0.0 ){	/* real only complex no. */
		z.re = exp_re;
		return( z );
		}
	z.re = exp_re*cos(z.im);
	z.im = exp_re*sin(z.im);
	return( z );
}

	complexf cfexp_( z )
	complexf *z;
	{ return( cfexp(*z) ); }


complexf cfsin( z )	/* sine of complex no. */
complexf z;
{
	complexf zs;
	float f, g;

	f = sin(z.re);
	g = cosh(z.im);
	zs.re = f*g;

	f = cos(z.re);
	g = sinh(z.im);
	zs.im = f*g;

	return( zs );
}

	complexf cfsin_( z )
	complexf *z;
	{ return( cfsin(*z) ); }


complexf cfcos( z )	/* cosine of complex no. */
complexf z;
{
	complexf zc;
	float f, g;

	f = cos(z.re);
	g = cosh(z.im);
	zc.re = f*g;

	f = -sin(z.re);
	g = sinh(z.im);
	zc.im = f*g;

	return( zc );
}

	complexf cfcos_( z )
	complexf *z;
	{ return( cfcos(*z) ); }


double cfabs( z )	/* complex absolute value */
complexf z;
{
	double s;

	s = z.re*z.re + z.im*z.im;
	return( sqrt(s) );
}

	double cfabs_( z )
	complexf *z;
	{ return cfabs(*z); }


	/* support for complex library functions */

double arg_cpx( z )	/* direct angle (in radians) of a complex no */
complexf z;
{
	double argc;
	if ( IS_FPZERO(z.im) )
		argc = acos( z.re/cfabs(z) );
	else if ( IS_FPZERO(z.re) )
		argc = asin( z.im/cfabs(z) );
	else {
		argc = atan( z.im/z.re );
		if( z.re < 0. ) {
			if( z.im < 0. ) 
				argc -= _PI_;	/* move from I to III quadrant */
			else
				argc += _PI_;	/* move from IV to II quadrant */
		}
	}
#ifdef DEBUG
	printf( "arg_cpx( (%f,%f) ) = %f)\n", z.re,z.im, argc );
#endif
	return( argc );
}
