/*  fmt.c -- function support for translated F77 format stmts
				by FOR_C (TM) from COBALT BLUE
      copyright Cobalt Blue, Inc., 1988 - 1994
                ALL RIGHTS RESERVED
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <fcrt.h>
#include <fcio.h>

#define SKIPDIGIT(p)	while( isdigit(*p) || isspace(*p) )	p++
#define SKIPSPACE(p)	while( isspace(*p) )	p++

/* NOTE: 
	A transl. format statement consists of a static (or external) 
		char array in the calling function that includes the ()'s 
		enclosing the original F77 FORMAT stmt.
	All left ('s inside the format are pushed on the paren[] stack with
		a repeat count of 1 if none is specified.  
	When the ) is detected,
		the repeat count is decremented by 1.  If the repeat const. >0, 
		then 'cf' is reset to paren[], (the '(').  If the repeat const == 0,
		then pdep is decremented by 1.
 */

struct {	/* field desciptor */
	int rept; /* no. of times the field descriptor is repeated */
	FEDIT fe; /* field edit descriptor */
	} fldesc;

	/* NOTE: As long as the F77 C I/O functs access formats in the same 
		way as a translated READ/WRITE stmt, then 1 field descriptor
		structure will suffice.  IF calls to the functs are ever made
		with intermixed format refs, THEN a separate field descriptor
		will need to be maintained for each format.
	 */


#ifdef PROTOTYPES
		/* local function prototypes */
	void fmtwarn( char*, char* ); /* output FORMAT warning message */
	int io_dirbeg( int, long, char* );	/* init. unit for direct I/O */
	void rddir_ubuf(UNITSTRU*); /* fill unit buffer w/next record */
	void rdseq_ubuf(UNITSTRU*); /* fill unit buffer w/next line */
	char *sedit( char*, F_FORMAT* ); /* eval S edit & retn end */
	char *tedit( char*, UNITSTRU* ); /* eval T edit & retn end */
	void term_rec( UNITSTRU*, int ); /* terminate record on format control */
#else
	void fmtwarn();
	int io_dirbeg();
	void rdseq_ubuf(), rddir_ubuf(), term_rec();
	char *sedit(), *tedit();
#endif

long curfmtno = 0;
char *curfmt = NULL;

#ifdef STATUS_WARNINGS
	char not_opn[]    = "Unit %d is NOT opened";
	char not_opndir[] = "Unit %d is not opened for DIRECT I/O";
	char not_opnfmt[] = "Unit %d is not opened for FORMATTED I/O";
	char not_opnseq[] = "Unit %d is not opened for SEQUENTIAL I/O";
	char stat_msg[120];
#	define STATUSWARN( m, u ) sprintf(stat_msg,m,u); f_warn(stat_msg);
#else
#	define STATUSWARN( m, u ) ; /*no op*/
#endif

				/*		*		*		*		*/

char *_fmt( stmt, fmts )	/* retn ptr to fmt str corres to stmt no. */
long stmt;
F77FMT *fmts;
{
	char msg[80];
	int i;

	curfmtno = stmt;	/* save as the 'current' format ref */

	for( i = 0; fmts[i].stmtn; i++ ){
		if( stmt == fmts[i].stmtn )
			return( fmts[i].fmt );
		}

	sprintf( msg, "format reference wasn't found for '%ld'", stmt );
	f_err( msg );
	return( "" );	/* to satisfy lint */
}

				/*		*		*		*		*/


void getunitfd( us, fe ) /* get & retn the field descriptor */
UNITSTRU *us;
FEDIT *fe; /* field edit descriptor */
{
F_FORMAT *f = us->ff;

/* NOTE: the scale factor will have been processed by io_fmtuntl() */

	if( fldesc.rept ){	/* ck for a repeated field */
		fldesc.rept--;
		fe->fd = fldesc.fe.fd;
		fe->w = fldesc.fe.w;
		fe->d = fldesc.fe.d;
		fe->x = fldesc.fe.x;
		fe->s = fldesc.fe.s;
		return;
		}
	else if( isdigit(*f->cf) ){	/* ck for a new repeated field */
		if( (fldesc.rept=atoi(f->cf)) <=0 )
			fmterr( "repeat const. must be >= 0", f->cf );
		SKIPDIGIT( f->cf );
		SKIPSPACE( f->cf );
		}
	else
		fldesc.rept = 1; /* default, rept field desc. 1 time */

	/* reset for cks below */
	fldesc.fe.w = 0; /* width must be > 0 */
	fldesc.fe.d = -1; /* decimal precision must be >= 0 */
	fldesc.fe.x = 2; /* default expo width */
	fldesc.fe.s = NULL; /*ptr to str (for B edit originally)*/

	fldesc.fe.fd = TOUPPER( *f->cf );	/* get the field descriptor */
	if( !isalpha(fldesc.fe.fd) )
		fmterr( "expecting letter for field descriptor", f->cf );
	f->cf++;				/* adv to the next char */
	SKIPSPACE( f->cf );

#ifdef PRIME_BUSEDIT
	if( fldesc.fe.fd == 'B' ){ /* PRIME Business edit */
		int n; /*count chars for bus_edit[]*/
		char *p;
		static char bus_edit[MAX_FMTLEN+1];
		bus_edit[0] = '\0';
		if( *f->cf != '\'' )
			fmterr( "expecting ' following B edit", f->cf );
		f->cf++; /*adv past sgl quote*/
		for( n=0, p=bus_edit; *f->cf &&  *f->cf != '\''; n++ ){
			if( n >= MAX_FMTLEN )
				fmterr( "B edit string is too big", f->cf );
			*p++ = *f->cf++;
			}
		*p = '\0';
		if( *f->cf != '\'' )
			fmterr( "missing closing ' in B edit", f->cf );
		fe->fd = fldesc.fe.fd;
		fe->s = fldesc.fe.s = bus_edit;
		return;
		}
#endif

	if( isdigit(*f->cf) ){	/* ck for the field width */
		if( (fldesc.fe.w=atoi(f->cf)) <= 0 )
			fmterr( "field width must be > 0", f->cf );
		SKIPDIGIT( f->cf );
		if( *f->cf == '.' ){ /* ck for the decimal precision */
			f->cf++;
			SKIPSPACE( f->cf );
			if( !isdigit(*f->cf) )
				fmterr( "expecting decimal precision", f->cf );
			fldesc.fe.d = atoi( f->cf );
			SKIPDIGIT( f->cf );
			}
		}
	else if( fldesc.fe.fd != 'A' && fldesc.fe.fd != 'Q' )
		fmterr( "expecting field width", f->cf );
		/* NOTE: only the A & Q edits may be given w/o a field width */

	switch( fldesc.fe.fd ){
		case 'I':
			if( fldesc.fe.w <= 0 )
				fldesc.fe.w = 12; /* assumed default width */
			break;
		case 'D':
		case 'E':
		case 'F':
		case 'G':
			if( fldesc.fe.d < 0 )
				fmterr("missing decimal precision w/floating point edit", 
					f->cf );
			if(fldesc.fe.fd=='E' || fldesc.fe.fd=='G'){
				/* ck for expo field */
				if( *f->cf == 'e' || *f->cf == 'E' ){
					f->cf++;	/* adv to digit */
					SKIPSPACE( f->cf );
					if( !isdigit(*f->cf) )
					  fmterr("expecting digit for exponent field",f->cf);
					fldesc.fe.x = atoi( f->cf );
					if( fldesc.fe.x <= 0 )
						fmterr("illegal exponent field width",f->cf);
					SKIPDIGIT( f->cf );
					}
				}
			break;
		case 'A':
		case 'L':
			if( fldesc.fe.d >= 0 )
				fmterr(
				 "the no. of digits can't be specified w/'A' or 'L' edit",
				 f->cf );
			break;
#ifdef DG_F77
		case 'B': /* Binary edit extension: Bw */
			if( fldesc.fe.w <= 0 )
				fldesc.fe.w = 12; /* assumed default width */
			break;
#endif
		case 'O': /* VAX Octal edit extension: Ow */
			if( fldesc.fe.w <= 0 )
				fldesc.fe.w = 12; /* assumed default width */
			break;
		case 'Q': /* VAX extension (pg. 12.24) */
			if( fldesc.fe.w != 0 )
				fmterr(
				 "Q edit can't be specified w/width or precision",
				 f->cf );
			break;
		case 'Z': /* VAX Hex edit extension: Zw */
			if( fldesc.fe.w <= 0 )
				fldesc.fe.w = 12; /* assumed default width */
			break;
		default:
			fmterr( "unsupported field descriptor", f->cf );
		}

	fldesc.rept--; /* decr the rept count, since it was used once */
	fe->w = fldesc.fe.w;
	fe->d = fldesc.fe.d;
	fe->x = fldesc.fe.x;
	fe->fd = fldesc.fe.fd;
	return;
}

	/*	*	*	*	*/

int io_dirbeg( u, rn, fmt )	/* init. unit for direct I/O */
int u;		/* the unit */
long rn;	/* the F77 record no. */
char *fmt;	/* the format */
{

	rn--;	/* adjust for C offsets */

	if( rn < 0 )
		unit_s[u].status = BAD_RECORD_NO;
	/* jump to the correct record */
	else if( fseek(unit_s[u].fp,rn*unit_s[u].reclen,SEEK_SET) )
		unit_s[u].status = BAD_RECORD_NO;	/* since unable to re-position */
	else
		io_fmtini( u, fmt );	/* init. the format struct */

	return( unit_s[u].status );
}

	/*	*	*	*	*/

void io_fmtini( u, fmt )	/* init. F77 format struct */
int u;
char *fmt;	/* pts to null str if no format specified */
{
	char msg[80];
	F_FORMAT *f;
	f = unit_s[u].ff;

	/* reset the field descriptor struct for upcoming format */
	fldesc.rept = fldesc.fe.fd = fldesc.fe.w = fldesc.fe.d = 0;

	for( ; fmt && isspace(*fmt); fmt++ )
		;	/* skip leading blanks */
	if( fmt == NULL || *fmt == '\0' ){	/* NO format given */
		sprintf(msg,"missing format in read/write initialization for unit %d",
			u );
		f_err( msg );
		}
	else if( *fmt != '(' ){
		sprintf( msg, "expecting '(' for format beginning: %.20s", fmt );
		f_err( msg );
		}

	/* init. the fmt */
	f->fmt = f->lp = curfmt = fmt;
	f->cf = fmt + 1;

	/* the ()'s enclosing the FORMAT stmt are included in the char array 
		and form the 1st () group (w/a repeat count of 1) */
	f->pdep = 0;
	f->prept[0] = 1;	/* only 1 repeat of the entire format (by def.) */
	f->paren[0] = fmt;
	f->lprept = 1; /*3/31/92*/
	f->lpdep = 0; /*3/31/92*/

	/* NOTE: init. the scale factor, blank & sign control w/each format! */
	f->scalef = 0;
	f->cblank = unit_s[u].blank;
	f->signc = NORMSIGN;

return;
}

	/*	*	*	*	*/

		/* FORMAT EDITING CONTROL */

char fw_gt0[] = "field width must be > 0";
char ill_digit[] = "illegal digit in format";
char premat_endrd[] = "premature end of format w/READ";
char premat_endwrt[] = "premature end of format w/WRITE";


void io_fmtuntl(us,finish,wrtf) /*process until field descriptor detected*/
UNITSTRU *us;
int finish,	/* if TRUE, terminate output when ':' detected */
	wrtf;	/* TRUE if write mode, FALSE if read mode */
{
char *nc;	/* ptr to 'next' char in the fmt (for peak ahead use) */
int digit,	/* flag indicating a digit was detected */
	r,	/* repeat const */
	sf,	/* scale factor */
	w;	/* field width */
F_FORMAT *f = us->ff;	/* ptr to F77 format info struct */

if( fldesc.rept > 0 )
	return;	/* since a repeated field descriptor is waiting */

while( *f->cf ){
	SKIPSPACE( f->cf );
	/* ck. field separators */
	if( *f->cf == '/' ){
		/* terminate record */
		term_rec( us, wrtf );
		if( us->status || us->ateof )
			return;
		f->cf++;
		continue;
		}
	else if( *f->cf == ',' ){
		f->cf++;
		continue;
		}
	else if( *f->cf == ':' ){
		if( finish )
			return;
		f->cf++;	/* colon ignored on input */
		continue;
		}
	/* else, NOT a field separator */

	/* ck edit control & field descriptors */
	nc = f->cf;
	w = r = sf = 0;	digit = NO;
	if( *nc=='-' || *nc=='+' || isdigit(*nc) ){ /* ck for digits */
		digit = YES;
		if( (*nc=='-' || *nc=='+') && !isdigit(*(nc+1)) )
			fmterr( "illegal '+-' in format", f->cf );
		/* assign digit to all possibilities */
		w = r = sf = atoi( nc );
		if( *nc=='-' || *nc=='+' )
			nc++;
		SKIPDIGIT( nc );
		}
	SKIPSPACE( nc );

	if( wrtf && us->setbuf )
		setubuf( us );

	switch( (int) *nc ){ /*switch on edit or field descriptors*/

		case '\'':	/* string const */
		case '"':
			{ int quote = *nc;
			if( !wrtf )
				fmtwarn("apostrophe edit skipped on input",f->cf);
			if( digit )
				fmterr( ill_digit, f->cf ); /*"illegal digit in format"*/
			/* skip 1st '\'', then write until closing '\'' */
			for( nc++ ; *nc; nc++ ){
				if( *nc == quote ){
					if( *(nc+1) == quote ){	/* escaped quote? */
						if( wrtf )
						  WRTCH( quote, us )
						nc++;	/* adv to 2nd quote */
						}
					else
						break;	/* end of the sgl quoted str reached */
					}
				else if( wrtf )
					WRTCH( *nc, us )
				}
			}
			break;

		case '(':	/* () fmt group */
			f->pdep++;
			if( f->pdep > MAX_PAREN )
				fmterr("()'s in format nested too deeply",f->cf);
			if( !digit )
				r = 1;	/* simple () grouping */
			if( (f->prept[f->pdep]=r) <= 0 )
				fmterr( "repeat count must be > 0", f->cf );
			f->paren[f->pdep] = nc;
			if( f->pdep <= 1 ){ /*3/31/92*/
				f->lp = nc;
				f->lprept = r;
				f->lpdep = f->pdep; /*3/31/92*/
				} /*3/31/92*/
			break;

		case '$': /*suppress \n at end of output - common extension*/
		case '\\': /* MS F77 version */
#ifdef DG_F77
		case '#':
#endif
			if( wrtf ){
				if( *us->ubuf != '0' && *us->ubuf != '1' )
					*us->ubuf = '$'; /*modify carriage control char*/
				}
			/* ignore on reads */
			break;

		case ')':	/* end of ()'d fmt group */
			if( digit )
				fmterr( ill_digit, f->cf ); /*"illegal digit in format"*/
			f->prept[f->pdep]--;	/* decr the repeat */
			if( f->prept[f->pdep] <= 0 ){
				f->pdep--;	/* decr the ()'s depth */
				if( f->pdep < 0 ){	/* ck if at end of the fmt */
					if( finish ){
						curfmtno = 0;	/* reset */
						return;	/* END of FORMAT, NORMAL EXIT */
						}
					/* STILL MORE ITEMS IN THE 'WRITE' STMT,
						RE-USE THE FORMAT STARTING AT THE LAST OPEN PAREN */
					nc = f->lp;	/* restart at last '(' */
					/* reset the () ptrs & counters */
					f->prept[f->pdep=f->lpdep] = f->lprept; /*3/31/92*/
					f->paren[f->pdep] = f->lp;
					term_rec( us, wrtf );	/* terminate record */
					if( us->status || us->ateof )
						return;
					}
				}
			else	/* fmt group still repeated */
				nc = f->paren[f->pdep];
			break;

		case 'X': /*skip field*/
		case 'x':
			if( w < 0 )	/* field width */
				fmterr( fw_gt0, f->cf ); /*"field width must be > 0"*/
			else if( w == 0 )
				w = 1; /* HP extension */
			if( w <= (us->ubufend - us->nxtu) )
				us->nxtu += w;
			else
				us->nxtu = us->ubufend; /* don't skip past end */
			break;

		case 'P': /*scale factor control*/
		case 'p':
			f->scalef = sf;
			break;

		case 'H': /*Hollerith field*/
		case 'h':
			if( !wrtf )
				fmtwarn("Hollerith edit skipped on input",f->cf);
			if( w <= 0 )
				fmterr( fw_gt0, f->cf ); /*"field width must be > 0"*/
			for( ; w > 0  &&  *nc; w-- ){
				nc++;
				if( wrtf )
				  WRTCH( *nc, us )
				}
			if( w > 0 )
				fmterr("premature end of format w/Hollerith field",f->fmt);
			break;

		case 'T':
		case 't':
			nc = tedit( nc, us ); /*evaluate T edit & retn end*/
			break;

		case 'S':
		case 's':
			nc = sedit( nc, f ); /*evaluate S edit & retn end*/
			break;

		case 'B': /*blank control*/
		case 'b':
			nc++; /*adv to 2nd char*/
			if( *nc == 'N' || *nc == 'n' )
				f->cblank = BLANK_NULL;
			else if( *nc == 'Z' || *nc == 'z' )
				f->cblank = BLANK_ZERO;
#if defined(PRIME_BUSEDIT)
			else if( *nc == '\'' ){ /*Prime Business edit*/
				nc--; /*move back to the B*/
				goto fld_descr;
				}
#elif defined(DG_F77)
			else if( isdigit(*nc) ){ /*DG Binary edit*/
				nc--; /*move back to the B*/
				goto fld_descr;
				}
#endif
			else
				f_warn("expecting 'N' or 'Z' following 'B' edit");
			break;

		case '\0': /* "premature end of format" */
			fmterr((wrtf ? premat_endwrt : premat_endrd), f->cf);
			break;

		default: /*all else*/
fld_descr:
			if( isalpha(*nc) ){ /*assume to a field descriptor*/
				if( r < 0 )
					fmterr( "illegal repeat const", f->cf );
				return; /*NORMAL EXIT WHEN IN THE MIDDLE OF A FORMAT*/
				}
			else
				fmterr("unexpected char(s) in format",f->cf);
			break;
		}	/* end of switch */

	f->cf = ++nc; /* incr ptr & reset saved position in fmt */
	}	/* end if edit control loop */

fmterr( (wrtf ? premat_endwrt : premat_endrd), f->cf );
	/*"premature end of format"*/
return;
}

	/*	*	*	*	*/

void rd_dirbeg( u, rn, fmt ) /* init. unit for direct reading */
int u;		/* the unit */
long rn;	/* the F77 record no. */
char *fmt;	/* the format */
{
	unit_s[u].status = 0;	/* clear any errors for new 'stmt' */

	/*NOTE: Can't automatically open a unit for DIRECT access, 
	 *   since the record length isn't specified; hence an error
	 *   will result with an UNopened unit!
	 */

	if( !inqu_opened(u) ){
		unit_s[u].status = NOT_OPEN;
		STATUSWARN( not_opn, u )
		}
	else if( unit_s[u].form != FORMATTED ){
		unit_s[u].status = NOT_OPEN_FOR_FORMAT;
		STATUSWARN( not_opnfmt, u )
		}
	else if( unit_s[u].access != DIRECT ){
		unit_s[u].status = NOT_OPEN_FOR_DIRECT;
		STATUSWARN( not_opndir, u )
		}
	else if( unit_s[u].form == FORMATTED && *fmt == '*' ){
		unit_s[u].status = NO_DIRECT_LSTDIR_IO;
#ifdef STATUS_WARNINGS
		sprintf(stat_msg,"Unit %d; List Directed DIRECT READ attempted", u);
		f_warn( stat_msg );
#endif
		}
	else if( io_dirbeg(u,rn,fmt) == 0 )	/* init. unit for direct I/O */
		rddir_ubuf( &unit_s[u] );	/* read in next record */

	return;
}

	/*	*	*	*	*/

void rd_seqbeg( u, fmt )	/* init. format struct for sequential I/O */
int u;
char *fmt;	/* pts to null str if no format specified */
{
	unit_s[u].status = 0;	/* clear any errors for new 'stmt' */

	if( !inqu_opened(u) ){
		unit_s[u].status = NOT_OPEN;
		STATUSWARN( not_opn, u )
		}
	else if( unit_s[u].form != FORMATTED ){
		unit_s[u].status = NOT_OPEN_FOR_FORMAT;
		STATUSWARN( not_opnfmt, u )
		}
	else if( unit_s[u].access != SEQUENTIAL ){
		unit_s[u].status = NOT_OPEN_FOR_SEQUENTIAL;
		STATUSWARN( not_opnseq, u )
		}
	else{
		io_fmtini( u, fmt );
		rdseq_ubuf( &unit_s[u] );	/* read in next record (i.e., line) */
		}

	return;
}

	/*	*	*	*	*/

char *sedit( nc, f )	/* eval S edit & return end */
char *nc;	/* pts to 'S' on entry */
F_FORMAT *f;
{		/* format:  S or SP or SS */
	int signchr;

	signchr = TOUPPER( *(nc+1) );	/* next char */

	if( signchr == 'P' ){		/* SP */
		f->signc = PLUSSIGN;
		nc++;	/* pt to 'P' */
		}
	else if( signchr == 'S' ){	/* SS */
		f->signc = NORMSIGN;
		nc++;	/* pt to 'S' */
		}
	else					/* S */
		f->signc = NORMSIGN;
		/* ANY other char after 'S' edit is left for io_fmtuntl() */

return( nc );
}


	/*	*	*	*	*/

char *tedit( nc, us )	/* eval T edit & return end */
char *nc;	/* pts to 'T' on entry */
UNITSTRU *us;
{
	int dir, w;

	nc++;	/* adv. to next char */
	if( isdigit(*nc) )
		dir = 'A';	/* absolute position */
	else{
		dir = TOUPPER( *nc );
		nc++;	/* adv. to digit */
		}
	w = atoi( nc );
	SKIPDIGIT( nc );
	nc--;	/* back up to the last digit */
	if( dir == 'A' ){	/* absolute */
		if( w >= us->reclen )
			us->nxtu = us->ubufend - 1;	/* will cause error */
		else
			us->nxtu = us->ubuf + w - 1;
		}
	else if( dir == 'L' ){	/* rel. left */
		if( w >= (us->nxtu - us->ubuf) )
			us->nxtu = us->ubuf;
		else
			us->nxtu -= w;
		}
	else if( dir == 'R' ){	/* rel. right */
		if( w >= (us->ubufend - us->nxtu +1) )
			us->nxtu = us->ubufend - 1;	/* will cause error */
		else
			us->nxtu += w;
		}
	else if( dir == 'T' ){	/* tab extension */
		w = w*8;	/* calc tab pos */
		if( w >= us->reclen )
			us->nxtu = us->ubufend - 1;	/* will cause error */
		else
			us->nxtu = us->ubuf + w - 1;
		}
	else
		f_warn( "illegal char in T edit" );

return( nc );
}

	/*	*	*	*	*/

void term_rec( us, wrtf )	/* terminate record on format control */
UNITSTRU *us;
int wrtf;
{

	/* insure an initiallized record in case record adv (/ edit) */
	if( wrtf && us->setbuf )
		setubuf( us );

	if( us->access==SEQUENTIAL ){
		if( wrtf && us->intu ){ /*internal write*/
			/*advance to the next 'record'*/
			us->ubuf += us->reclen + 1;
			us->ubufend = us->ubuf + us->reclen;
			/*NOTE: multi-'record' string assumed*/
			us->setbuf = YES; /*forces init. prior to actual write*/
			}
		else if( wrtf ){
			wrtubuf( us ); /*write out the buffer contents*/
			us->setbuf = YES; /*forces init. prior to actual write*/
			}
		else
			rdseq_ubuf( us ); /*read a new record (ie., line)*/
		}
	else{ /*direct access*/
		if( wrtf ){
			wrtubuf( us ); /*write out the buffer contents*/
			us->setbuf = YES; /*forces init. prior to actual write*/
			}
		else
			rddir_ubuf( us ); /*read in next record*/
		}

	return;
}

void setubuf( us ) /* initialize unit buffer to blanks for write */
UNITSTRU *us;
{ /* (must be called prior to writing each record) */

	memset( us->ubuf, ' ', us->reclen );
	us->nxtu = us->ubuf;
	*us->ubufend = '\0';

	us->setbuf = NO;
	return;
}

	/*	*	*	*	*/

void wrt_dirbeg( u, rn, fmt )	/* init. unit for direct writing */
int u;		/* the unit */
long rn;	/* the F77 record no. */
char *fmt;	/* the format */
{
	unit_s[u].status = 0;	/* clear any errors for new 'stmt' */

	/*NOTE: Can't automatically open a unit for DIRECT access, 
	 *   since the record length isn't specified; hence an error
	 *   will result with an UNopened unit!
	 */

	if( !inqu_opened(u) ){
		unit_s[u].status = NOT_OPEN;
		STATUSWARN( not_opn, u )
		}
	else if( unit_s[u].access != DIRECT ){
		unit_s[u].status = NOT_OPEN_FOR_DIRECT;
		STATUSWARN( not_opndir, u )
		}
	else if( unit_s[u].form != FORMATTED ){
		unit_s[u].status = NOT_OPEN_FOR_FORMAT;
		STATUSWARN( not_opnfmt, u )
		}
	else{
		unit_s[u].setbuf = YES; /* forces init. prior to actual write */
		io_dirbeg( u, rn, fmt );	/* init. unit for direct I/O */
		}

	return;
}

	/*	*	*	*	*/

void wrt_seqbeg( u, fmt ) /* init. format struct for sequential I/O */
int u;
char *fmt;	/* pts to null str if no format specified */
{

	unit_s[u].status = 0;	/* clear any errors for new 'stmt' */

	if( !inqu_opened(u) ){
#ifdef AUTO_OPEN_UNIT
		auto_openunit(u,FORMATTED);
#else
		unit_s[u].status = NOT_OPEN;
		STATUSWARN( not_opn, u )
		return;
#endif
		}

	if( unit_s[u].form != FORMATTED ){
		unit_s[u].status = NOT_OPEN_FOR_FORMAT;
		STATUSWARN( not_opnfmt, u )
		}
	else if( unit_s[u].access != SEQUENTIAL ){
		unit_s[u].status = NOT_OPEN_FOR_SEQUENTIAL;
		STATUSWARN( not_opnseq, u )
		}
	else{
		unit_s[u].setbuf = YES; /* forces init. prior to actual write */
		io_fmtini( u, fmt );
		}

	return;
}

			/*	*	*	*	*/

	/*	* FORMATTED INPUT Low Level Unit Buffer Routines *	*/

void rdseq_ubuf(us)	/* fill the unit buffer w/the next line (ie., record) */
UNITSTRU *us;
{		/* NOTE: Called at the START of every READ stmt transl
			 & under format, or list directed input control (when at eol) */
	char *p;


	if( us->ateof || us->status )
		return;
	else if( us->intu ){ /* internal read */
		us->ubuf += us->reclen + 1;
		us->ubufend = us->ubuf + us->reclen;
		/* NOTE: a multi-'record' string is assumed */
		}
	else{
		memset( us->ubuf, '\0', us->reclen );
		if( fgets(us->ubuf,us->reclen,us->fp) == NULL ){
			/* NOTE: fgets() incls the \n (which is deleted below) */
			if( ferror(us->fp) )
				us->status = READ_ERR; /* error reading from phys. device */
			else{
				us->status = -1;
				us->ateof = YES;
				}
			return;
			}
		for( p=us->ubuf; *p && *p != '\n'; p++ )
			;
#ifdef DG_F77
		if( us->inclnl && *p=='\n' )
			p++; /*keep the newline*/
		*p = '\0';	/*(normally) delete any newline*/
		/*count the chars read in*/
		us->chars_proc = strlen( us->ubuf );
#else
		*p = '\0';	/* delete any newline */
#endif
		}

	us->nxtu = us->ubuf; /*reset*/

	return;
}


void rddir_ubuf( us )	/* fill the unit buffer with the next record */
UNITSTRU *us;
{		/* NOTE: Called at the START of every direct READ stmt transl
			 & under format control */
	int n;

	us->nxtu = us->ubuf;	/* reset */
	memset( us->ubuf, '\0', us->reclen );

	if( us->status )
		return;
	else if( (n=fread(us->ubuf,sizeof(char),us->reclen,us->fp)) < us->reclen )
		us->status = READ_ERR;	/* error reading from phys. device */

#ifdef DG_F77
	us->chars_proc = n;
#endif
return;
}

	/*	*	*	*	*/

extern FILE *rterr_fp;
void _fmtmsg(); /* output FORMAT message */

void fmterr( msg1, fmtloc ) /* output FORMAT error message & quit */
char *msg1, *fmtloc;
{
	if( rterr_fp==NULL )
		rterr_fp = RTERR_FP; /*insure valid fp (in case reset)*/
	fputs( "FCRT ERROR: ", rterr_fp );
	_fmtmsg( msg1, fmtloc );
	exit(1);
}

void fmtwarn( msg1, fmtloc ) /* output FORMAT warning message */
char *msg1, *fmtloc;
{
	if( rterr_fp==NULL )
		rterr_fp = RTERR_FP; /*insure valid fp (in case reset)*/
	fputs( "FCRT WARNING: ", rterr_fp );
	_fmtmsg( msg1, fmtloc );
	return;
}

void _fmtmsg( msg1, fmtloc ) /* output FORMAT message */
char *msg1, *fmtloc;
{
	char msg[30], *p;

	fputs( msg1, rterr_fp );
	fputc( '\n', rterr_fp );
	if( curfmtno > 0 ){
		/* output the FORMAT number if known */
		sprintf( msg, "%ld FORMAT", curfmtno );
		fputs( msg, rterr_fp );
		}
	p = curfmt;
	if( p == NULL )
		p = fmtloc;
	for( ; *p ; p++ ){
		if( p == fmtloc )
			fputs( "<<?>>", rterr_fp );
		fputc( *p, rterr_fp );
		}
	fputc( '\n', rterr_fp );
	return;
}
