/* fprtunf.c - UNformatted Output with headers and trailers!
                      by COBALT BLUE (TM)
 */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1994
 *               ALL RIGHTS RESERVED
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>
#include <fcio.h>
#include <unfio.h>	/* UNFormatted I/O header */

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifdef NEED_COMPLEX
#include <complx.h>
#endif

		/*		*		*		*		*/

		/* PRIVATE variables and macros */
static char *pbuf,	/* ptr to the unformatted print buffer */
	*beg,	/* beginning of unformatted output in the buffer */
	*end,	/* absolute end of unformatted output in the buffer */
	*nxt;	/* next location for unformatted output in buffer */
static size_t pbufsiz;	/* size of the unf print buffer */

static long count= 0L, /*no. of bytes written out for current record*/
       unfprt_count= 0L, /*no. of bytes written out with last record*/
       os_start = -1L; /*starting offset of a partial record*/

static FILE *ofp;	/* the output file ptr */

#ifdef PROTOTYPES
	int dump_buf( int ); /*dump the buffer to the file - returns 0 if OK*/
#else
	int dump_buf();
#endif
#define PARTIAL	1
#define COMPLETE 0

	/* macros to get arg of type 't' & place it into the buffer.
	 *   (for args passed by value only, i.e., no ptrs)
	 * NOTE: 't' must be a valid C type!
	 * 		UNFWRT() must be enclosed in {}! ! !
	 */
#define UNFWRT(t)	{ DECLVAL( t )	\
					GETARG( t ) 	\
					CK_NOUT( t ) }

#define DECLVAL(t)	t v, *vp;
#define GETARG(t)	v = va_arg( args, t );
#define CK_NOUT(t)	CKBUF( sizeof(t) )	\
                    vp = (t*)nxt; \
                    *vp++ = v; \
                    nxt = (char*)vp;

/* ck the buffer for space, & dump it if not enough room */
#define CKBUF(siz)	if( (nxt+siz-1) >= end ) \
						if( dump_buf(PARTIAL) ) goto wrterr;

/* OUTCHR() first cks the buffer for space, 
 *   then copies 'c' into the next location in the buffer
 */
#define OUTCHR( c )    CKBUF( sizeof(char) )  *nxt++ = c

#define reset_exts     { nxt=NULL; count=0L; os_start= -1L; }


	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */
#if defined(__STDC__) || ANSI_C_TARGET
long fprtunf( FILE *fp, char *cs, ... )
#else
long fprtunf( fp, cs, va_alist )
FILE *fp;
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{
char *c,	/* local ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int prefix;
long cnt;	/* number of chars in the record (local copy) */
va_list args;	/* arg ptr */


	/* Initialization */
if( pbuf == NULL ){	/* ck for buffer existance */
	/* create buffer on 1st call */
	if( setprtunf((char*)malloc(MAX_UNFBUF),MAX_UNFBUF) )
		return( -1 );	/* unable to allocate buffer */
	}
if( nxt==NULL )
	nxt = beg;
if( fp==NULL || (fp != ofp && count > 0) ){ /*fp Error Checks*/
	reset_exts
	return( -2 );	/* file pointer error */
	/* (either NULL, or different from a prior continued output) */
	}
ofp = fp;	/* save file ptr in case record is continued */

#if defined(__STDC__) || ANSI_C_TARGET
va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
va_start(args);	/* initialize variable args */
#endif

/* read the control string - and the associated output */
for( c=cs; *c ; c++ ){
	if( *c == ' ' || *c == '\t' )	/* ignore spaces & tabs only */
		continue;
	ALIGN( nxt, beg );
	if( *c == '%' ){	/* syntax:  %[prefix]type  */
		c++;	/* advance past % */

		if( *c == 'l' || *c == 'h' || *c == 'L' )	/* ck for prefix */
			prefix = *c++;	/* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( (int) *c ){

			case '&':	/* obj being output, get size & address */
			case 'A':
				{ size_t sz;	/* size of ptr objects being output */
				sz = va_arg( args, size_t );	/* get the size in bytes */
				p = (char*) va_arg( args, void* );	/* get ptr to the obj */
				for( ; sz > 0; sz-- ){
					OUTCHR( *p++ );
					}
				/* NOTE: prefix is ignored */
				}
				break;

			case 's':	/* null terminated string output */
				p = va_arg( args, char* );	/* get str */
				for( ; *p; p++ ){
					OUTCHR( *p );
					}
				OUTCHR( '\0' );	/* incl the string termination */
				/* NOTE: prefix is ignored */
				break;

			case 'c':	/* char output */
				UNFWRT( char )
				/* NOTE: prefix is ignored */
				break;

			case 'd':	/* integer output */
			case 'i':
				if( prefix == 'l' )
					UNFWRT( long )
				else if( prefix == 'L' )
					UNFWRT( dbl_long )	/* Cray F77 - 64 bits */
				else if( prefix == 'h' ){
					/* short (passed as int - output as short) */
					DECLVAL( short )
					GETARG( int )
					CK_NOUT( short )
					}
				else
					UNFWRT( int )
				break;

			case 'f':	/* floating pt output */
				if( prefix == 'L' )
#ifdef LONG_DBL
					UNFWRT( longdbl )	/* long double */
#else
					{
					puts( "fprtunf() compiled withOUT long double support" );
					exit( 1 );
					}
#endif
				else if( prefix == 'l' )
					UNFWRT( double )
				else{	/* float (passed as double - output as float) */
					DECLVAL( float )
					GETARG( double )
					CK_NOUT( float )
					}
				break;

			case 'z':	/* complex output */
#ifdef NEED_COMPLEX
				if( prefix == 'l' )
					UNFWRT( COMPLEX16 )
				else
					UNFWRT( COMPLEX )
#else
				puts( "fprtunf() compiled withOUT complex support" );
				exit( 1 );
#endif
				break;

			case '+': /* %+ indicates this record is continued!*/
				cnt = count + (nxt - beg);
				return(cnt); /*output is continued, so don't write record yet*/
				/* NOTE: %+ automatically terminates the prt list */
				/* NOTE: prefix is ignored */
				/*break;*/

			case '\0':
				goto wrterr; /*premature termination of cs following '%'*/

			default: /*any other char following '%' is output literally*/
				OUTCHR( *c );
				break;
			}	/* end of switch */
		}	/* end of if-% */
	else{ /*any other char is output literally*/
		OUTCHR( *c );
		}
	}

/* write out the buffer, SINCE the record is complete */
dump_buf( COMPLETE );
/* finally, reset the fprtunf() externals and return */
cnt = count; /*save for the return*/
reset_exts
return( cnt );	/* output preformed satisfactory */
				/* NORMAL EXIT ! ! ! */

wrterr:	/* error actually writing the data to the file */
	reset_exts
	return( -3 );	/* write error */
}


	/*		*		*		*		*/

int dump_buf( p )	/* dump the buffer to the file, retn 0 if sucessful */
int p;	/* if TRUE, then a partial record, else complete record */
{
	char *b,	/* location for beginning of write from buffer */
		 *e;	/* write up to location 'e' */
	size_t cnt;	/* number of bytes (i.e., chars) to write out */


	if( p ){ /*partial record output ! ! !*/
		count += nxt - beg; /*keep track of the bytes in record*/
		if( os_start == -1 ){ /*1st partial output */
			/*save current offset as beginning of current record*/
			os_start = ftell( ofp );
			b = pbuf; /*reserves space for header*/
			}
		else /*subsequent partial output*/
			b = beg;
		e = nxt;   /*write out to current end of buffer*/
		nxt = beg; /*reset nxt only*/
		}

	else{  /*record completed (p==NO)*/
		ALIGN( nxt, beg );
		count += nxt - beg; /*keep track of the bytes in record*/
		if( os_start != -1 ){ /*previous partial output*/
			long os;
			os = ftell( ofp ); /*save current offset*/
			/*write size of record (ie., count) to os_start*/
			fseek( ofp, os_start, SEEK_SET );
			if( fwrite(&count,sizeof(long),1,ofp) != 1 )
				return( 1 ); /*write error*/
			/*& re-position to 'current' offset*/
			fseek( ofp, os, SEEK_SET );
			b = beg; /*beginning of write from pbuf[]*/
			}
		else{ /*NO prior output*/
			/*place size of record in beginning of buffer*/
			b = pbuf;
			*(long*)b = count;
			}

		/* place record count at end of record (for trailer)
		 *  and advance nxt
		 */
		e = nxt;
		*(long*)e = count;
		e += sizeof(long);
		unfprt_count = count; /*save record size for len_unfprt()*/
		}

	/*write out buffer (from designated beginning to end)*/
	cnt = (size_t)( e - b );
	if( fwrite(b,sizeof(char),cnt,ofp) != cnt )
		return( 1 );	/* write error */

return( 0 );	/* successful write */
}

	/*    *    *    *    */

long len_unfprt() /*return record length of unformatted print*/
{                 /*corres. to last fprtunf(), or fseekwrt() call*/
	return( unfprt_count );
}

	/*    *    *    *    */

int setprtunf( bufptr, size )	/* initialize the fprtunf() buffer */
char *bufptr;
size_t size;
{
	if( bufptr == NULL || size == 0 )
		return( 1 );	/* error */

	if( pbuf )	/* free any existing buffer */
		free( pbuf );
	pbuf = bufptr;
	pbufsiz = size;

	/* mark the beginning & end of unformatted output in buffer */
	beg = pbuf + sizeof(long);
	end = pbuf + pbufsiz - sizeof(long);

	return( 0 );	/* buffer successfully initialized */
}
