/* fscanld.c - F77 Style, List Directed formatted Input! */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1993
 *               ALL RIGHTS RESERVED
 */

/*#define DEBUG*/

#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>
#include <fcio.h>

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifdef STRICT_LDREAD
	/*strict list directed read - data must match input fields*/
	char *ck_ldint();
#else /*pass the string through untouched*/
#	define ck_ldint(s)	s
#endif

#ifdef NEED_COMPLEX
#include <complx.h>
COMPLEX16 dzr;	/* repeated complex constant */
int getld_cpx();	/* get COMPLEX list directed input */
#endif

	/* PRIVATE local functions & macros */
#ifdef PROTOTYPES
	int atolog(char*);
	void get_ldstr(char*,int,int); /*get list directed input string*/
	int nxtline(void); /*read the next line for list directed input*/
	char *prep_fp(char*); /*prep. fp number, (map F77 style nos. to C fmt)*/
	char *rdld_str(char*,int,int);	/* get list directed input */
	void skip_rd(void);	/* skip leading spaces & read next line if necessary */
	int _scanld( char*, va_list );
#else
	char *rdld_str(), *prep_fp();
	int atolog();
	void get_ldstr(), skip_rd();
	int nxtline();
	int _scanld();
#endif

/* macro to assign the data value to an argument, given,
	- type 't' 
	- function call 'fc' to actually read & return the input
 */
#define ASSGN_LD(t,fc) 	{ t *tp; \
			tp = va_arg( args, t* ); \
			for( ; nele > 0; nele-- ) { \
				p = rdld_str(str,MAX_LDSTR-1,0); \
				if( *p ) *tp = fc; \
				tp++; } }

static FILE *ldfp;

static char buf[MAXLINE+1],	/* list dir scan line buffer */
			*nxt;	/* ptr to the list directed line buffer */

static int scanld_err;

	/* Input List Directed Parameters */
#define MAX_LDSTR	36	/* max str size for numeric input */
#define MAX_LDREPT 80	/* max repeated str size */
long ld_r;	/* repeat const */
int cld_r;	/* flag for complex list directed repeat */
char ld_rstr[MAX_LDREPT];	/* repeated data str */

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */

#if defined(__STDC__) || ANSI_C_TARGET
int fscanld( FILE *fp, char *cs, ... )
#else
int fscanld( fp, cs, va_alist ) /* list directed fmt'd INput */
FILE *fp;
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{
	int nf=0;  /* number of fields actually read & assigned */
	va_list args;  /* arg ptr */

	/* Initialization */
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs); /* initialize variable args */
#else	/* UNIX System V C */
	va_start(args); /* initialize variable args */
#endif

	if( fp == NULL ){	/* fp Checks */
		scanld_err = NOT_OPEN;
		return( nf );	/* file pointer error */
		}

	if(	strncmp(cs,"%+",2) == 0 ){
		if( fp != ldfp ){
			scanld_err = READ_ERR;
			return( nf ); /* error, can't continue w/different fp */
			}
		else if( scanld_err )
			return( nf ); /* prior error, don't continue */
		/* continued list directed input - read only when necessary */
		cs += 2; /* skip %+ */
		}
	else{
		ldfp = fp; /* save file ptr for ck in case read is continued */
		/* read the next line & reset the buffer pointer */
		scanld_err = 0;	/* reset */
		nxtline();
		ld_r = cld_r = 0;	/* reset any repeated values */
		ld_rstr[0] = '\0';
		}
		/* end of initialization */

	/* process the control list & the var arg list */
	nf = _scanld( cs, args );

	va_end( args );
	return( nf ); /* return the no. of fields read */
}  /*=================  end of fscanld()  =================*/


	/*	*	*	*	*/

#if defined(__STDC__) || ANSI_C_TARGET
int sscanld( char *sbuf, char *cs, ... )
#else
int sscanld( sbuf, cs, va_alist ) /* list directed fmt'd INput */
char *sbuf;
char *cs; /* control string (indicates type & no. of args) */
va_dcl
#endif
{
	int nf=0;  /* dummy return value for _scanld() */
	va_list args;  /* arg ptr */

	/* Initialization */
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs); /* initialize variable args */
#else		/* UNIX System V C */
	va_start(args); /* initialize variable args */
#endif

	scanld_err = 0; /* reset */

	if( sbuf == NULL ){
		scanld_err = READ_ERR;
		return( 0 );
		}

	/* read the next line & reset the buffer pointer */
	ldfp = NULL;
	nxt = sbuf;
	/* get input solely from the supplied buffer */
	buf[0] = '\0';
	ld_r = cld_r = 0; /* reset any repeated values */
	ld_rstr[0] = '\0';
	/* end of initialization */

	/* process the control list & the var arg list */
	nf = _scanld( cs, args );

	va_end( args );
	return( nxt - sbuf ); /* return the no. of chars read */
}  /*=================  end of sscanld()  =================*/


	/*	*	*	*	*/

/* process control string for list directed args */

#if defined(__STDC__) || ANSI_C_TARGET
int _scanld( char *cs, va_list args )
#else
int _scanld( cs, args )
char *cs;	/* control string (indicates type & no. of args) */
va_list args;
#endif
{
char str[MAX_LDSTR];	/* str for list directed numeric input */
char *c,		/* local ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int nf=0,		/* number of fields actually read & assigned */
	blankpad,	/* flag indicating 'B' edit was chosen for str input */
	prefix;
size_t nele;		/* the number of elements to read w/a given object */

/* read the control string - and the associated INput */
for( c=cs; *c ; c++ ){
	blankpad = 0;

	if( isspace(*c) )	/* ignore whitespace */
		continue;

	else if( *c == '%' ){ /* syntax:  %[prefix]type  */

		c++;	/* advance past % */

		if( *c == 'A' ){ /* ck for array indication prefix */
			c++; /* advance past 'A' */
			nele = va_arg( args, size_t );	/* get the no. of elements */
			}
		else
			nele = 1;	/* default */

		if( *c == 'l' || *c == 'L' || *c == 'h' ) /*ck for prefix*/
			prefix = *c++; /* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( *c ){

			case 'B':
				blankpad = 1; /* blank pad strings */
			case 'S':
				blankpad = 1; /* blank pad strings Even w/S */
				{ size_t lensiz;
				/* get char ptr and following char length size */
				p = (char*) va_arg( args, void* ); /* get ptr to the obj */
				lensiz = va_arg( args, size_t ); /* get the char length */
				for( ; nele > 0; nele-- ){
					rdld_str( p, lensiz-1, 1 );
					if( blankpad )
						strpad( p, ' ', lensiz-1 );
					p += lensiz; /* advance the str */
					}
				}
				break;

			case 's': /* null terminated string INput - C style! */
				p = va_arg( args, char* );	/* get target str location */
				rdld_str( p, INT_MAX, 1 );
				/* NOTE: prefixes (including 'A') are ignored;
						 '\0' is copied to target */
				break;

			case 'c':	/* char INput */
				p = va_arg( args, char* ); /* get target char location */
				for( ; nele > 0; nele-- ){
					rdld_str( str, 1, 0 );
					if( *str )
						*p = *str;
					p++;
					}
				/* NOTE: prefix is ignored */
				break;

			case 'b': /*INTEGER*1 INput*/
				ASSGN_LD( byte, atoi( ck_ldint(str) ) )
				break;

			case 'd':	/* integer INput */
			case 'i':
				if( prefix == 'l' )
#ifdef DEBUG
						{ long *tp;
						tp = va_arg( args, long* );
						for( ; nele > 0; nele-- ) {
							p = rdld_str( str, MAX_LDSTR-1, 0 );
							if( *p )
								*tp = atol( ck_ldint(str) );
							tp++; /* adv ptr */
							}
						}
#else
					ASSGN_LD( long, atol( ck_ldint(str) ) )
#endif
				else if( prefix == 'L' )
					ASSGN_LD( dbl_long, atol( ck_ldint(str) ) )
				else if( prefix == 'h' )
					ASSGN_LD( short, atoi( ck_ldint(str) ) )
				else	/* default int */
					ASSGN_LD( int, atoi( ck_ldint(str) ) )
				break;

			case 'f':	/* floating pt INput */
				if( prefix == 'L' )
					ASSGN_LD( longdbl, atof(prep_fp(str)) )
				else if( prefix == 'l' )
					ASSGN_LD( double, atof(prep_fp(str)) )
				else
					ASSGN_LD( float, (float) atof(prep_fp(str)) )
				break;

			case 'T':	/* LOGICAL8 object */
				ASSGN_LD( LOGICAL8, atolog( str ) )
				break;

			case 't':	/* LOGICAL16 or LOGICAL32 object */
				if( prefix == 'l' )
					ASSGN_LD( LOGICAL32, atolog( str ) )
				else if( prefix == 'L' )
					ASSGN_LD( LOGICAL64, atolog( str ) )
				else
					ASSGN_LD( LOGICAL16, atolog( str ) )
				break;

#ifdef NEED_COMPLEX
			case 'z':	/* complex INput */
				if( prefix == 'l' ){
					COMPLEX16 *tp;
					tp = va_arg( args, COMPLEX16* );
					for( ; nele > 0; nele-- ) {
						if( getld_cpx() ) 
							*tp = dzr;
						tp++;	/* adv ptr */
						}
					}
				else{
					COMPLEX *tp;
					tp = va_arg( args, COMPLEX* );
					for( ; nele > 0; nele-- ) {
						if( getld_cpx() ){
							tp->re = dzr.re;
							tp->im = dzr.im;
							}
						tp++;	/* adv ptr */
						}
					}
				break;
#endif

			case '\0':	/* premature termination of cs following '%' */
				scanld_err = ILLEGAL_INPUT;
				break;

			default:	/* any other char following '%' must be matched */
				for( ; *nxt==' ' || *nxt=='\t'; nxt++ )
					; /* skip white space in input buffer */
				if( *c != *nxt++ )	/* input char doesn't match control str */
					scanld_err = ILLEGAL_INPUT;
				break;
		}	/* end of switch */
		if( scanld_err )
			break;	/* out of for_loop */
		nf++;	/* indicate a field was correctly read */
	}	/* end of if-% */
	else{	/* any other non-white char must be matched exactly */
		rdld_str( str, 1, 0 );
		if( *c != *str )	/* input char doesn't match control str */
			scanld_err = ILLEGAL_INPUT;
			break;
	}
	if( scanld_err )
		break;	/* out of for_loop */
} /* end of for-loop */

return( nf );	/* return the no. of fields read */

}  /*=================  end of _scanld()  =================*/


	/*	*	*	*	*/

int scanlderr()	/* return the scanld() error status */
{	/* if >0 then error, if <0 then eof, 0 if no error */
	return( scanld_err );
}

#ifdef STRICT_LDREAD
char *ck_ldint( str ) /*ck list directed int input for correctness*/
char *str;
{
	char *s=str;
	if( *s=='+' || *s=='-' )
		s++; /*skip +- sign for check*/
	for( ; *s && scanld_err==0; s++ ){
		if( !isdigit(*s) )
			scanld_err = ILLEGAL_INT_INPUT;
		}
	return( str ); /*str passed through*/
}
#endif

	/*	*	*	*	*/


int atolog( s )	/* get formatted logical */
char *s;
{	/* NOTE: a logical true const is defined as either 'T' or 't' & may be
		optionally preceded by spaces or a '.', any chars after the T are 
		ignored;  ANY OTHER char is considered FALSE!
	  Return: 1 is returned if True, else 0.
	  The F77 logical constants are: "T", "F", ".TRUE.", ".FALSE."
	 */

	if( *s == '.' )	/* ignore optional leading '.' */
		s++;	
	/* determine the logical input 
		(based only on the 1st char, the rest are ignored) */
	if( *s == 'T' || *s == 't' )
		return( 1 );
	else if( *s=='F' || *s=='f' )
		return( 0 );
#ifdef STRICT_LDREAD
	else /*invalid char in data field*/
		scanld_err = ILLEGAL_LOGICAL_INPUT;
#endif

	/* any other input is considered FALSE */
	return( 0 );
}

	/*	*	*	*	*/

char *prep_fp( str ) /*prep. fp number, i.e., map F77 style nos. to C fmt*/
char *str;
{
char *s;

	/* find 1st non-digit (& non-decimal pt) */
	s = (str[0]=='-' || str[0]=='+'? str+1 : str);
	for( ; isdigit(*s) || *s == '.'; s++ )
		;
	/* end of field, or possible exponent */

	/* map F77 style expo (dD+-) to C fmt */
	if( *s == 'D' || *s == 'd' || *s == '+' )
		*s = 'e';
	else if( *s == '-' )
		inschr( 'e', s );
#ifdef STRICT_LDREAD
	else if( *s=='E' || *s=='e' )
		; /*no chg required*/
	else if( *s ) /*any other char is illegal*/
		scanld_err = ILLEGAL_FP_INPUT;
#endif
	return( str );
}


	/*	*	*	*	*/

#ifdef NEED_COMPLEX

	/*		[rept_const*](real_const, imag_const)  [,][ ] 	*/


int getld_cpx()	/* get COMPLEX list directed input */
{	/* returns TRUE if complex constant read, else FALSE */
	char istr[MAX_LDSTR];
	char *p;		/* peak ahead ptr */

	if( ld_r > 0 ){	/* return repeated data */
		ld_r--;
		if( cld_r == 0 ){
			/* consider the non-complex str as the complex real component */
			dzr.re = atof( ld_rstr );
			dzr.im = 0.;
			cld_r = 1;	/* now that a complex constant is defined */
			}
		return( TRUE );
		}

	/* reset so 0z or null returned with '/' input */
	dzr.re = dzr.im = 0.;
	ld_rstr[0] = '\0'; istr[0] = '\0';

	/* skip leading spaces (& read next line if necessary) */
	skip_rd();

	if( isdigit(*nxt) ){	/* ck if repeated data:	"digit *" */
		/* peak ahead to see if followed by '*' */
		for( p=nxt; isdigit(*p); p++ )
			;
		if( *p == '*' ){	/* then a repeat const */
			ld_r = atol( nxt );
			nxt = p+1;	/* data starts immediately after "r*" */
			}
		else
			ld_r = 1L;	/* NOT a repeat const */
		}
	else if( *nxt == '/' )
		return( FALSE ); /*remaining entries are null*/

	if( *nxt == '(' ){	/* get '(' */
		nxt++;
		skip_rd();
		get_ldstr(ld_rstr,MAX_LDSTR-2,0); /*get REAL input*/
		if( *nxt == ',' )	/* get ',' */
			nxt++;
		skip_rd();
		get_ldstr( istr, MAX_LDSTR-2, 0); /*get IMAG input*/
		if( *nxt == ')' )	/* get ')' */
			nxt++;
		}
	else if( *nxt == ',' )
		; /*null field*/
	else{ /* treat entry as real component */
#ifdef STRICT_LDREAD
		scanld_err = ILLEGAL_COMPLEX_INPUT; /*invalid complex no.*/
#endif
		get_ldstr( ld_rstr, MAX_LDSTR-2, 0 );
		}

	/* determine the complex number input */
	dzr.re = atof( prep_fp(ld_rstr) );
	dzr.im = atof( prep_fp(istr) );

	if( ld_r > 1L )
		ld_r--;	/* decrement the repeat const */
	cld_r = 1;	/* since a complex repeat will be defined */

	/* eat trailing ',' */
	if( *nxt == ',' )
		nxt++;	/* eat comma */

	return( ld_rstr[0] ? TRUE : FALSE );
}

#endif

		/*	*	LIST DIRECTED INPUT MGT.	*	*/


char *rdld_str( str, w, strflag )	/* get list directed input */
char *str;
int w;	/* max. str length */
int strflag;
{
	char *r,	/* ptr to repeated str */
		*s=str,	/* working copy of str */
		*p;		/* peak ahead ptr */

	cld_r = 0;	/* disable the complex list directed flag */
	if( ld_r > 0L ){	/* return repeated data str */
		if( !ld_rstr[0] && strflag )
			; /* don't chg orig str */
		else{ /* copy the repeated char data to str 's' */
			for( r=ld_rstr; w-- > 0  &&  (*s++ = *r++); )
				;
			}
		if( --ld_r <= 0L )
			ld_rstr[0] = '\0';
		return( str );
		}

	/* skip leading spaces (& read next line if necessary) */
	skip_rd();

	if( isdigit(*nxt) ){	/* ck if repeated data:	"digit *" */
		/* peak ahead to see if followed by '*' */
		for( p=nxt; isdigit(*p); p++ )
			;
		if( *p == '*' ){	/* then a repeat const */
			ld_r = atol( nxt );
			nxt = p+1;	/* data starts immediately after "r*" */
			}
		else
			ld_r = 1L;	/* NOT a repeat const */
		}
	else if( *nxt == '/' ){	/* ck if remaining entries are null */
		*s = '\0';
		return( str );
		}

	get_ldstr( s, w, strflag );	/* get the list directed input string */

	if( ld_r > 1L ){
		strncpy( ld_rstr, s, MAX_LDREPT-1 );
		ld_rstr[MAX_LDREPT-1] = '\0';	/* insurance */
		}
	ld_r--;	/* decrement the repeat const */

	/* eat trailing ',' */
	if( *nxt == ',' )
		nxt++;	/* eat comma */

	return( str );
}

	/*	*	*	*	*/


void get_ldstr( s, w, strflag )	/* get list directed input string */
char *s;	/* save string in 's' */
int w;	/* max. str length */
int strflag;	/* if TRUE, don't null-terminate unquoted input */
{	/* NOTE: the calling funct will decide if this is valid input */
	int c, quote, bksl_chrs();

	if( *nxt == '\'' || *nxt == '"' ){	/* ck for sgl/dbl quoted str input */
		quote = *nxt++;	/* save & then skip lead quote */
		while( *nxt ){
			c = *nxt;
			if( c == '\n' && ldfp ){
				if( nxtline() )
					continue;
				c = *nxt;
				}

			if( c == '\\' ){	/* allow for C-style back slash char seqs */
				nxt++;	/* move past back slash */
				if( *nxt == '\n' && ldfp ){
					if( nxtline() )
						continue;
					}
				c = bksl_chrs( *nxt );	/* map C bk slash chars */
				}
			else if( c == quote ){	/* ck for embedded quote, '' or "" */
				nxt++;	/* move past quote */
				if( *nxt != quote )
					break;	/* terminating quote (normal loop exit) */
				}
			if( w-- > 0 )	/* copy str subj to size limitations */
				*s++ = c;
			nxt++;
			}	/* end of while loop */
		*s = '\0';
		}
	else{	/* assume simple space or comma separated (or '/' terminated) */
		int n = 0;
		while( *nxt && *nxt != ',' && !isspace(*nxt) && *nxt != '/' 
			 && *nxt != '(' && *nxt != ')' ){
			if( w-- > 0 ) {
				*s++ = *nxt;
				n++;
				}
			nxt++;
			}
		if( n || !strflag )
			*s = '\0';
		}

	/* skip trailing spaces */
	for( ; *nxt == ' ' || *nxt == '\t'; nxt++ )
		;

	return;
}

			/*	*	*	*	*/

int bksl_chrs( c )	/* map C bk slash chars sequences to sgl chars */
int c;
{	/* NOTE: the \ has already been detected */

	switch( c ){
		case 'a':	return( '\a' );
		case 'b':	return( '\b' );
		case 'f':	return( '\f' );
		case 'n':	return( '\n' );
		case 'r':	return( '\r' );
		case 't':	return( '\t' );
		case 'v':	return( '\v' );
		default:
			/* typically: '  "  \  ?
				which are returned as themselves */
			break;
		}

	return( c );
}


			/*	*	*	*	*/

void skip_rd()	/* skip leading spaces (& read next line if necessary) */
{

	while( *nxt && isspace(*nxt) ){
		if( *nxt == '\n' && ldfp ){
			nxtline();
			continue;
			}
		nxt++;
		}
	return;
}

int nxtline() /*read the next line for list directed input*/
{             /* returns true if EOF or ERR */

	if( fgets((nxt=buf),MAXLINE,ldfp) == NULL ){
		buf[0] = '\0';
		if( ferror(ldfp) )
			scanld_err = READ_ERR;
		else
			scanld_err = -1;
		return 1;
		}
	return 0;
}

