/* fscanunf.c - UNformatted Input with headers and trailers! */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1995
 *               ALL RIGHTS RESERVED
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>
#include <fcio.h>
#include <unfio.h>	/* UNFormatted I/O header */

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>	/* ANSI variable arg macros */
#else
#include <varargs.h>
#endif

#ifdef NEED_COMPLEX
#include <complx.h>
#endif

	/*		*		*		*		*/

		/* PRIVATE variables and macros */

static char *sbuf, /* ptr to the scan unformatted buffer */
	*end,	/* current end of the record in the buffer */
	*nxt;	/* next location for unformatted input in buffer */
static size_t sbufsiz; /* size of the scan unf buffer */


static FILE *ifp; /* the input file ptr */

#ifdef PROTOTYPES
	int fill_buf(); /*fill the buffer from the file - returns 0 if OK*/
#else
	int fill_buf();
#endif
#define NEWREC	1
#define CONTINUE 0

	/* macros to get arg of type 't' from the buffer & assign it to the 
		location pointed to.
	   NOTE: 't' must be a valid C type!
	   		UNFRD() must be enclosed in {}! ! !
	 */
#define UNFRD(t)	{ DECLTARGT( t )	\
					GETARG( t ) 	\
					CK_NIN( t ) }

#define DECLTARGT(t) t *tp;
#define GETARG(t)	tp = va_arg( args, t* );
#define CK_NIN(t)	CKBUF( sizeof(t) )	\
					*tp = *(t*)nxt; \
					nxt += sizeof(t);

/* ck buffer if next object is completely present - and refill if not */
#define CKBUF(sz)	if( (nxt+sz-1) >= end ) \
						goto rderr;


	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */


#if defined(__STDC__) || ANSI_C_TARGET
int fscanunf( FILE *fp, char *cs, ... )
#else
int fscanunf( fp, cs, va_alist )/* unformatted variable length record INput */
FILE *fp;
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{
char *c,	/* working ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int nf=0,	/* number of fields actually read & assigned */
	prefix;
va_list args;	/* arg ptr */

	/* Initialization */
if( sbuf == NULL ){	/* ck for buffer existance */
	/* create buffer on 1st call */
	if( (sbuf=malloc(sbufsiz=MAX_UNFBUF)) == NULL )
		memerr("unable to allocate buffer for fscanunf()");
	}

#if defined(__STDC__) || ANSI_C_TARGET
va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
va_start(args);	/* initialize variable args */
#endif

for( ; *cs == ' ' || *cs == '\t'; cs++ )	/* skip leading blanks/tabs */
	;
if( fp == NULL )	/* fp Checks */
	return( -2 );	/* file pointer error */
else if( fp != ifp  &&  (strncmp(cs,"%+",2)==0) ){
	fprintf(stderr,"ERROR: fscanunf() read continuation without prior read\n");
	return( -2 ); /* error, can't continue w/different fp */
	}
ifp = fp; /* save file ptr for ck in case read is continued */
if( strncmp(cs,"%+",2) == 0 ){ /* ck control str for continuation */
	cs += 2;	/* skip %+ */
	/* ck that something is in the buffer */
	if( nxt >= end )
		return( nf ); /* since there's nothing in the buffer */
	/* NOTE: the entire record is assumed to be in the buffer
	 * (i.e., in memory), and the file offset advanced
	 * to the next record!  Hence, if nxt is at the end of
	 * of the buffer, there's nothing left to read, so return.
	 */
	}
else{	/* read the next record */
	if( fill_buf() )
		return( nf );
	}
	/* End of Initialization */

/* read the control string - and the associated INput */
for( c=cs; *c ; c++ ){
	if( *c == ' ' || *c == '\t' ) /* ignore spaces & tabs only */
		continue;
	ALIGN( nxt, sbuf );
	if( *c == '%' ){ /* syntax:  %[prefix]type  */
		c++;	/* advance past % */

		if( *c == 'l' || *c == 'h' || *c == 'L' ) /* ck for prefix */
			prefix = *c++; /* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( (int) *c ){

			case '&': /* obj being INput, get size & address */
			case 'A':
				{ size_t sz;
				sz = va_arg( args, size_t ); /* get the size in bytes */
				p = (char*) va_arg( args, void* ); /* get ptr to the obj */
				for( ; sz > 0; sz-- ){
					CKBUF( sizeof(char) );
					*p++ = *nxt++;
					}
				}
				/* NOTE: prefix is ignored */
				break;

			case 's': /* null terminated string INput */
				p = va_arg( args, char* );	/* get target str location */
				do {
					CKBUF( sizeof(char) );
					} while( (*p++ = *nxt++) != '\0' );
				/* NOTE: prefix is ignored; '\0' is copied to target */
				break;

			case 'c': /* char INput */
				UNFRD( char )
				/* NOTE: prefix is ignored */
				break;

			case 'd': /* integer INput */
			case 'i':
				if( prefix == 'l' )
					UNFRD( long )
				else if( prefix == 'L' )
					UNFRD( dbl_long )
				else if( prefix == 'h' )
					UNFRD( short )
				else	/* default int */
					UNFRD( int )
				break;

			case 'f': /* floating pt INput */
				if( prefix == 'L' )
#ifdef LONG_DBL
					UNFRD( longdbl )	/* long double */
#else
					{
					puts( "fscanunf() compiled withOUT long double support" );
					exit( 1 );
					}
#endif
				else if( prefix == 'l' )
					UNFRD( double )
				else
#ifdef DEBUG
					{ /* expanded macro logic for debugging */
					float *tp;
					tp = va_arg( args, float* );
					if( (nxt+sizeof(float)-1) >= end )
						goto rderr;
					*tp = *(float*)nxt;
					nxt += sizeof(float);
					}
#else
					UNFRD( float )
#endif
				break;

			case 'z': /* complex INput */
#ifdef NEED_COMPLEX
				if( prefix == 'l' )
					UNFRD( COMPLEX16 )
				else
					UNFRD( COMPLEX )
#else
				puts( "fscanunf() compiled withOUT complex support" );
				exit( 1 );
#endif
				break;

			case '+': /* ignore %+ since NOT at the beginning of cs */
				break;

			case '\0':
				goto rderr;	/* premature termination of cs following '%' */

			default: /* any other char following '%' must be matched */
				CKBUF( sizeof(char) );
				if( *c != *nxt++ )
					goto rderr; /* input char doesn't match control str */
				break;
			}	/* end of switch */
		nf++; /* indicate a field was correctly read */
		}	/* end of if-% */
	else{ /* any other char must be matched exactly */
		CKBUF( sizeof(char) );
		if( *c != *nxt++ )
			goto rderr; /* input char doesn't match control str */
		}
	}

	return( nf );

rderr: /* error reading data from the file (or in the control str) */
	/*NOTE: may wish to add a warning message here. */
	return( nf );	/* return the no. of fields read */
}

	/*    *    *    *    */

static long unfrd_recsiz; /*record size assoc. w/last fscanunf() call*/
                           /*(or last fadvunf() call)*/

	/*    *    *    *    */

int fadvunf( fp ) /*advance to next unformatted record header*/
FILE *fp;        /*returns 0 if skipped, EOF if error or EOF*/
{
	/* NOTE: file offset assumed to be positioned at the beginning
			 of the record header ! ! ! */

	if( fread(&unfrd_recsiz,sizeof(long),1,fp) != 1 )
		return( EOF );	/*EOF or error (use ferror() to discern)*/

	if( fseek(fp,unfrd_recsiz + sizeof(long),SEEK_CUR) )
		return( EOF );	/*seek error*/

	return( 0 );
}

	/*    *    *    *    */

int fill_buf() /* fill the unit buffer w/the next record */
{
	/* UNformatted sequential record:	N_ITEMS record_contents N_ITEMS  */
	long tohdr; /* diff betw current file offset & next header */
	/* NOTE:  tohdr	= recsiz + TRAILER_SIZ 
	 *              = recsiz + sizeof(long)
	 */

	/* read the N_ITEMS Header */
	if( fread(&unfrd_recsiz,sizeof(long),1,ifp) != 1 )
		return( 1 ); /* either EOF or error (use ferror() to discern) */
	/* the file offset is now positioned at the start of the record itself */
	/* init tohdr, */
	tohdr = unfrd_recsiz + sizeof(long);	/* record_size + trailer_size */

	if( tohdr > sbufsiz ){ /*record exceeds current buffer*/
		/* so make the buffer bigger! */
		if( (sbuf=realloc(sbuf,sbufsiz=tohdr)) == NULL )
			memerr("unable to reallocate fscanunf() buffer for large record");
		}
	else if( tohdr < MAX_UNFBUF  && sbufsiz > MAX_UNFBUF ){
		/* shrink large buffer from prior read */
		if( (sbuf=realloc(sbuf,sbufsiz=MAX_UNFBUF)) == NULL )
			memerr("unable to downsize fscanunf() buffer");
		}

	/* read the record contents (plus the N_ITEMS trailer) into buffer */
	if( fread(sbuf,sizeof(char),(size_t)tohdr,ifp) != tohdr )
		return( 1 ); /* read error */

	nxt = sbuf; /* start processing buffer at the beginning */
	end = sbuf + unfrd_recsiz; /* mark the record end */
	return( 0 );
}

long len_unfscan() /*retn the length of the last fscanunf() record read*/
{
	return( unfrd_recsiz );
}

