/* inqu.c - support for translated F77 inquire stmt by FOR_C (TM) */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1994
 *               ALL RIGHTS RESERVED
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fcrt.h>
#include <fcio.h>

	/* translations generated from the F77 INQUIRE stmt,

		INQUIRE 'ilist'		C funct
		---------------		--------------
		ACCESS = acc		inqu_acc(u,acc);
		BLANK = blnk		inqu_blank(u,blnk);
		DIRECT = dir		inqu_dir(u,dir);
		ERR = s				if( UERR(u) ) goto L_s;		-- macro
		EXIST =	ex			ex = !access(filename,0);
		  "		 "			ex = inqu_existu( u );
		FORM = fm			inqu_form(u,fm);
		FORMATTED = fmt		inqu_fmt(u,fmt);
		IOSTAT = ios		ios = UERR(u);   [or,  ios = ferror(UFP(u));]
		NAME = fn			inqu_name(u,fn);
		NAMED = nmd			nmd = inqu_named(u);
		NEXTREC nr			nr = ftell(UFP(u))/URECL(u) + 1;
		NUMBER = num		num = inqu_unit(fn);
		OPENED = od			od = inqu_opened(u);
		RECL = rcl			rcl = URECL(u);
		SEQUENTIAL = seq	inqu_seq(u,seq);
		UNFORMATTED = unf	inqu_unf(u,unf);
	 */

				/*	*	*	*	*/


void inqu_acc( u, acc )	/* get the access assoc. w/a unit */
int u;						/* F77 equivalent */
char *acc;
{
	if( !inqu_opened(u) )
		*acc = '\0';
	if( unit_s[u].access == SEQUENTIAL )
		strcpy( acc, "SEQUENTIAL" );
	else if( unit_s[u].access == DIRECT )
		strcpy( acc, "DIRECT" );
	return;
}


void inqu_blank( u, blnk )	/* get the blank treatment setting */
int u;
char *blnk;
{
	if( !inqu_opened(u) )
		*blnk = '\0';
	if( unit_s[u].blank == BLANK_NULL )
		strcpy( blnk, "NULL" );
	else if( unit_s[u].blank == BLANK_ZERO )
		strcpy( blnk, "ZERO" );
	return;
}


void inqu_dir( u, dir )	/* store "YES" in dir if DIRECT access */
int u;
char *dir;
{
	if( !inqu_opened(u) )
		*dir = '\0';
	if( unit_s[u].access == DIRECT )
		strcpy( dir, "YES" );
	else
		strcpy( dir, "NO" );
	return;
}


int inqu_existu( u )	/* return 1 if unit exists */
int u;
{
	if( inqu_opened(u) )
		return( 1 );

	return( 0 );
}


void inqu_fmt( u, fmt )	/* store "YES" in fmt if FORMATTED io */
int u;
char *fmt;
{
	if( !inqu_opened(u) )
		*fmt = '\0';
	if( unit_s[u].form == FORMATTED )
		strcpy( fmt, "YES" );
	else
		strcpy( fmt, "NO" );
	return;
}


void inqu_form( u, form )	/* return the form mode assoc. w/a unit */
int u;
char *form;
{
	if( !inqu_opened(u) )
		*form = '\0';
	if( unit_s[u].form == FORMATTED )
		strcpy( form, "FORMATTED" );
	else if( unit_s[u].form == UNFORMATTED )
		strcpy( form, "UNFORMATTED" );
	return;
}

	/*	*	*	*	*/

void inqu_name( u, fn )	/* get the filename assoc. w/a unit */
int u;
char *fn;
{
	if( !inqu_opened(u) )
		*fn = '\0';
	else if( unit_s[u].fname && *unit_s[u].fname )	/* a named unit */
		strcpy( fn, unit_s[u].fname );
	else
		*fn = '\0';	/* unnamed */
	return;
}


int inqu_named( u )	/* return 1 if the unit has a filename assoc. w/it */
int u;
{
	if( !inqu_opened(u) )
		;	/* retn 0 (below) for not open/valid */
	else if( unit_s[u].fname && *unit_s[u].fname != '\0' )
		return( 1 );

	return( 0 );
}


int inqu_opened( u )	/* retn 1 if unit is opened */
int u;	/* unit to ck */
{

	if( u == _ERR_ )	/* (when inqu_unit() fails) */
		return( 0 );

	if( !VALID_UNIT(u) && u != STDIN_UNIT && u != STDOUT_UNIT
#ifdef STD_PRN
	    && u != STDPRN_UNIT
#endif
	   )
		badunit( u );

	/* if a non-NULL file ptr is assoc. w/this unit, then it is opened! */
	if( IS_INTERNAL(u) )
		; /*don't check internal units*/
	else if( unit_s[u].fp == NULL ){
		if( u==STDIN_UNIT || u==STDOUT_UNIT )
		  f_warn("failure to pre-open STDIN/OUT_UNIT - missing f77ini() call");
		return( 0 );
		}

	return( 1 );
}


void inqu_seq( u, seq )	/* store "YES" in seq if SEQUENTIAL access */
int u;
char *seq;
{
	if( !inqu_opened(u) )
		*seq = '\0';
	if( unit_s[u].access == SEQUENTIAL )
		strcpy( seq, "YES" );
	else
		strcpy( seq, "NO" );
	return;
}


void inqu_unf( u, unf )	/* store "YES" in unf if UNFORMATTED io */
int u;
char *unf;
{
	if( !inqu_opened(u) )
		*unf = '\0';
	if( unit_s[u].form == UNFORMATTED )
		strcpy( unf, "YES" );
	else
		strcpy( unf, "NO" );
	return;
}


int inqu_unit( fn )	/* retn the unit assoc. w/a filename */
char *fn;
{
	int i;
	char *p;

	p = prep_fn( fn ); /* trim any leading & trailing blanks */
	
	for( i=0; i < MAX_UNITS; i++ ){
		if( inqu_opened(i) && unit_s[i].fname ){
			if( strcmpip(p,unit_s[i].fname) == 0 )
				return( i );
			}
		}

	return( _ERR_ );
}
