/* ntstr.c - convert length specified memory to null terminated string */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1996
 *               ALL RIGHTS RESERVED
 */

/*Note: ntstr() was renamed to ntS(), 1/31/96, for compactness and
 * uppercased for uniqueness.
 */

#include <stdio.h>	/* for NULL definition */
#include <stdlib.h>

#include <fcrt.h>

#define MX_NTS	255
static char *_nts[MX_NTS];
static int next=0;
	/* rotate through _nts[] ptrs to avoid clobbering prior values when
	 * used in an argument list
	 */

char *ntS( obj, n )
void *obj;
long n;	/*the size of the object to null terminate*/
{
	int i;
	char *o=obj, *t;

	if( n <= 0 )
		return( "" ); /*null string since invalid length*/
	else if( obj == NULL )
		return( "" );

	next = (next >= MX_NTS) ? 0 : next; /*cycle through the index*/
		                               /*(round robin style)*/
	if( _nts[next] )
		free( _nts[next] ); /*free prior string*/

	if( (_nts[next]=(char*)malloc((int)(n+1))) == NULL )
		f_err( "ntS: can't allocate space" );
		/*NOTE: +1 to allow for \0 at the end of the new string*/

	for( i=0,t=_nts[next]; i < n; i++ ){
		/*copy the object to the null terminated space*/
		if( *o )
			*t = *o;
		else
			*t = ' '; /*convert interior nulls to blanks*/
		t++; o++;
		}
	*t = '\0'; /*null terminate the string*/

	return( _nts[next++] );
}


void free_nts() /*free up the entire _nts[] array*/
{              /*(normally called at exit),   added 10/14/96*/
	int i;

	for( i=0; i < MX_NTS; i++ ){
		if( _nts[i] )
			free( _nts[i] ); /*free any allocated space*/
		_nts[i] = NULL; /*just in case not called at exit*/
		}
	next = 0;

	return;
}
