/* opncls.c - support for translated F77 open/close stmts by FOR_C (TM) */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1994
 *               ALL RIGHTS RESERVED
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fcrt.h>
#include <fcio.h>

	/* funct calls from F77 translation to C,
		F77 OPEN 'olist'	C translation generated
		--------------		-----------------------
		[unit=]u			open_unit(u,filename,sta,fm,acc,rl);
		FILE=filename		 "
		STATUS=sta			 "
		ACCESS=acc, RECL=rl	 "
		FORM=fm				 "
		BLANK=blnk			unit_blank(u,blnk);
		IOSTAT=ios			ios = UERR(u); [or, ios = ferror(UFP(u));]
		IOSTAT=ios			ios = inqu_stat(u);
		ERR=s				if( UERR(u) )	goto L_s;

		F77 CLOSE 'clist'	C translation generated
		--------------		----------------------
		[unit=]u			close_unit(u,sta);
		STATUS=sta			 "
		IOSTAT=ios			inqu_stat(u,ios);
		ERR=s				if( UERR(u) )	goto L_s;

	 */

/* typical FORTRAN-IV pre-opened units - change if necessary */
#define IN_UNIT 5
#define OUT_UNIT 6
  /*NOTE: IN_UNIT2 & OUT_UNIT2 may be defined as additional 
   *  pre-opened units.  For Data General F77, they are Units 11
   *  and 10 respectively.
   */

	/* local functions */
#ifdef PROTOTYPES
	void _set_status(int,char*); /*merely reset f77typ given status*/
	void ini_units(void); /*init. the F77 units: STDIN_UNIT & STDOUT_UNIT*/
	char *trim(char*); /*trim leading & trailing spaces, retn trimmed str*/
#else
	void _set_status();
	void ini_units();
	char *trim();
#endif

	/*	*	*	*	*/

void auto_openunit( u, form ) /* (possibly) auto. open unit w/temp file */
int u; /* unit number */
int form; /* FORMATTED or UNFORMATTED */
{
	char msg[120];

#ifdef INSURANCE
	if( u == _ERR_ )
		return;
	else if( inqu_opened(u) )
		return; /* the norm - valid fp on opened unit */
	/*NOTE: inqu_opened() cks for a valid unit number*/
#else
	/* assume unit to be valid and unopened;
	 * initial call from wrt_seqbeg() makes a prior inqu_opened() call
	 */
#endif

	if( u==STDIN_UNIT || u==STDOUT_UNIT ){
		f_warn("pre-opening STDIN/OUT_UNIT - since missing f77ini() call");
		ini_units(); /*init. the F77 units: STDIN_UNIT & STDOUT_UNIT*/
		return;
		}

	/* warn the user we're about to open the unit */
	if( form==FORMATTED )
		sprintf( msg, 
		  "opening temp seq., formatted file on UNopened unit %d", u );
	else
		sprintf( msg, 
		  "opening temp seq., UNformatted file on UNopened unit %d", u );
	f_warn( msg );

	if( open_unit( u,"","",
	          (form==FORMATTED?"FORMATTED":"UNFORMATTED"),"",0) != _ERR_ )
		return;

	sprintf( msg, "can't open temp. file for unit %d", u );
	f_err( msg );
}

	/*	*	*	*	*/

void badunit( u ) /* report a bad unit error and quit */
int u;
{
	char msg[80];
	sprintf( msg, "bad unit number used (%d), valid units: 1 to %d",
		u, MAX_UNITS-1 );
	f_err( msg ); /*no return!*/
}

void *badunitp( u )
int u;
{
	badunit( u );
	return( (void*)NULL );
}

       /* * * * */

 /* * *CLOSE functs* * */

       /* * * * */

	static int _cls_unit(u,sta) /*close w/o ck for valid or opened unit*/
	              /*returns 1 if successfully closed, else _ERR_*/
	int u; /*F77 unit*/
	char *sta; /*CLOSE status*/
	{
	/*NOTE: _close_unit() is an internal routine only, it does
	 *      not make any calls to f_err(), thus it can be used
	 *      in clos_units() without worry of recursion.
	 */
		unit_s[u].status = 0; /* reset */
		/* NOTE: by reseting the unit status any prior errors are 
				effectively forgotten */

		if( unit_s[u].fp==NULL )
			; /*already closed*/
		else if( (unit_s[u].fp==stdin || unit_s[u].fp==stdout)
			&& (u==IN_UNIT || u==OUT_UNIT) )
			; /* DON'T CLOSE stdin or stdout - merely zero */
#ifdef INOUT2_UNIT
		else if( (unit_s[u].fp==stdin || unit_s[u].fp==stdout)
		        && (u==IN_UNIT2 || u==OUT_UNIT2) )
			; /* DON'T CLOSE stdin or stdout - merely zero */
#endif
		else if( fclose(unit_s[u].fp) == _ERR_ ){
			unit_s[u].status = UNABLE_TO_CLOSE;
			return( _ERR_ );
			}
		unit_s[u].fp = NULL; /* unit is now 'free' */

		if( unit_s[u].fname==NULL || *unit_s[u].fname=='\0' )
			; /* no filename to delete */
		else if( *sta && *sta != '\0' ){
			/*closing status has precedence over opening type*/
			if( strcmpip(sta,"KEEP") == 0 )
				; /* don't delete the file w/its closing */
			else if( strcmpip(sta,"DELETE")==0 ){
				if( remove(unit_s[u].fname) )
					unit_s[u].status = UNABLE_TO_DELETE;
					/* ANSI C: remove() returns non-ZERO if error */
				}
			else
				unit_s[u].status = BAD_CLOSE_STATUS;
			}
		else if( unit_s[u].f77typ==SCRATCH ){
			/* delete files opened with scratch attribute */
			if( remove(unit_s[u].fname) )
				unit_s[u].status = UNABLE_TO_DELETE;
			}
		if( unit_s[u].status )
			return( _ERR_ );

		reset_unit( u ); /*reset basic unit params*/

		return( 1 );
	}

	void reset_unit( u ) /*reset basic unit params after close*/
	int u; /*unit number*/
	{
		if( VALID_UNIT(u) ){
			unit_s[u].fp = NULL; /*ins, in case called directly*/
			if( unit_s[u].fname )
				free( unit_s[u].fname );
			unit_s[u].fname = NULL;
			if( unit_s[u].ubuf )
				free( unit_s[u].ubuf );	/* rele the unit buffer */
			unit_s[u].ubuf = unit_s[u].ubufend = NULL;
			if( unit_s[u].ff )
				free( (char*) unit_s[u].ff ); /* rele the format struct */
			unit_s[u].ff = NULL;
			}
		return;
	}

void clos_units() /*close all the units (left) open - Public funct */
{				
	int u, err;
	for( u=0, err=0; u < MAX_UNITS; u++ ){
		if( unit_s[u].fp )
			err = _cls_unit(u,"")==_ERR_ ? YES : err;
		}
	if( err )
		fputs("unable to properly close one or more units",stderr);
		/*NOTE: f_err() isn't called to eliminate recursion possibility*/
	return;
}

int close_unit( u, sta )  /*CLOSE unit u - Public function*/
int u;                   /*returns 1 if successfully closed, else _ERR_*/
char *sta;
{
	if( !inqu_opened(u) || 
		u >= INTERNAL_UNIT ) /* can't CLOSE an internal file */
		return( _ERR_ ); /* or other pre-attached unit */
	return( _cls_unit(u,sta) );
}

         /*  *  *  *  */

/* * OPEN a file on a Unit * */

         /*  *  *  *  */

int open_sequnit( u, pfn, sta, fm, pos ) /* open a file on a F77 unit */
int   u;	/* unit to open */
char *pfn,	/* the (possibly padded) filename */
	*sta,	/* open state: OLD, NEW, SCRATCH, UNKNOWN ("" for default) */
			/*    extensions: RO, WO */
	*fm,	/* format mode ("" for default) */
	*pos;	/* F90 position: 'REWIND', 'ASIS', 'APPEND' */
{		/* open_sequnit() returns 1 if successfully opened, else _ERR_ */
	int reopen=NO;
	char mode[5], fn[120];

	if( pfn==NULL || *pfn=='\0' ) /* no name given, create name */
		sprintf( fn, "scr%d.dat", u );
	else{
		strcpy(fn,prep_fn(pfn)); /*trim blanks (& lowercase) filename*/
		if( fn[0] == '\0' )
			sprintf( fn, "scr%d,dat", u );
		}

	if( !VALID_UNIT(u) )
		badunit( u );

	if( inqu_opened(u) ){
		if( strcmp(fn,unit_s[u].fname) == 0 )
			reopen = YES;
		else /*different names! close & re-open w/new name*/
			close_unit( u, "" );
		}
	else
		unit_s[u].status = 0; /*not currently open, start w/fresh slate*/
	if( unit_s[u].status )
		return( _ERR_ );

	/* initialize unit structure */
	ini_unitstru( u, fn, fm, "", 0 );

	if( unit_s[u].status )
		return( _ERR_ );

	/* NOTE: all files are opened for both reading AND writing! */
	unit_s[u].action = READWRITE; /*default*/

	if( reopen ){
		/*don't actually re-open the file, since already open*/
		_set_status( u, sta ); /*merely reset the f77typ*/
		}
	else if( strcmpip(sta,"OLD")==0  /* open for reading */
	       || strcmpip(sta,"RO")==0 || strcmpip(sta,"READ")==0 ){
		unit_s[u].f77typ = OLD;
		strcpy( mode, "r+" );
		if( strcmpip(sta,"RO")==0 || strcmpip(sta,"READ")==0 ){
			strcpy( mode, "r" ); /*Read Only - F90 ACTION='READ'*/
			unit_s[u].action = READONLY;
			}
		if( unit_s[u].form==UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
			unit_s[u].status = UNABLE_TO_OPEN_OLD;
		/*file successfully opened*/
		}
	else if( strcmpip(sta,"NEW")==0  /*open for writing*/
	         || strcmpip(sta,"WO")==0 || strcmpip(sta,"WRITE")==0 ){
		unit_s[u].f77typ = NEW;
		strcpy( mode, "w+" );
		if( strcmpip(sta,"WO")==0 || strcmpip(sta,"WRITE")==0 ){
			strcpy( mode, "w" ); /* WRITE ONLY - F90 ACTION='WRITE' */
			unit_s[u].action = WRITEONLY;
			}
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
			unit_s[u].status = UNABLE_TO_CREATE_NEW;
		/* NOTE: any existing file is deleted */
		}
	else if( strcmpip(sta,"SCRATCH") == 0 ){ /* open scratch file */
		unit_s[u].f77typ = SCRATCH;
		strcpy( mode, "w+" );
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
			unit_s[u].status = UNABLE_TO_CREATE_SCRATCH;
		/* NOTE: any existing file is deleted; 
		 *   the main difference betw. NEW and SCRATCH is that 
		 *   SCRATCH files are deleted by default when they are 
		 *   closed
		 */
		}
	else if( *sta == '\0'    /* the default, "" */
		     || strcmpip(sta,"UNKNOWN") == 0 ){
		/* open unknown (existance) file */
		unit_s[u].f77typ = UNKNOWN;
		/* first try to open it as an existing file */
		strcpy( mode, "r+" );
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL ){
			/* then try to open it as a new file */
			strcpy( mode, "w+" );
			if( unit_s[u].form == UNFORMATTED )
				strcat( mode, "b" ); /* open in binary mode */
			if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
				unit_s[u].status = UNABLE_TO_OPEN_UNKNOWN;
			}
		}
	else	/* other status */
		unit_s[u].status = BAD_OPEN_STATUS;

	if( unit_s[u].access==APPEND )
		unit_s[u].access = SEQUENTIAL;
		/*APPEND considered the same as SEQUENTIAL*/

	/*evaluate the POSITION attribute*/
	/* NOTE: alternate names are from DG F77 */
	if( strcmpip(pos,"REWIND")==0 || strcmpip(pos,"START")==0 ){
		if( fseek(unit_s[u].fp,0L,SEEK_SET) != 0 )
			unit_s[u].status = CANT_REPOSITION;
		}
	else if( strcmpip(pos,"ASIS")==0 || strcmpip(pos,"CURRENT")==0 )
		; /*don't chg the file position*/
	else if( strcmpip(pos,"APPEND")==0 || strcmpip(pos,"END")==0 ){
		if( fseek(unit_s[u].fp,0L,SEEK_END) != 0 )
			unit_s[u].status = UNABLE_TO_SEEK_TO_EOF;
		}

	if( unit_s[u].status )
		return( _ERR_ );
	return( TRUE ); /* unit successfully opened & initialized! */
}

	void _set_status( u, sta ) /*merely reset the f77typ given status*/
	int u;
	char *sta;
	{
		if( strcmpip(sta,"OLD")==0  /* open for reading */
	       || strcmpip(sta,"RO")==0 || strcmpip(sta,"READ")==0 )
			unit_s[u].f77typ = OLD;
		else if( strcmpip(sta,"NEW")==0  /*open for writing*/
		         || strcmpip(sta,"WO")==0 || strcmpip(sta,"WRITE")==0 )
			unit_s[u].f77typ = NEW;
		else if( strcmpip(sta,"SCRATCH") == 0 ) /* open scratch file */
			unit_s[u].f77typ = SCRATCH;
		else if( *sta=='\0'    /* the default, "" */
			     || strcmpip(sta,"UNKNOWN") == 0 )
			unit_s[u].f77typ = UNKNOWN;
		else	/* other status */
			unit_s[u].status = BAD_OPEN_STATUS;
		return;
	}

	/*	*	*	*	*/

int open_unit( u, pfn, sta, fm, acc, rl ) /* open a file on a F77 unit */
int   u;	/* unit to open */
char *pfn,	/* the (possibly padded) filename */
	*sta,	/* open state: OLD, NEW, SCRATCH, UNKNOWN ("" for default) */
			/*    extensions: RO, WO */
	*fm,	/* format mode ("" for default) */
	*acc;	/* the access mode ("" for default) */
int rl; 	/* the record length (if direct access) */
{		/* open_unit() returns 1 if successfully opened, else _ERR_ */
char mode[5], fn[120];

	if( pfn==NULL || *pfn=='\0' ) /* no name given, create name */
		sprintf( fn, "scr%d.dat", u );
	else{
		strcpy(fn,prep_fn(pfn)); /*trim blanks (& lowercase) filename*/
		if( fn[0] == '\0' )
			sprintf( fn, "scr%d,dat", u );
		}

	if( !VALID_UNIT(u) )
		badunit( u );

	if( inqu_opened(u) ){
		if( strcmp(fn,unit_s[u].fname) == 0 )
			return( TRUE ); /* same name, take no action */
		/* different names! close & re-open w/new name */
		close_unit( u, "" );
		}
	else
		unit_s[u].status = 0; /*not currently open, start w/fresh slate*/
	if( unit_s[u].status )
		return( _ERR_ );

	/* initialize unit structure */
	ini_unitstru( u, fn, fm, acc, rl );

	if( unit_s[u].status )
		return( _ERR_ );

	/* free unit at this point */
		/* NOTE: all files are opened for both reading AND writing! */
	unit_s[u].action = READWRITE; /*default*/
	if( strcmpip(sta,"OLD")==0 
	       || strcmpip(sta,"RO")==0 ){ /* open for reading */
		unit_s[u].f77typ = OLD;
		strcpy( mode, "r+" );
		if( strcmpip(sta,"RO")==0 || strcmpip(sta,"READ")==0 ){
			strcpy( mode, "r" ); /* Read Only - F90 ACTION='READ'*/
			unit_s[u].action = READONLY;
			}
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
			unit_s[u].status = UNABLE_TO_OPEN_OLD;
			/* file MUST exist */
		else{ /*file successfully opened*/
			if( unit_s[u].access == APPEND ){
				/*VAX/others - extension*/
				if( fseek(unit_s[u].fp,0L,SEEK_END) != 0 )
					unit_s[u].status = UNABLE_TO_SEEK_TO_EOF;
				/*NOTE: APPEND is treated the same as SEQUENTIAL
				 *  after initially positioning to EOF.
				 */
				}
			}
		}
	else if( strcmpip(sta,"NEW")==0 
	         || strcmpip(sta,"WO")==0 ){ /*open for writing*/
		unit_s[u].f77typ = NEW;
		strcpy( mode, "w+" );
		if( strcmpip(sta,"WO") == 0 )
			strcpy( mode, "w" ); /* WRITE ONLY extension */
			/* (still considered NEW -- temporary) */
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
			unit_s[u].status = UNABLE_TO_CREATE_NEW;
		/* NOTE: any existing file is deleted */
		}
	else if( strcmpip(sta,"SCRATCH") == 0 ){ /* open scratch file */
		unit_s[u].f77typ = SCRATCH;
		strcpy( mode, "w+" );
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
			unit_s[u].status = UNABLE_TO_CREATE_SCRATCH;
		/* NOTE: any existing file is deleted; 
		 *   the main difference betw. NEW and SCRATCH is that 
		 *   SCRATCH files are deleted by default when they are 
		 *   closed
		 */
		}
	else if( *sta == '\0'    /* the default, "" */
		     || strcmpip(sta,"UNKNOWN") == 0 ){
		/* open unknown (existance) file */
		unit_s[u].f77typ = UNKNOWN;
		/* first try to open it as an existing file */
		strcpy( mode, "r+" );
		if( unit_s[u].form == UNFORMATTED )
			strcat( mode, "b" ); /* open in binary mode */
		if( (unit_s[u].fp=fopen(fn,mode)) == NULL ){
			/* then try to open it as a new file */
			strcpy( mode, "w+" );
			if( unit_s[u].form == UNFORMATTED )
				strcat( mode, "b" ); /* open in binary mode */
			if( (unit_s[u].fp=fopen(fn,mode)) == NULL )
				unit_s[u].status = UNABLE_TO_OPEN_UNKNOWN;
			}
		}
	else	/* other status */
		unit_s[u].status = BAD_OPEN_STATUS;

	if( unit_s[u].access==APPEND )
		unit_s[u].access = SEQUENTIAL;
		/*now consider APPEND the same as SEQUENTIAL*/

	if( unit_s[u].status )
		return( _ERR_ );
	return( TRUE ); /* unit successfully opened & initialized! */
}

	/*	*	*	*	*/

void ini_unitstru( u, fn, fm, acc, rl )	/* unit struct initialization */
int u;	/* unit to initialize */
char *fn,	/* the filename */
	*fm,	/* format mode ("" for default) */
	*acc;	/* the access mode ("" for default) */
int rl;	/* the record length (if direct access) */
{

	if( unit_s[u].status )
		return;

	if( unit_s[u].fname != NULL )
		free( unit_s[u].fname );
	if( (unit_s[u].fname=iostrsav(fn)) == NULL ){
		unit_s[u].status = MEMERR_FILENAME;
		return;
		}

	if( *fm != '\0' ){ /* form given, ck it out */
		if( strcmpip(fm,"FORMATTED") == 0 )
			unit_s[u].form = FORMATTED;
		else if( strcmpip(fm,"UNFORMATTED") == 0 )
			unit_s[u].form = UNFORMATTED;
		else{
			unit_s[u].status = BAD_FORM;
			return;
			}
		}
	/* NOTE: when no form is given, the default type is determined 
	 *       according to the access method chosen (below) */

	if( *acc=='\0' || strcmpip(acc,"SEQUENTIAL")==0
	               || strcmpip(acc,"APPEND")==0 ){
	    if( strcmpip(acc,"APPEND")==0 )
			unit_s[u].access = APPEND;
			/* NOTE: APPEND is treated the same as SEQUENTIAL
			 *  after initially positionally at the EOF.
			 */
		else
			unit_s[u].access = SEQUENTIAL;
		if( *fm == '\0' ) /* NO form given, set default */
			unit_s[u].form = FORMATTED; /* sequential default */
		if( IS_INTERNAL(u) )
			unit_s[u].reclen = 0;	/* depends on str len */
			/* NOTE: ini_unitstru() is NOT called for INTERNAL
			 *       units, but the check is made just in case!
			 */
		else if( unit_s[u].form==FORMATTED ){
#ifdef DG_F77
			if( rl > 0 )
				unit_s[u].reclen = rl;
				/*MAXRECL can be used as alias to RECL for seq files*/
			else
#endif
				unit_s[u].reclen = SEQ_BUFSIZ;
			}
		else	/* UNformatted */
			unit_s[u].reclen = 1; /* seq unf buffer isn't used */
		}
	else if( strcmpip(acc,"DIRECT") == 0 ){
		unit_s[u].access = DIRECT;
		if( rl > 0 ){
			unit_s[u].reclen = rl;
			if( *fm == '\0' ) /* NO form given, set default */
				unit_s[u].form = UNFORMATTED; /* direct default */
			if( unit_s[u].form == UNFORMATTED ){
				if( reclen_wordsiz <= 0 )
					reclen_wordsiz = sizeof( float );
				/* record length in words */
				unit_s[u].reclen *= reclen_wordsiz;
				}
			}
		else{
			unit_s[u].status = BAD_RECLEN;
			return;
			}
		}
	else{
		unit_s[u].status = BAD_ACCESS;
		return;
		}

	/* create buffer using unit_s[u].reclen */
	if( unit_s[u].ubuf ){ /*release any existing buffer first*/
		free( unit_s[u].ubuf );	/* rele the unit buffer */
		unit_s[u].ubuf = NULL;
		}
	if(	 IS_INTERNAL(u) )
		;	/* internal units supply their own buffer */
	else if( (unit_s[u].ubuf=(char*)malloc(unit_s[u].reclen+1)) != NULL ){
		unit_s[u].ubufend = unit_s[u].ubuf + unit_s[u].reclen;
		unit_s[u].nxtu = unit_s[u].ubuf;
			/* set the unit buffer to ' ' */
		memset( unit_s[u].ubuf, ' ', unit_s[u].reclen );
		*unit_s[u].ubufend = '\0'; /* null terminate buffer */
		}
	else{
		unit_s[u].status = MEMERR_UNITBUF;
		return;
		}

	if( unit_s[u].ff )
		free( unit_s[u].ff );
	if( unit_s[u].form == FORMATTED ){
		/* make the format struct */
		if( (unit_s[u].ff=(F_FORMAT*)malloc(sizeof(F_FORMAT))) == NULL )
			unit_s[u].status = MEMERR_FORMAT;
		}
	else if( unit_s[u].form == UNFORMATTED )
		unit_s[u].ff = NULL;

	if( !IS_INTERNAL(u) )
		unit_s[u].intu = FALSE; /* note this isn't an internal unit*/
		/* NOTE: the 'intu' field is used when accessing the unit by
		 *       a pointer, thus the unit number isn't known */

	/* set defaults for newly opened unit */
	unit_s[u].action = READWRITE; /*allow both READ & WRITE to unit*/
	unit_s[u].ateof  = NO;
	unit_s[u].blank  = BLANK_NULL; /* default */
	unit_s[u].carrgc = NO; /* ignore carriage control chars in col 1 */
	unit_s[u].lock   = NO; /*allow multiple access*/
#ifdef PAD_SHORT_RECORD
	unit_s[u].pad    = YES; /*pad short records - F90 default*/
#else
	unit_s[u].pad    = NO;
#endif
#ifdef DG_F77
	unit_s[u].errlog = NO; /*log RT error msgs to unit (DG ext.)*/
	unit_s[u].inclnl = NO; /*don't include \n on input*/
#endif
	return;
}

	/*   *   *   *   */

void setu_action( u, action ) /*set the F90 I/O intent*/
int u; /*F77 unit no.*/
char *action; /*"YES" or "NO"*/
{
	if( VALID_UNIT(u)==NO )
		;
	else if( strcmpip(action,"READ")==0 )
		unit_s[u].action = READONLY;
	else if( strcmpip(action,"WRITE")==0 )
		unit_s[u].action = WRITEONLY;
	else if( strcmpip(action,"READWRITE")==0 )
		unit_s[u].action = READWRITE;
#ifdef DG_F77 /*support for DG's IOINTENT=ioi */
	else if( strcmpip(action,"INPUT")==0 )
		unit_s[u].action = READONLY;
	else if( strcmpip(action,"OUTPUT")==0 )
		unit_s[u].action = WRITEONLY;
	else if( strcmpip(action,"BOTH")==0 )
		unit_s[u].action = READWRITE;
#endif
	else
		; /*no action - undefined*/
	return;
}

	/*   *   *   *   */

int setu_blank( u, blnk )	/* set the unit blank handling attribute */
int u;
char *blnk;
{ /*NOTE: this routine was formerly called unit_blank()*/

	if( !inqu_opened(u) || unit_s[u].status )
		return( _ERR_ );

	if( strcmpip(blnk,"NULL") == 0 )
		unit_s[u].blank = BLANK_NULL;
	else if( strcmpip(blnk,"ZERO") == 0 )
		unit_s[u].blank = BLANK_ZERO;
	else{
		unit_s[u].status = BAD_BLANK;
		return( _ERR_ );
		}
	return( TRUE );
}

	/*	*	*	*	*/

void setu_crrgcntl( u, carrg ) /*set the carriage control flag on given unit*/
	/*NOTE: this is the transl. of the CARRIAGECONTROL= extension*/
int u;  /* unit */ 
char *carrg; /*'LIST', 'NONE', 'FORTRAN'*/
{
	/*NOTE: in this implimentation, LIST & NONE have the same impact*/
	if( !VALID_UNIT(u) )
		badunit( u );
	if( strcmpip(carrg,"FORTRAN")==0 )
		unit_s[u].carrgc = 1; /* use col 1 for carriage control */
	else /* disable carriage control */
		unit_s[u].carrgc = 0;
	return;
}

	/*   *   *   *   */

#ifdef DG_F77
	void setu_delimiter( u, delim ) /*set flag to incl/exclude the \n */
	int u; /*F77 unit no.*/              /*(a DG F77 feature)*/
	char *delim; /*"INCLUDE" or "EXCLUDE"*/
	{
		if( VALID_UNIT(u)==NO )
			;
		else if( strcmpip(delim,"INCLUDE")==0 )
			unit_s[u].inclnl = YES;
		else
			unit_s[u].inclnl = NO;
		return;
	}
#endif

	/*   *   *   *   */

#ifdef DG_F77
	void setu_errorlog( u, setting ) /*re-direct the runtime error msgs*/
	int u; /*F77 unit no.*/
	char *setting; /*"YES" or "NO"*/
	{
		if( VALID_UNIT(u)==NO )
			;
		else if( strcmpip(setting,"YES")==0 )
			unit_s[u].errlog = YES;
		else
			unit_s[u].errlog = NO;
		return;
	}
#endif

	/*   *   *   *   */

#ifdef DG_F77
void setu_exclusive( u, exclusive ) /*lock the file*/
int u; /*F77 unit no.*/
char *exclusive; /*"YES" or "NO"*/
{
	if( inqu_opened(u)==NO )
		;
	else if( strcmpip(exclusive,"YES")==0 ){
		/*if not locked, then lock the file associated with the unit*/
		unit_s[u].lock = YES;
		}
	else{ /*unlock the file*/
		/* if( unit_s[u].lock==YES ) then unlock the file */
		unit_s[u].lock = NO;
		}
	return;
}
#endif

	/*   *   *   *   */

void setu_pad( u, pad ) /*re-direct the runtime error msgs*/
int u; /*F77 unit no.*/
char *pad; /*"YES" or "NO"*/
{
	if( VALID_UNIT(u)==NO )
		;
	else if( strcmpip(pad,"YES")==0 )
		unit_s[u].pad = YES; /*default for F90*/
	else
		unit_s[u].pad = NO; /*std for F77*/
	return;
}

	/*  *  *  *  */

	/* * Unit initialization & buffer flushing * */

int crrgcntl = 1; /* when true, open stdOUT w/col 1 used for carriage cntl */

void ini_units() /* initialize the F77 units: STDIN_UNIT & STDOUT_UNIT */
{                /* (called from f77_ini() - & possibly auto_openunit()) */

		/* pre-defined units */

	/* standard input (keyboard) */
	unit_s[STDIN_UNIT].fp = stdin;
	unit_s[STDIN_UNIT].f77typ = OLD;
	unit_s[STDIN_UNIT].status =	0; /* NO errors */
	ini_unitstru( STDIN_UNIT, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	if( unit_s[STDIN_UNIT].status == MEMERR_UNITBUF
	   || unit_s[STDIN_UNIT].status == MEMERR_FORMAT )
		f_err( "allocating space for STDIN_UNIT" );

	/* standard output (screen) */
	unit_s[STDOUT_UNIT].fp = stdout;
	unit_s[STDOUT_UNIT].f77typ = NEW;
	unit_s[STDOUT_UNIT].status =	0; /* NO errors */
	ini_unitstru( STDOUT_UNIT, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	unit_s[STDOUT_UNIT].carrgc = crrgcntl; /* use col 1 for carriage control */
	if( unit_s[STDOUT_UNIT].status == MEMERR_UNITBUF
	   || unit_s[STDOUT_UNIT].status == MEMERR_FORMAT )
		f_err( "allocating space for STDOUT_UNIT" );
#ifdef STD_PRN
	/* standard printer (typically defined on MSDOS) */
	unit_s[STDPRN_UNIT].fp = stdprn;
	unit_s[STDPRN_UNIT].f77typ = NEW;
	unit_s[STDPRN_UNIT].status =	0; /* NO errors */
	ini_unitstru( STDPRN_UNIT, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	unit_s[STDPRN_UNIT].carrgc = crrgcntl; /* use col 1 for carriage control */
	if( unit_s[STDPRN_UNIT].status == MEMERR_UNITBUF
	   || unit_s[STDPRN_UNIT].status == MEMERR_FORMAT )
		f_err( "allocating space for STDPRN_UNIT" );
#endif

	/* internal I/O unit (char string) */
	unit_s[INTERNAL_UNIT].fp = NULL;
	unit_s[INTERNAL_UNIT].f77typ = UNKNOWN;
	unit_s[INTERNAL_UNIT].access = SEQUENTIAL;	/* by definition */
	unit_s[INTERNAL_UNIT].form = FORMATTED;
	unit_s[INTERNAL_UNIT].status =	0;	/* NO errors */
	unit_s[INTERNAL_UNIT].intu = TRUE;	/* denote INTERNAL unit */

	if( unit_s[STDIN_UNIT].status || unit_s[STDOUT_UNIT].status 
	   || unit_s[INTERNAL_UNIT].status )
		f_err( "initializing std input/output/internal units" );

	return;
}

void ini_inoutunits() /* initialize the F77 units: IN_UNIT & OUT_UNIT */
{                /* (called from f77_ini()) */

		/* pre-defined units */
	/* stardard input (keyboard) for most old FORTRAN-IV */
	unit_s[IN_UNIT].fp = stdin;
	unit_s[IN_UNIT].f77typ = OLD;
	unit_s[IN_UNIT].status =	0; /* NO errors */
	ini_unitstru( IN_UNIT, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	if( unit_s[IN_UNIT].status == MEMERR_UNITBUF
	   || unit_s[IN_UNIT].status == MEMERR_FORMAT )
		f_err( "allocating space for IN_UNIT" );

#ifdef IN_UNIT2
	/* stardard input (keyboard) for most old FORTRAN-IV */
	unit_s[IN_UNIT2].fp = stdin;
	unit_s[IN_UNIT2].f77typ = OLD;
	unit_s[IN_UNIT2].status =	0; /* NO errors */
	ini_unitstru( IN_UNIT2, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	if( unit_s[IN_UNIT2].status == MEMERR_UNITBUF
	   || unit_s[IN_UNIT2].status == MEMERR_FORMAT )
		f_err( "allocating space for IN_UNIT #2" );
#endif /*IN_UNIT2*/

	/* standard output (screen) for most old FORTRAN-IV */
	unit_s[OUT_UNIT].fp = stdout;
	unit_s[OUT_UNIT].f77typ = NEW;
	unit_s[OUT_UNIT].status =	0; /* NO errors */
	ini_unitstru( OUT_UNIT, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	unit_s[OUT_UNIT].carrgc = crrgcntl; /* use col 1 for carriage control */
	if( unit_s[OUT_UNIT].status == MEMERR_UNITBUF
	   || unit_s[OUT_UNIT].status == MEMERR_FORMAT )
		f_err( "allocating space for OUT_UNIT" );
#ifdef DG_F77
	unit_s[OUT_UNIT].errlog = YES; /*log RT error msgs to unit (DG ext.)*/
#endif

#ifdef OUT_UNIT2
	/* standard output (screen) for most old FORTRAN-IV */
	unit_s[OUT_UNIT2].fp = stdout;
	unit_s[OUT_UNIT2].f77typ = NEW;
	unit_s[OUT_UNIT2].status =	0; /* NO errors */
	ini_unitstru( OUT_UNIT2, "", "", "", 0 ); /* UN-named unit */
	    /* defaults: SEQUENTIAL, FORMATTED, & BLANK_NULL */
	unit_s[OUT_UNIT2].carrgc = crrgcntl; /* use col 1 for carriage control */
	if( unit_s[OUT_UNIT2].status == MEMERR_UNITBUF
	   || unit_s[OUT_UNIT2].status == MEMERR_FORMAT )
		f_err( "allocating space for OUT_UNIT #2" );
#endif

	if( unit_s[IN_UNIT].status || unit_s[OUT_UNIT].status
#ifdef INOUT2_UNIT
	   || unit_s[IN_UNIT2].status || unit_s[OUT_UNIT2].status
#endif
	   )
		f_err( "initializing pre-opened input/output units" );

	return;
}

	/*	*	*	*	*/

char *iostrsav( s )	/* save str 's' (like K&R's strsave()) */
char *s;	/* (except this generates fatal error if str can't be saved) */
{
	char *p;

	if( (p=malloc( (unsigned)(strlen(s)+1) )) == NULL )
		f_err( "unable to save string" );
	return( strcpy( p, s ) );
}  /* end of io_strsav */

	/*	*	*	*	*/

char *prep_fn( pfn ) /* trim (& maybe lowercase) filename */
char *pfn; /* possibly padded filename */
{
	static char fn[120];
	strlcpy( fn, pfn, 119 );
	trim( fn ); /* trim any leading and trailing blanks */
#if unix || sun || _IBMR2 || __sgi || hpuxpa || __clix__
	/* UNIX is case SENSITIVE */
#else	/* case INsensitive */
	strlwr( fn ); /* lowercase the filename for consistency */
#endif
	return( fn );
}

	/*	*	*	*	*/

void set_crrgcntl( u, set ) /*set the carriage control flag on a given unit*/
int u,   /*unit*/ 
    set; /*if true, set the carriage control on*/
{
	if( !VALID_UNIT(u) )
		badunit( u );
	if( set )
		unit_s[u].carrgc = 1; /* use col 1 for carriage control */
	else /* disable carriage control */
		unit_s[u].carrgc = 0;
	return;
}

	/*	*	*	*	*/

#define ISSPACE(c)	((c)==' ' || (c)=='\t' || (c)=='\n')

char *trim(str) /*trim leading & trailing spaces, retn trimmed str*/
char *str;
{
	char *s, *t,
		*lnb = 0;	/* ptr to last non-blank char */

	/* find 1st non-blank */
	for( s=str; ISSPACE(*s); s++ )
		;

	/*copy rest of str, noting last non-blank char*/
	for( t=str;  (*t= *s) != '\0';  s++, t++ ){
		if( ! ISSPACE(*t) )
			lnb = t; /*save location of non-blank*/
		}

	if( lnb ) /*ck if there was a non-blank char*/
		*(t=lnb+1) = '\0';	/*terminate str after last non-blank*/

	return( str );
}
