/* readfmt.c - support for translated F77 FORMATTED read stmts
 *                   by FOR_C (TM) from COBALT BLUE
 *  copyright Cobalt Blue, Inc., 1988 - 1995
 *             ALL RIGHTS RESERVED
 */

/* NOTE: the unit status flag is only set for things that might be expected
 *           to change at runtime of a particular application,
 *   in contrast, an fmterr()/f_err() call is made when programming related 
 *           errors are detected */

#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>	/* ANSI variable arg macros */
#else
#include <varargs.h>
#endif

#include <fcio.h>
#include <complx.h>

			/* * *  local functions & macros  * * */

/* #define DEBUG */

#define MAXSTR	80

UNITSTRU *i_up;	/* local input unit structure pointer */
FEDIT in; /* input field edit descriptor */

/* macro to assign the data value to an argument, given,
	- type 't' 
	- function call 'fc' to actually read & return the input
 */
#define ASSGN(t,fc) 	{ t *tp; \
						tp = va_arg( args, t* ); \
						for( ; nele > 0; nele-- ) \
							*tp++ = fc; }


#ifdef PROTOTYPES
	int prep_fpin(char[],int,int); /*prepare fp input for atof()*/
	int rdcunit(UNITSTRU*); /*return next char from unit for FORMATTED I/O*/
	void rdfchr(char*,unsigned); /*read formatted char*/
	int rdflog(char*); /*read formatted (f77) logical*/
	long rdfint(int,char*); /*read formatted long int*/
	double rdfdbl(char*); /*read formatted floating pt no.*/
	void rdfnumfld(char*,int,int);	/*get numeric input field*/
	void rdfchrfld(char*,int);	/*get char field*/
	char *inschr( int, char* );  /*insert char 'c' in front of 'str'*/
#else
	int prep_fpin();
	int rdcunit();
	void rdfchr();
	int rdflog();
	long rdfint();
	double rdfdbl();
	void rdfnumfld(), rdfchrfld();
	char *inschr();
#endif
	int _readfmt();

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */

#if defined(__STDC__) || ANSI_C_TARGET
int dreadf( int u, long rn, char *fmt, char *cs, ... )
#else
int dreadf( u, rn, fmt, cs, va_alist )
int u;
long rn; /* the F77 record number */
char *fmt; /* the F77 format str */
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{
	int nf=0; /* number of fields actually read & assigned */
	va_list args; /* arg ptr */

	/* Initialization */
	if( !VALID_UNIT(u) )
		badunit( u ); /* unit error */
	i_up = &unit_s[u]; /* init. the unit struct ptr */
	if( UACTION(u) == WRITEONLY ){
		i_up->status = READ_ON_WRITEONLY;
		return 0;
		}
	rd_dirbeg( u, rn, fmt ); /* init. the unit structure */
	if( i_up->status )
		return( 0 );
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
	va_start(args);	/* initialize variable args */
#endif
	/* end of initialization */

	nf = _readfmt( u, cs, args );
	va_end( args );
	return( nf );	/* return the no. of fields read */
}

	/*		*		*		*		*/

#if defined(__STDC__) || ANSI_C_TARGET
int readf( int u, char *fmt, char *cs, ... )
#else
int readf( u, fmt, cs, va_alist )
int u;
char *fmt; /* the F77 format str */
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{
	int nf=0; /* number of fields actually read & assigned */
	va_list args; /* arg ptr */

	/* Initialization */
	if( !VALID_UNIT(u) && u != STDIN_UNIT )
		badunit( u ); /* unit error */
	i_up = &unit_s[u]; /* init. the unit struct ptr */
	if( UACTION(u) == WRITEONLY ){
		i_up->status = READ_ON_WRITEONLY;
		return 0;
		}
	rd_seqbeg( u, fmt ); /* init. the unit structure */
	if( i_up->status )
		return( 0 );
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
	va_start(args);	/* initialize variable args */
#endif
	/* end of initialization */

	nf = _readfmt( u, cs, args );
	va_end( args );
	return( nf );	/* return the no. of fields read */
}

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */

#if defined(__STDC__) || ANSI_C_TARGET
int readfmt( int u, char *cs, ... )
#else
int readfmt( u, cs, va_alist )
int u;
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{
	int nf; 		/* no. of fields successfully input */
	va_list args;		/* arg ptr */

	if( !VALID_UNIT(u) && u != STDIN_UNIT && u != INTERNAL_UNIT )
		badunit( u );	/* unit error */
	i_up = &unit_s[u];	/* init. the unit stru */
	if( i_up->status )
			;	/* error already exists, preserving original error */
	else if( i_up->fp == NULL && i_up->intu == FALSE )
		i_up->status = NOT_OPEN;
	else if( i_up->form == UNFORMATTED )
		i_up->status = NOT_OPEN_FOR_FORMAT;
	if( i_up->status )
		return( 0 );
	if( i_up->ff->fmt == NULL )	/* format ck */
		f_err( "missing format in readfmt()" );

#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs);	/* initialize variable args */
#else				/* UNIX System V C */
	va_start(args);		/* initialize variable args */
#endif

	nf = _readfmt( u, cs, args );
	va_end( args );
	return( nf );		/* return the no. of fields read */

}

	/*	*	*	*	*/


#if defined(__STDC__) || ANSI_C_TARGET
int _readfmt( int u, char *cs, va_list args )
#else
int _readfmt( u, cs, args )
int u;	/* the unit no. */
char *cs;	/* the control str that identifies the variable args */
va_list args;
#endif
{
char str[MAXSTR];	/* str for list directed numeric input */
char *_c,	/* local ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int nf=0,	/* number of fields actually read & assigned */
	prefix;
size_t nele;	/* the number of elements to read w/a given object */

	/* Initialization */
	i_up = &unit_s[u];	/* init. the unit stru */
	/* end of initialization */

/* read the control string - and the associated INput */
for( _c=cs; *_c ; _c++ ){

	if( i_up->status )
		goto rderr;

	if( isspace(*_c) )	/* ignore whitespace */
		continue;

	else if( *_c == '%' ){	/* syntax:  %[prefix]type  */

		_c++;	/* advance past % */

		if( *_c == 'A' ){	/* ck for array indication prefix */
			_c++;	/* advance past 'A' */
			nele = va_arg( args, size_t );	/* get the no. of elements */
			}
		else
			nele = 1;	/* default */

		if( *_c=='l' || *_c=='L' || *_c=='h' ) /* ck for prefix */
			prefix = *_c++;	/* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( (int) *_c ){

			case 'S':
				{ size_t lensiz;
				/* get char ptr and following char length size */
				p = (char*) va_arg( args, void* ); /*get ptr to obj*/
				lensiz = va_arg( args, size_t ); /*get char length*/
				for( ; nele > 0; nele-- ){
					rdfchr( p, lensiz-1 );
					strpad( p, ' ', lensiz-1 );
					p += lensiz; /* advance the str */
					}
				}
				break;

			case 's':	/* null terminated string INput */
				p = va_arg( args, char* ); /* get target str location */
				rdfchr( p, INT_MAX );
				/* NOTE: prefixes (including 'A') are ignored;
						 '\0' is copied to target */
				break;

			case 'c':	/* character INput to BYTE */
				/* prefix ignored, since BYTE specified */
				ASSGN( char, rdfint(sizeof(char),str) )
				break;

			case 'd':	/* integer INput */
			case 'i':
				if( prefix == 'l' )
#ifdef DEBUG
						{ long *tp;
						tp = va_arg( args, long* );
						for( ; nele > 0; nele-- )
							*tp++ = rdfint(sizeof(long),str);
						}
#else
					ASSGN( long, rdfint(sizeof(long),str) )
#endif
				else if( prefix == 'L' )
					ASSGN( dbl_long, rdfint(sizeof(dbl_long),str) )
				else if( prefix == 'h' )
					ASSGN( short, rdfint(sizeof(short),str) )
				else	/* default int */
					ASSGN( int, rdfint(sizeof(int),str) )
				break;

			case 'f':	/* floating pt INput */
				if( prefix == 'L' )
					ASSGN( longdbl, rdfdbl(str) )
				else if( prefix == 'l' )
#ifdef DEBUG
						{ double *tp;
						tp = va_arg( args, double* );
						for( ; nele > 0; nele-- )
							*tp++ = rdfdbl(str);
						}
#else
					ASSGN( double, rdfdbl(str) )
#endif					
				else
					ASSGN( float, (float) rdfdbl(str) )
				break;

			case 'T':	/* LOGICAL8 object */
				ASSGN( LOGICAL8, rdflog( str ) )
				break;

			case 't':	/* LOGICAL16 or LOGICAL32 object */
				if( prefix == 'l' )
					ASSGN( LOGICAL32, rdflog( str ) )
				else if( prefix == 'L' )
					ASSGN( LOGICAL64, rdflog( str ) )
				else
					ASSGN( LOGICAL16, rdflog( str ) )
				break;

			case 'z':	/* complex INput */
				if( prefix == 'l' ){
				 	COMPLEX16 *dzp;
					dzp = va_arg( args, COMPLEX16* );
					for( ; nele > 0; nele-- ){
						dzp->re = rdfdbl(str);
						dzp->im = rdfdbl(str);
						dzp++;
						}
					}
				else{
				 	COMPLEX *zp;
					zp = va_arg( args, COMPLEX* );
					for( ; nele > 0; nele-- ){
						zp->re = rdfdbl(str);
						zp->im = rdfdbl(str);
						zp++;
						}
					}
				break;

			case '\0':
				goto rderr; /*premature termination of cs following '%'*/

			default: /*any other char following '%' is IGNORED!!!*/
				break;
			}	/* end of switch */
		nf++;	/* indicate a field was correctly read */
		}	/* end of if-% */
	else	/* any other non-white char is ignored!!! */
		;
	}	/* end of cs for-loop */

rderr:	/* error reading the data from the file */

	return( nf );	/* return the no. of fields read */
}

	/*	*	*	*	*/

void rdfchr( str, mx )	/* read formatted char from unit */
char *str;	/* storage for the input string */
unsigned mx;/* max str length (excluding '\0') if not given w/A field */
{
int junk;

if( mx < 1 )
	f_err( "input str length must be >0" );

	io_fmtuntl( i_up, 0, 0 ); /*read until field descr detected*/
	getunitfd( i_up, &in ); /*get field descr*/
	in.w = ( in.w==0 ) ? mx : in.w; /*use str length if none in format*/
	if( in.fd != 'A' )
		fmterr( "expecting 'A' edit field", i_up->ff->cf );
	if( in.w >= mx ){
		for( ; in.w > mx; in.w-- ){ /* skip leftmost 'w-mx' chars */
			junk = rdcunit( i_up );
			if( i_up->status )
				break;
			}
		rdfchrfld( str, mx ); /* get rightmost chars from field */
		}
	else{	/* w < mx, so left justify & follow by 'mx-w' blanks */
		rdfchrfld( str, in.w ); /* get field (& place in left of str) */
		while( *str )
			str++;	/* move to the EOS */
		*str = '\0';
		}

return;
}

	/*	*	*	*	*/

void rdfchrfld( s, w )	/* get char field */
char *s;	/* target location for the numeric input */
int w;	/* field width (>0) */
{

	if( i_up->status )
		return;

	while( w-- > 0  &&  (*s++ = rdcunit(i_up)) ){
		if( i_up->status )
			break;
		}
	*s = '\0';

	return;
}

	/*	*	*	*	*/

int rdflog( str )	/* read formatted logical */
char *str;
{	/* NOTE: a logical const is defined as either 'T' or 'F' optionally
		preceded by spaces or a '.', any chars after T/F are ignored */
	/* logical consts: "T", "F", ".TRUE.", ".FALSE." */
int c;
int l;	/* logical input (to retn) */

	/* get the logical input str */
	io_fmtuntl( i_up, 0, 0 ); /*read until field descr detected*/
	getunitfd( i_up, &in ); /*get field descr*/
	if( in.fd == 'Q' ){	/* VAX extension */
		return( strlen(i_up->nxtu) ); /*no. of chars left in record*/
		}
	else if( in.fd != 'L' )
		fmterr( "expecting 'L' edit field", i_up->ff->cf );
	/* get the input field (less lead blanks) */
	rdfnumfld( str, MAXSTR-1, in.w );
	 /* NOTE: the trailing blank conversion to '0' won't affect the
	  * the logical input since only the 1st (non-'.') char in the
	  * field is used
	  */

	if( (c=str[0]) == '.' )	/* ignore optional leading '.' */
		c = str[1];	
	/* determine the logical input 
		(based only on the 1st char, the rest are ignored) */
	if( c == 'T' || c == 't' )
		l = TRUE;
	else if( c == 'F' || c == 'f' )
		l = FALSE;
	else{
		i_up->status = ILLEGAL_LOGICAL_INPUT;
		l = FALSE;
		}

	return( l );
}

	/*	*	*	*	*/

long rdfint( siz, str )	/* read formatted (long) int from unit */
int siz;
char *str;
{
char *c;
char *cp;	/* char input - used w/char hiding */
short *sp;	/* short input - used w/char hiding */
int *ip;	/* default int input - used w/char hiding */

	io_fmtuntl( i_up, 0, 0 ); /*read until field descr detected*/
	getunitfd( i_up, &in ); /*get field descr*/
	if( in.fd == 'I'
	    || in.fd == 'B' /*binary edit extension (DG)*/
	    || in.fd == 'O' /*VAX F77 octal edit*/
	    || in.fd == 'Z' /*VAX F77 hex edit*/ ){
		/* get the input field (less lead blanks) */
		rdfnumfld( str, MAXSTR-1, in.w );
		}
	else if( in.fd == 'A' ){ /*treat chars as integers (1 char/int)*/
		if( in.w == 1  ||  in.w == 0 )
			return( (long)rdcunit(i_up) ); /*assign char to long int*/
			/*NOTE: w==0 when A edit w/o a precision & char*1 object is
			 *      translated as char literal (aka, byte), hence assume
			 *      length of 1!
			 */
		else if( in.w > 1 && in.w <= 4 ){ /*FORTRAN-IV char hiding*/
			rdfchrfld( str, in.w );
			strpad( str, ' ', 4 );
			/* return the int formed by the str */
			if( siz == sizeof(char) ){
				cp = (char*)str;
				return( (long) *cp );
				}
			else if( siz == sizeof(short) ){
				sp = (short*)str; 
				return( (long) *sp );
				}
			else if( siz == sizeof(int) ){
				ip = (int*)str;
				return( (long) *ip );
				}
			else
				return( *(long*)str );
			}
		else
			fmterr("unsupported 'A' format width with integer input",
			         i_up->ff->cf );
		}
	else if( in.fd == 'Q' ){	/* VAX extension */
		return( strlen(i_up->nxtu) ); /*no. of chars left in record*/
		}
	else
		fmterr("expecting 'I', 'B', 'O', 'Z', 'Q', or 'A' edit field",
		        i_up->ff->cf );

	/* ck out the integer input - given I, O, or Z edits */
	if( *str == '-' || *str == '+' )
		c = str+1;	/* skip the sign */
	else
		c = str;
	for( ; *c; c++ ){
		if( in.fd == 'I' ){
			if( !isdigit(*c) )
				i_up->status = ILLEGAL_INT_INPUT;
			}
		else if( in.fd == 'B' ){
			if( !ISBINARY(*c) )
				i_up->status = ILLEGAL_OCT_INPUT;
			}
		else if( in.fd == 'O' ){
			if( !ISOCTAL(*c) )
				i_up->status = ILLEGAL_OCT_INPUT;
			}
		else if( in.fd == 'Z' ){
			if( !isxdigit(*c) )
				i_up->status = ILLEGAL_HEX_INPUT;
			}
		}

	if( in.fd == 'B' )
		return( strtol(str,(char**)NULL,2) );
	else if( in.fd == 'O' )
		return( strtol(str,(char**)NULL,8) );
	else if( in.fd == 'Z' )
		return( strtol(str,(char**)NULL,16) );
	return( atol(str) );	/* convert to long int */
}

	/*	*	*	*	*/

double rdfdbl( str )	/* read formatted floating pt no. */
char *str;
{
	int sf;	/* positive scale factor */
	double f;

	io_fmtuntl( i_up, 0, 0 ); /*read until field descriptor detected*/
	getunitfd( i_up, &in ); /*get field descr*/
	in.fd = (in.fd=='G') ? 'F' : in.fd; /*'G' field equiv to 'F' on input*/
	if( in.fd == 'A' ){ /*char hiding extension*/
		/*NOTE: only Hollerith hiding in 8 byte dbls is supported,
		 * since floats are promoted to doubles, thus messing up
		 * any hidden chars.
		 */
		rdfchrfld( str, in.w );
#ifdef ZERO_DBL_HOLL
		memset( &f, 0, 8 );
#else /*pad w/' '*/
		strpad( str, ' ', 8 );
#endif
		memcpy( &f, str, (in.w < 8 ? in.w : 8) );
		return f;
		}
	else if( in.fd != 'F' && in.fd != 'E' && in.fd != 'D' )
		fmterr("expecting 'D', 'E', 'F', or 'G' edit field",i_up->ff->cf);
	/* get the input field (less lead blanks) */
	rdfnumfld( str, MAXSTR-2, in.w );
	i_up->status = prep_fpin(str,in.fd,in.d); /* prepare for atof() */

	f = atof(str);
	if( strchr(str,'e') == NULL ){
		/* apply scale factor only when NO exponent is present, 
		 * according to the formula: 
		 *  external_value = internal_value * 10Escale_factor
		 * or,
		 *  internal_value = external_value / 10Escale_factor
		 * where 'external_value' represents the input, hence the scale
		 * factor is applied opposite to that applied on output.
		 */
		for( sf=abs(i_up->ff->scalef); sf > 0 ; sf-- )
			f = ( i_up->ff->scalef > 0 ? f/10 : f*10 );
		}
	return( f );
}

	int prep_fpin( str, fd, d )	/* prepare fp input for atof() */
	char str[];
	int fd,	/* field descriptor */
		d;	/* field decimal precision */
	{	/* NOTE: called by rdf_dbl() & complex list directed input */
	char *s, *si, *se;
	int dec,	/* decimal pt. flag */
		status=0;	/* error status */

	if( str[0] == '\0' ) /*3/31/92*/
		return(status); /* null str - zero requires no prep */

	if( str[0] == '-' || str[0] == '+' )
		si = &str[1]; /* ignore negation for str preparation */
	else
		si = str;
	/* find 1st non-digit (& non-decimal pt) */
	for( s=si, dec=NO; isdigit(*s) || *s == '.'; s++ ){
		if( *s == '.' ){
			if( dec )
				status = ILLEGAL_FP_INPUT;
			dec = YES; /* only 1 decimal pt allowed */
			}
		}
	se = s; /* end of field, or location of possible exponential */
	if( *se == '\0' || *se == ' ' )
		;
	else if( se == si  || (*se && !strchr("eEdD-+",*se)) )
		status = ILLEGAL_FP_INPUT;
		/* either the 1st char isn't a digit OR last isn't an expo */

	if( !status ){
		/* if NO decimal pt, ck if one should be inserted 
			(either 'F' field & no expo, OR not 'F' field) */
		if( dec==NO && ((fd == 'F' && *se == '\0') || (fd != 'F')) ){
			/*3/31/92*/
			if( (se-si) >= d ){
				/*insert '.' d spaces left of end (or expo)*/
				inschr( '.', se-d );
				se++; /* incr. since '.' inserted to the left */
				}
			else{ /* insert 0's, then '.' left of end */
				int i;
				for( i=d-(se-si); i > 0; i-- ){
					inschr( '0', si );
					se++; /* incr. since '0' insert */
					}
				inschr( '.', si ); 
				se++; /* incr. since '.' inserted */
				}
			/* NOTE: THE F77 FEATURE OF INSERTING A DECIMAL POINT WHEN ONE IS
			NOT PRESENT IS VERY DANGEROUS!  IT WOULD BE BEST TO PRODUCE A
			WARNING OR ERROR MESSAGE AT THIS POINT AND CHANGE THE INPUT FILE
			RATHER THAN PROPOGATE THIS OLD FORTRAN FEATURE TO C. */
			}
		if( *se == 'D' || *se == 'd' || *se == 'E' /*stdF77*/
		    || *se=='Q' || *se=='q' )  /*PRIME F77 quad prec*/
			*se = 'e';
		else if( *se == '+' || *se == '-' )
			inschr( 'e', se );
		}

	return( status );
	}

	/*	*	*	*	*/

void rdfnumfld( s, mx, w ) /* get formatted numeric input field */
 /* skips leading, & ignores trailing blanks (or maps to '0' on BLANK cntl) */
char *s;	/* target location for the numeric input */
int mx,	/* string size */
	w; /* field width */
{
	int c,	/* char read in */
		lead=YES; /* lead space flag */

	while( w-- > 0 ){
		c = rdcunit(i_up);
		if( i_up->status )
			break;	/* error condition */
		if( c == ',' )
			break;	/* common extension: terminate input field on ',' */
			/*NOTE: next field starts after ','. (see VAXF77 Sect 12-5)*/
		if( isspace(c) ){
			if( lead ) /*3/31/92*/
				continue;	/* skip leading blanks */
			else if( i_up->ff->cblank == BLANK_NULL )
				continue;	/* skip trailing blanks */
			c = '0';	/* convert trailing blanks to '0' */
			}
		if( mx-- > 0 )	/* don't exceed str size */
			*s++ = c;
		lead = NO;	/* no more leading spaces since non-blanks reached */
		}
	*s = '\0';

	return;
}

	/*	*	*	*	*/

#ifdef rdcunit
#	undef rdcunit
#endif

int rdcunit( us ) /*return next char from unit (us) for FORMATTED I/O */
UNITSTRU *us;
{
	int c;
	if( us->status )
		c = 0; /*don't return any char on prior error*/
	else if( *us->nxtu=='\0' || *us->nxtu=='\n' ){
		/*reached end of record*/
		if( us->pad && us->form==FORMATTED )
			c = SHORT_FILLCHR;
		else if( us->pad && us->form==UNFORMATTED )
			c = '\0';
		else{
			us->status = SHORT_RECORD;
			c = 0;
			}
		}
	else
		c = *us->nxtu++;
	return( c );
}

