/* sprtunf.c - UNformatted Output to buffer! */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1995
 *               ALL RIGHTS RESERVED
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>	/* ANSI variable arg macros */
#else
#include <varargs.h>
#endif

#include <fcio.h>
#include <unfio.h>	/* UNFormatted I/O header */

#ifdef NEED_COMPLEX
#include <complx.h>
#endif

		/*		*		*		*		*/

		/* PRIVATE variables and macros */

	/* macros to get arg of type 't' & place it into the buffer.
		(for args passed by value only, i.e., no ptrs)
	   NOTE: 't' must be a valid C type!
	 */
#define UNFWRT(t)	{ DECLVAL( t )	\
					GETARG( t ) 	\
					CK_NOUT( t ) }

#define DECLVAL(t)	t v, *vp;
#define GETARG(t)	v = va_arg( args, t );
#define CK_NOUT(t)	vp = (t*)nxt; \
					*vp++ = v; \
					nxt = (char*)vp;

/* OUTCHR() merely copies 'c' into the next location in the buffer */
#define OUTCHR( c )	*nxt++ = c

/* NOTE: There isn't a CKBUF() macro in sprtunf.c since the end of the 
	buffer isn't known! */

	/*		*		*		*		*/

static long unfdprt_count; /*no. of chars written w/last fseek_wrt()*/


int fseek_wrt( fp, os, n, buf )	/* seek & write a buffer to a file */
FILE *fp;
long os;	/* abs offset from the BOF */
size_t n;		/* number of chars to write (normally buf size) */
char *buf;
{
	/* NOTE: fseek_wrt() is normally called after sprtunf() to
	 *       output the record to a file at a given location
	 */

	if( os >= 0 ){	/* re-position the file offset */
		if( fseek(fp,os,SEEK_SET) )
			return( -1 ); /*error: unable to re-position*/
		}
	/* else, write to the current offset if negative offset */

	unfdprt_count = fwrite(buf,sizeof(char),n,fp);
	if( unfdprt_count < n )
		return( -2 ); /*error writing to phys. device*/

	memset(buf, '\0', n); /*reset the buffer*/

	return( 0 ); /*successful*/
}

	long len_unfdprt() /*return record length of last fseek_wrt() write*/
	{
		return( unfdprt_count );
	}

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */

		/* sprtunf() -- unformatted OUTput to buffer */

#if defined(__STDC__) || ANSI_C_TARGET
size_t sprtunf( char *bp, char *cs, ... )
#else
size_t sprtunf( bp, cs, va_alist )
char *bp;	/* buffer pointer */
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{	/* returns the number of chars written to the buffer */

char *nxt;	/* next location for unformatted output in buffer */
char *c,	/* local ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int prefix;
va_list args;	/* arg ptr */

	/* Initialization */
if( bp == NULL )	/* ck for buffer existance */
	return( 0 );
nxt = bp;

#if defined(__STDC__) || ANSI_C_TARGET
va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
va_start(args);	/* initialize variable args */
#endif

/* read the control string - and the associated output */
for( c=cs; *c ; c++ ){
	if( *c == ' ' || *c == '\t' )	/* ignore spaces & tabs only */
		continue;
	ALIGN( nxt, bp );
	if( *c == '%' ){	/* syntax:  %[prefix]type  */
		c++;	/* advance past % */

		if( *c == 'l' || *c == 'h' || *c == 'L' )	/* ck for prefix */
			prefix = *c++;	/* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( (int) *c ){

			case '&':	/* obj being output, get size & address */
			case 'A':
				{ size_t sz;	/* size of ptr objects being output */
				sz = va_arg( args, size_t );	/* get the size in bytes */
				p = (char*) va_arg( args, void* );	/* get ptr to the obj */
				for( ; sz > 0; sz-- ){
					OUTCHR( *p++ );
					}
				/* NOTE: prefix is ignored */
				}
				break;

			case 's':	/* null terminated string output */
				p = va_arg( args, char* );	/* get str */
				for( ; *p; p++ ){
					OUTCHR( *p );
					}
				OUTCHR( '\0' );	/* incl the string termination */
				/* NOTE: prefix is ignored */
				break;

			case 'c':	/* char output */
				UNFWRT( char )
				/* NOTE: prefix is ignored */
				break;

			case 'd':	/* integer output */
			case 'i':
				if( prefix == 'l' )
					UNFWRT( long )
				else if( prefix == 'L' )
					UNFWRT( dbl_long )
				else if( prefix == 'h' ){
					/* short (passed as int - output as short) */
					DECLVAL( short )
					GETARG( int )
					CK_NOUT( short )
					}
				else
					UNFWRT( int )
				break;

			case 'f':	/* floating pt output */
				if( prefix == 'L' )
#ifdef LONG_DBL
					UNFWRT( longdbl )	/* long double */
#else
					{ puts( "sprtunf() compiled withOUT long double support" );
					exit( 1 ); }
#endif
				else if( prefix == 'l' )
					UNFWRT( double )
				else{	/* float (passed as double - output as float) */
					DECLVAL( float )
					GETARG( double )
					CK_NOUT( float )
					}
				break;

			case 'z':	/* complex output */
#ifdef NEED_COMPLEX
				if( prefix == 'l' )
					UNFWRT( COMPLEX16 )
				else
					UNFWRT( COMPLEX )
#else
				puts( "sprtunf() compiled withOUT complex support" );
				exit( 1 );
#endif
				break;

			case '\0':
				goto wrterr;	/* premature termination of cs following '%' */

			default:	/* any other char following '%' is output literally */
				OUTCHR( *c );
				break;
			}	/* end of switch */
		}	/* end of if-% */
	else{	/* any other char is output literally */
		OUTCHR( *c );
		}
	}

wrterr:	/* error actually writing the data to the buffer */

	ALIGN( nxt, bp );
	return( (size_t)(nxt - bp) );	/* write error */
}
