/* sscanunf.c - UNformatted Input with headers and trailers! */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1995
 *               ALL RIGHTS RESERVED
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

/* 
	#define DEBUG
 */

#include <fcrt.h>

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>	/* ANSI variable arg macros */
#else
#include <varargs.h>
#endif

#include <fcio.h>
#include <unfio.h>	/* UNFormatted I/O header */

#ifdef NEED_COMPLEX
#include <complx.h>
#endif

	/*		*		*		*		*/

		/* PRIVATE variables and macros */

	/* macros to get arg of type 't' from the buffer & assign it to the 
		location pointed to.
	   NOTE: 't' must be a valid C type!
	   		UNFRD() must be enclosed in {}! ! !
	 */
#define UNFRD(t)	DECLTARGT( t )	\
					GETARG( t ) 	\
					CK_NIN( t )

#define DECLTARGT(t) t *tp;
#define GETARG(t)	tp = va_arg( args, t* );
#define CK_NIN(t)	*tp = *(t*)nxt; \
					nxt += sizeof(t);


	/*		*		*		*		*/

static long unfdrd_recsiz; /*record size assoc. w/last sscanunf() call*/
    /*(actually read is read by fseek_rd())*/
	/*NOTE: this quantity is returned by len_unfdrd()*/

int fseek_rd( fp, os, n, buf )	/* seek & read chars into a buffer */
FILE *fp;
long os;	/* abs offset from the BOF */
size_t n;	/* max. number of chars to read (normally buf size) */
char *buf;
{
	/* NOTE: fseek_rd() is normally used as a precessor to sscanunf()
	 *       for reading unformatted records of a fixed size
	 */

	if( os >= 0 ){	/* re-position the file offset */
		if( fseek(fp,os,SEEK_SET) )
			return( -1 );	/* error: unable to re-position */
		}
	/* else, read from the current offset if negative offset */

	memset( buf, '\0', n );	/* reset the buffer */

	unfdrd_recsiz = fread(buf,sizeof(char),n,fp);
	if( unfdrd_recsiz < n )
		return( -2 );	/* error reading from phys. device */

	return( 0 );	/* successful */
}

	long len_unfdscan() /*return record length of last fseek_rd() read*/
	{
		return( unfdrd_recsiz );
	}

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGUMENTS ANSI FUNCTION */

		/* sscanunf() -- unformatted INput from buffer */

#if defined(__STDC__) || ANSI_C_TARGET
size_t sscanunf( char *bp, char *cs, ... )
#else
size_t sscanunf( bp, cs, va_alist )
char *bp;	/* buffer ptr */
char *cs;	/* control string (indicates type & no. of args) */
va_dcl
#endif
{	/* returns the number of chars read from the buffer */

char *nxt;	/* next location for unformatted input in buffer */
char *c,	/* working ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int prefix;
va_list args;	/* arg ptr */

	/* Initialization */
if( bp == NULL )	/* ck for buffer existance */
	return( 0 );

#if defined(__STDC__) || ANSI_C_TARGET
va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
va_start(args);	/* initialize variable args */
#endif

nxt = bp;
for( ; *cs == ' ' || *cs == '\t'; cs++ )	/* skip leading blanks/tabs */
	;
	/* End of Initialization */

/* read the control string - and the associated INput */
for( c=cs; *c ; c++ ){
	if( *c == ' ' || *c == '\t' )	/* ignore spaces & tabs only */
		continue;
	ALIGN( nxt, bp );
	if( *c == '%' ){	/* syntax:  %[prefix]type  */
		c++;	/* advance past % */

		if( *c == 'l' || *c == 'h' || *c == 'L' )	/* ck for prefix */
			prefix = *c++;	/* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( (int) *c ){

			case '&':	/* obj being INput, get size & address */
			case 'A':
				{ size_t sz;
				sz = va_arg( args, size_t );	/* get the size in bytes */
				p = (char*) va_arg( args, void* );	/* get ptr to the obj */
				for( ; sz > 0; sz-- ){
					*p++ = *nxt++;
					}
				}
				/* NOTE: prefix is ignored */
				break;

			case 's':	/* null terminated string INput */
				p = va_arg( args, char* );	/* get target str location */
				do {
					} while( (*p++ = *nxt++) != '\0' );
				/* NOTE: prefix is ignored; '\0' is copied to target */
				break;

			case 'c':	/* char INput */
				{ UNFRD( char ) }
				/* NOTE: prefix is ignored */
				break;

			case 'd':	/* integer INput */
			case 'i':
				if( prefix == 'l' ){
					UNFRD( long )
					}
				else if( prefix == 'L' ){
					UNFRD( dbl_long )
					}
				else if( prefix == 'h' ){
					UNFRD( short )
					}
				else{	/* default int */
					UNFRD( int )
					}
				break;

			case 'f':	/* floating pt INput */
				if( prefix == 'L' ){
#ifdef LONG_DBL
					UNFRD( longdbl )	/* long double */
#else
					puts( "sscanunf() compiled withOUT long double support" );
					exit( 1 );
#endif
					}
				else if( prefix == 'l' ){
					UNFRD( double )
					}
				else{
#ifdef DEBUG
					/* expanded macro logic for debugging */
					float *tp;
					tp = va_arg( args, float* );
				/*	if( (nxt+sizeof(float)-1) >= end )
						goto rderr;    DISABLED SINCE END ISN'T KNOWN!
				 */
					*tp = *(float*)nxt;
					nxt += sizeof(float);
#else
					UNFRD( float )
#endif
					}
				break;

			case 'z':	/* complex INput */
#ifdef NEED_COMPLEX
				if( prefix == 'l' ){
					UNFRD( COMPLEX16 )
					}
				else{
					UNFRD( COMPLEX )
					}
#else
				puts( "sscanunf() compiled withOUT complex support" );
				exit( 1 );
#endif
				break;

			case '\0':
				goto rderr;	/* premature termination of cs following '%' */

			default:	/* any other char following '%' must be matched */
				if( *c != *nxt++ )
					goto rderr;	/* input char doesn't match control str */
				break;
			}	/* end of switch */
		}	/* end of if-% */
	else{	/* any other char must be matched exactly */
		if( *c != *nxt++ )
			goto rderr;	/* input char doesn't match control str */
		}
	}

rderr:	/* error reading the data from the file (or in the control str) */

	ALIGN( nxt, bp );
	return( (size_t)(nxt - bp) );	/* return the no. of chars read */
}
