/* strmgt.c - str mngt. for transl. F77 char exprs by FOR_C (TM)
 */
/*
 *    copyright Cobalt Blue, Inc., 1988 - 1995
 *               ALL RIGHTS RESERVED
 */

 /*	contains:	f_concat()	-- F77 str concat operation
 				f_strcmp()	-- F77 str comparisons
				f_strncpy()	-- F77 str assignment transl.
				f_subs()	-- extracts F77 substr
				f_subscpy()	-- assigns F77 substrs
				ini_chrtmp()-- initializes the char temps
				rel_chrtmp()-- releases the space assoc. w/chrtmps
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <fcrt.h>

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>	/* ANSI variable arg macros */
#else
#include <varargs.h>
#endif

#define EXCESS	50
	/* NOTE: reduce EXCESS to reallocate more often and minimize
		unused large blocks of memory, more runtime (& possibly
		more memory fragmentation) result from smaller EXCESSes */

#ifdef PROTOTYPES
	void ct_alloc(CHRTMP*,long);	/* allocate space for char temporary */
#else
	void ct_alloc();
#endif

char null_ct_[] = "NULL CHRTMP pointer in f_concat() or f_subs()";

	/*		*		*		*		*/

void ct_alloc( tmp, len )	/* allocate space for char temporary */
CHRTMP *tmp;
long len;	/* the new tmp string length */
{
	/* NOTE: the chrtmp array is initialized to 0 size & NULL ptrs */

	if( len > 65534 )
		f_err( "temp string requires too much space" );

	if( tmp->s != NULL )
		free( tmp->s );	/* free the current temp space */

	/* allocate new space for the temp */
	if( (tmp->s=malloc((unsigned)len+1)) == NULL )
		f_err( "unable to allocate space for str temp" );
	tmp->siz = (unsigned) len;
	return;
}

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGRUMENTS ANSI FUNCTION */

/* concat strs, retn the addr of tmp space */
#if defined(__STDC__) || ANSI_C_TARGET
char *vcpyncat( char *targ, int tlen, ... )
#else
char *vcpyncat( targ, tlen, va_alist )
char *targ; int tlen; /*target string & length*/
va_dcl
#endif
{
	char *str;
	va_list args;	/* arg ptr */

	if( targ == NULL )
		f_err( "null target string passed to vcpyncat()" );

	/* copy & concatenate the strings to targ */
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,tlen); /* initialize variable args */
#else /* UNIX System V C */
	va_start(args); /* initialize variable args */
#endif

	*targ = '\0'; /*reset, so that the 1st concat will actually be a copy*/
	for( str=va_arg(args,char*); str!=NULL; str=va_arg(args,char*) ){
		targ = strncat( targ, str, tlen );
		if( strlen(targ) >= tlen )
			break;
		}
	va_end(args); /*reset variable args*/

	strpad( targ, ' ', tlen ); /*pad w/spaces to conform to F77 logic*/
	return( targ );
}

	/*		*		*		*		*/

		/* VARIABLE NUMBER OF ARGRUMENTS ANSI FUNCTION */

/* concat strs, retn the addr of tmp space */
#if defined(__STDC__) || ANSI_C_TARGET
char *f_concat( CHRTMP *tmp, ... )
#else
char *f_concat( tmp, va_alist )
CHRTMP *tmp;
va_dcl
#endif
{
	char *str, *t;
	long len=0L;	/* the total length of the tmp */
	va_list args;	/* arg ptr */

	if( tmp == NULL )
		f_err( null_ct_ );

	/* 1st - determine the necessary size of the temp */
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,tmp);	/* initialize variable args */
#else	/* UNIX System V C */
	va_start(args);	/* initialize variable args */
#endif

	for( str=va_arg(args,char*); str!=NULL; str=va_arg(args,char*) )
		len += strlen(str);
	va_end(args);	/* reset variable args */
	if( len == 0 ){
		if( tmp->s == NULL )
			ct_alloc( tmp, len );
		*(tmp->s) = '\0';	/* nothing is being concatenated */
		return( tmp->s );	/* (shouldn't occur in F77 char ops) */
		}

	/* 2nd - free the current temp space */
	if( tmp->siz < len || (tmp->siz - len) > EXCESS )
		ct_alloc( tmp, len );

	/* 3rd - concatenate the strings to temp */
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,tmp);	/* initialize variable args */
#else	/* UNIX System V C */
	va_start(args);	/* initialize variable args */
#endif

	*(t=tmp->s) = '\0';	/* initialize the local ptr */
	for( str=va_arg(args,char*); str!=NULL; str=va_arg(args,char*) )
		t = strcate( t, str );
		/* NOTE: strcate() retns the new end of 't', rather than
			the beginning, for faster repeated concatenations */

	return( tmp->s );
}

	/*		*		*		*		*/

int f_strcmp( left, right )	/* F77 str comparison */
char *left, *right;	/* retns <0 if l < r, 0 if equal, or >0 if l > r */
{
 /* f_strcmp() effectively blank fills the shorter str for the comparison,
		but otherwise functions the same as strcmp() */
	int l, r;

	if( left==NULL || right==NULL )
		return( left ? 1 : (right? -1: 0) );

	while( *left || *right ){
		l = *left ? *left++ : ' ';
		r = *right ? *right++ : ' ';
		if( l != r )
			return( l - r );
			/* NOTE: This function assumes an ASCII character set!
				If not, then the character set should be designed
				so that the values of the characters correspond to
				the position of the letters in the alphabet, (i.e.,
				'a' should be < 'b'), otherwise the function will 
				need to be re-written as appropriate. */
		}

	return( 0 );	/* equal strings */
}

	/*		*		*		*		*/

char *f_strncpy( to, from, n )	/* F77 str assignment transl. */
char *to, *from;
int n;	/* no. of chars to copy, (blank filled if length(from) < n) */
{
	char *t=to;

	if( to==NULL || from==NULL )
		return( NULL );	/* no chg or action if NULL strs given */
	if( n <= 0 )
		return( NULL );	/* undefined */

	while( n>0 && *from ){	/* copy the string (subject to n) */
		*t++ = *from++;
		n--;
		}
	while( n > 0 ){	/* blank fill (since length(from) < n) */
		*t++ = ' ';
		n--;
		}

	*t = '\0';
	return( to );
}

	/*		*		*		*		*/

	/* NOTE: f_subs() is no longer generated, starting w/FOR_C v3.1 */

char *f_subs(tmp,s,start,end)	/* copy substr to tmp space, retn tmp addr */
CHRTMP *tmp;
char *s;	/* the string containing the substring */
int start, end; /* substring bounds ASSUME C range, (ie., start at 0) */
{
	char *b, *t;
	long len=0L;	/* substring length, (& hence tmp length) */

	/* input checks */
	if( tmp == NULL )
		f_err( null_ct_ );
	if( s == NULL )
		return( NULL );	/* NO chg or action if null substr given */

	/* 1st - determine the substring length (& tmp size) */
	start = MAX(start,0);
	len = MAX(end-start+1,0);

	/* 2nd - free the current temp space if necessary */
	if( tmp->siz < len || (tmp->siz - len) > EXCESS )
		ct_alloc( tmp, len );

	/* 3rd - copy substring to temp space */
	for( b=s+start,t=tmp->s; len ; len-- ){
		if( *b )
			*t++ = *b++;
		else
			*t++ = ' ';	/* blank fill rest of substr */
		}
	*t = '\0';

	return( tmp->s );
}

	/*		*		*		*		*/

char *f_subscpy(to,start,end,len,from)	/* F77 substr assignment transl. */
char *to, *from;
int start, end, /* substring bounds target string */
	len;	/* target string length */
{
	/* NOTE: f_subscpy() null terminates the target string, the end of
			the target string is determined from "to+len"! ! !
			The substring bounds assume C range (i.e., start at 0).
	 */
	char *t;
	int l;
	long n;

	/* input checks */
	if( to == NULL || from == NULL )
		return( NULL );	/* NO chg or action if null str(s) given */

	/* 1st - determine the substr length/size, ltd to target str length */
	start = MAX(start,0);
	if( end >= 0 )
		end = MIN(end,len-1);
	else	/* no end given */
		end = len-1;
	n = MAX(end - start + 1,0);	/* substr length */

	/* insure string reaches the substring (if var. length strs used) */
	for( l=strlen(to); l < start; l++ )
		to[l] = ' '; /*blank pad (to eliminate premature termination)*/

	/* 2nd - copy substring */
	t = to + start;
	for( ; n>0 ; n-- ){	/* copy the string (subject to n) */
		if( *from )
			*t++ = *from++;
		else
			*t++ = ' ';	/* blank fill rest of substr */
		}

	*(to+len) = '\0';	/* insure the string itself is terminated */
	return( to );
}

	/*		*		*		*		*/

void ini_chrtmp(ct,n)	/* initialize the chrtmp array */
CHRTMP ct[];
int n;
{
	int i;

	for( i=0; i < n; i++ ){
		ct[i].siz = -1;
		ct[i].s = NULL;
		}
	return;
}

	/*		*		*		*		*/

void nulltermsa( sa, siz, nstrs )	/* null term. an array of strings */
char *sa;	/* ptr to the string array */
int siz,	/* size of the strings in the array */
	nstrs;	/* no. of strings in the array */
{
	int n;

	for( n=1; n <= nstrs; n++ )
		*(sa+n*siz-1) = '\0';
	return;
}

	/*		*		*		*		*/

void rel_chrtmp( ct, n )	/* release the space assoc. w/chrtmp array */
CHRTMP ct[];
int n;
{
	int i;

	for( i=0; i < n; i++ ){
		if( ct[i].s != NULL )
			free( ct[i].s );
		ct[i].s = NULL;
		ct[i].siz = -1;
		}

	return;
}

	/*		*		*		*		*/

char *strcate( to, from ) /* add 'from' to 'to' & retn the new END of 'to' */
char *to, *from;
{
	while( *to )	/* find end of 'to' */
		to++;
	while( (*to++ = *from++) )	/* add 'from' , (EOS is copied) */
		;
	return( --to );	/* retn ptr to the EOS */
}

