/* wrtfmt.c - support for literally translated F77 FORMATTED write stmts
 *                          by FOR_C (TM) from COBALT BLUE
 *   copyright Cobalt Blue, Inc., 1988 - 1995
 *             ALL RIGHTS RESERVED
 */

/* NOTE: the unit status flag is only set for things that might be expected
 *          to change at runtime of a particular application,
 *  in contrast, a fmterr()/f_err() call is made when programming related 
 *          errors are detected */

#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

	/* get definition of size_t */
#ifdef __HIGHC__
#include <stdefs.h>
#else
#include <stddef.h>
	/* NOTE: Some compilers typedef 'size_t' in stdlib.h &/or string.h too */
#endif

#include <fcrt.h>
#include <fcio.h>

#if defined(__STDC__) || ANSI_C_TARGET
#include <stdarg.h>	/* ANSI variable arg macros */
#else
#include <varargs.h>
#endif

			/* * *  local functions & macros  * * */

#define MAX_LOGLEN	 7	/* 7 chars in max input: ".FALSE." */
#define MAX_LONGLEN 17  /* big enough to hold 32 bit long in octal */
#define MAX_DBLLEN	21	/*20 chars mx, (14 digits prec, +2 sgn & dec, +5 expo)*/
#define MAX_EXPLEN	7 /* allows for width of 7 in exponent field */

#define OVERFLOW_CHAR	'*' /*std F77*/

#ifdef PROTOTYPES
	void fpedit(double,FEDIT*,int,int,char*); /*F77 fp edit*/
	void fp_busedit( double, char*, char* );
	void int_busedit( long, char*, char* );
	void intedit(long,int,int,int,int,char*); /*format F77 Im.n edit*/
	void _wrtbyte(int); /*write formatted byte to unit*/
	void _wrtstr(char*); /*write formatted char str to unit*/
	void _wrtlog(int); /*write formatted logical to unit*/
	void _wrtlong(long,int,int); /*write formatted (long) int to unit*/
	void _wrtdbl(double); /*write formatted floating pt no. to unit*/
#else
	void fpedit();
	void fp_busedit();
	void int_busedit();
	void intedit();
	void _wrtbyte();
	void _wrtstr();
	void _wrtlog();
	void _wrtlong();
	void _wrtdbl();
#endif
int _wrtfmt();

char edigit[] = "0123456789abcdef"; /* support for octal,decimal, & hex */

UNITSTRU *o_up;	/* local output unit structure pointer */

FEDIT out; /* output field edit */

/* macro to write out the argument value, given,
 *  - type 't' 
 *  - function call 'fc' to actually do to writes
 */
#define WRTFMT(t,funcall) \
	{ t v, *vp; \
	if( nele ){ vp = va_arg( args, t* ); \
		 for( ; nele > 0; nele-- ) funcall( *vp++ ); } \
	else { v = va_arg( args, t ); funcall( v ); } }
/*special short int/logical version*/
#define WRTFMT_SHORT(funcall) \
	{ short v, *vp; \
	if( nele ){ vp = va_arg( args, short* ); \
		 for( ; nele > 0; nele-- ) funcall( *vp++ ); } \
	else { v = va_arg( args, int ); funcall( v ); } }

#define wrtlog(v)	_wrtlog((int)v)
#define wrtlong(v)	_wrtlong((long)v,sizeof(long),1)
#define wrtint(v)	_wrtlong((long)v,sizeof(int),1)
#define wrtshort(v)	_wrtlong((long)v,sizeof(short),1)

	/*	*	*	*	*/


#if defined(__STDC__) || ANSI_C_TARGET
int dwritef( int u, long rn, char *fmt, char *cs, ... )
#else
int dwritef( u, rn, fmt, cs, va_alist )
int u;     /* unit no. */
long rn;   /* record number */
char *fmt; /* F77 format str */
char *cs;  /* control str that identifies the variable args */
va_dcl
#endif
{
	int nf=0; /* no. of fields successfully output */
	va_list args; /* arg ptr */

	/* Initialization */
	if( !VALID_UNIT(u) )
		badunit( u );	/* unit error */
	o_up = &unit_s[u]; /* init. the unit struct ptr */
	if( UACTION(u) == READONLY ){
		o_up->status = WRITE_ON_READONLY;
		return 0;
		}
	wrt_dirbeg( u, rn, fmt ); /* init. the unit struct */
	if( o_up->status )
		return( 0 );
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
	va_start(args);	/* initialize variable args */
#endif
	/* end of initialization */

	nf = _wrtfmt( u, cs, args );
	va_end( args );
	return( nf ); /* return the no. of fields written */
}


	/*	*	*	*	*/


#if defined(__STDC__) || ANSI_C_TARGET
int writef( int u, char *fmt, char *cs, ... )
#else
int writef( u, fmt, cs, va_alist )
int u;     /* unit no. */
char *fmt; /* F77 format str */
char *cs;  /* control str that identifies the variable args */
va_dcl
#endif
{
	int nf=0; /* no. of fields successfully output */
	va_list args; /* arg ptr */

	/* Initialization */
	if( !VALID_UNIT(u) && u != STDOUT_UNIT
#ifdef STD_PRN
	     && u != STDPRN_UNIT 
#endif
	   )
		badunit( u );	/* unit error */
	o_up = &unit_s[u]; /* init. the unit struct ptr */
	if( UACTION(u) == READONLY ){
		o_up->status = WRITE_ON_READONLY;
		return 0;
		}
	wrt_seqbeg( u, fmt ); /* init. the unit struct */
	if( o_up->status )
		return( 0 );
#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs);	/* initialize variable args */
#else	/* UNIX System V C */
	va_start(args);	/* initialize variable args */
#endif
	/* end of initialization */

	nf = _wrtfmt( u, cs, args );
	va_end( args );
	return( nf ); /* return the no. of fields written */
}


	/*	*	*	*	*/


#if defined(__STDC__) || ANSI_C_TARGET
int wrtfmt( int u, char *cs, ... )
#else
int wrtfmt( u, cs, va_alist )
int u;    /* unit number */
char *cs; /* control str identifying variable args */
va_dcl
#endif
{
	int nf;   /* no. of fields successfully output */
	va_list args;  /* arg ptr */

	/* Initialization */
	if( !VALID_UNIT(u) && u != STDOUT_UNIT 
#ifdef STD_PRN
	     && u != STDPRN_UNIT 
#endif
	     && u != INTERNAL_UNIT )
		badunit( u ); /* unit error */
	o_up = &unit_s[u]; /* init. the unit struct ptr */
	if( o_up->status )
		; /* error already exists, preserving original error */
	else if( o_up->fp == NULL && o_up->intu == FALSE )
		o_up->status = NOT_OPEN;
	else if( o_up->form == UNFORMATTED )
		o_up->status = NOT_OPEN_FOR_FORMAT;
	if( o_up->status )
		return( 0 );
	if( o_up->ff->fmt == NULL )	/* format ck */
		f_err( "missing format in wrtfmt()" );
	/* end of initialization */

#if defined(__STDC__) || ANSI_C_TARGET
	va_start(args,cs);	/* initialize variable args */
#else				/* UNIX System V C */
	va_start(args);		/* initialize variable args */
#endif

	nf = _wrtfmt( u, cs, args );
	va_end( args );
	return( nf );  /* return the no. of fields written */

}

	/*	*	*	*	*/


#if defined(__STDC__) || ANSI_C_TARGET
int _wrtfmt( int u, char *cs, va_list args )
#else
int _wrtfmt( u, cs, args )
int u;    /* unit no. */
char *cs; /* control str that identifies the variable args */
va_list args;
#endif
{
char *c,	/* local ptr to the control str */
	*p;		/* ptr to objs (& strs) being output */
int nf=0, 	/* no. of fields successfully output */
	prefix;
size_t nele; /* no. of elements - if array passed */

o_up = &unit_s[u];	/* init. the unit stru */

/* read the control string - and the associated OUTput */
for( c=cs; *c ; c++ ){
	out.fd = out.w = out.d = out.x = 0;
	if( o_up->status )
		goto wrterr;

	if( *c==' ' || *c=='\t' )	/* ignore whitespace */
		continue;

	else if( *c=='\n' ){	/* end of the write stmt reached */
		if( o_up->setbuf )
			setubuf( o_up ); /* init. buffer to ' 's */
		io_fmtuntl( o_up, YES, 1 );	/* finish off the fmt */
		term_rec( o_up, YES ); /* flush unit buffer */
		break;
		}

	else if( *c=='%' ){	/* syntax:  %[prefix]type  */

		if( o_up->setbuf )
			setubuf( o_up ); /* init. buffer to ' 's */

		c++;	/* advance past % */

		if( *c=='A' ){ /* expect address of array */
			c++;
			nele = va_arg( args, size_t );
			}
		else
			nele = 0; /* NO array passed; arg is pbv */

		if( *c=='l' || *c=='L' || *c=='h' ) /* ck for prefix */
			prefix = *c++; /* save prefix & advance ptr */
		else
			prefix = '\0';

		switch( (int) *c ){

			case 'S':
			case 's':	/* null terminated string OUTput */
				io_fmtuntl(o_up,NO,1); /* write until field descr */
				getunitfd( o_up, &out ); /*get field descr*/
				if( nele )
				  f_err( "array of strings isn't supported" );
				/*NOTE: can't accept array of strs, 
				 * unless the string length is known (e.g, passed)
				 */
				p = va_arg( args, char* );	/* get str */
				_wrtstr( p );
				break;

			case 'c':	/* sgl char & logical OUTput */
				{ char *chp; int ch; /*WRTFMT expanded*/
				if( nele ){ 
					chp = va_arg( args, char* );
					for( ; nele > 0; nele-- )
					  _wrtbyte( *chp++ );
					}
				else
					{ ch = va_arg( args, int ); 
					_wrtbyte( ch ); }
				}
				break;

			case 't':	/* logical OUTput */
				if( prefix=='L' )
					WRTFMT( LOGICAL64, wrtlog )
				if( prefix=='l' )
					WRTFMT( LOGICAL32, wrtlog )
				else if( prefix == 'h' )
					WRTFMT_SHORT( wrtlog )
				else	/* default */
					WRTFMT( LOGICAL32, wrtlog )
				break;

			case 'T':
				WRTFMT( LOGICAL8, wrtlog )
				break;

			case 'd':	/* integer OUTput */
			case 'i':
				if( prefix=='l' || prefix=='L' )
					WRTFMT( long, wrtlong )
				else if( prefix == 'h' )
					WRTFMT_SHORT( wrtshort )
				else	/* default int */
					WRTFMT( int, wrtint )
				break;

			case 'f':	/* floating pt OUTput */
			case 'g':
				if( prefix == 'L' )
					WRTFMT( longdbl, _wrtdbl )
				else if( prefix=='l' || nele==0 ) /*double*/
					WRTFMT( double, _wrtdbl )
				else{ /*float array case*/
					float *vp;
					vp = va_arg( args, float* );
					for( ; nele > 0; nele-- )
					  _wrtdbl( *vp++ );
					}
				break;

			case '\0':
				goto wrterr;	/* premature termination of cs following '%' */

			default:	/* ignore any other char following '%' */
				break;
			}	/* end of switch */
		nf++;	/* indicate a field was correctly written */
		}	/* end of if-% */
	else
		;	/* any other non-white char is ignored! */

	if( o_up->status )
		break;	/* terminate loop on error */
	}	/* end of cs for-loop */

wrterr:	/* error writing the data to the file */

	return( nf );	/* return the no. of fields written */
}

	/*	*	*	*	*/

void _wrtbyte( ch )
int ch;
{
	io_fmtuntl(o_up,NO,1); /* write until field descr */
	getunitfd( o_up, &out ); /*get field descr*/

	if( out.fd=='I' || out.fd=='B' || out.fd=='O' || out.fd=='Z' ){
		/* byte output of char */
		_wrtlong( (long)ch, sizeof(byte), 0 );
		return;
		}

	while( (out.w--) > 1 )	/* pad w/' ' */
	 	WRTCH( ' ', o_up )
	if( out.fd == 'A' || out.fd == 'L' )
		WRTCH( ch, o_up )
	else if( out.fd != 'Q' )
		f_warn("unknown field descriptor w/char output");
	/* NOTE: the VAX 'Q' edit (used w/int & logical types), 
	 *  is ignored on output, causing the output to be skipped
	 */
	/* NOTE: the field descriptor type is ignored, 
	 *  since %c may be used for both A and L edits
	 */
	return;
}

	/*	*	*	*	*/

void _wrtstr( s )	/* write formatted char str to unit */
char *s;	/* char str to output */
{
int len;	/* str length */

	if( out.fd != 'A' )
		fmterr( "expecting 'A' edit field", o_up->ff->cf );
	len = strlen( s );
	if( out.w && out.w > len ){
		while( (out.w--) > len )	/* pad w/' ' */
			WRTCH( ' ', o_up )
		/* output the char str */
		while( *s && !o_up->status )
			WRTCH( *s++, o_up )
		}
	else if( out.w && out.w <= len ){
		while( out.w-- )	/* output the leftmost w chars */
			WRTCH( *s++, o_up )
		}
	else{	/* no width given, simply output the string */
		while( *s && !o_up->status )
			WRTCH( *s++, o_up )
		}

return;
}

	/*	*	*	*	*/

void _wrtlog( ch )
int ch;
{
	io_fmtuntl(o_up,NO,1); /* write until field descr */
	getunitfd( o_up, &out ); /*get field descr*/
	while( (out.w--) > 1 )	/* pad w/' ' */
	 	WRTCH( ' ', o_up )
	if( out.fd == 'L' )
		WRTCH( (ch ? 'T':'F'), o_up )
	else if( out.fd != 'Q' )
		f_warn("unknown field descriptor w/logical output");
	return;
}

	/*	*	*	*	*/

void _wrtlong( l, siz, procfmt ) /* write formatted (long) int to unit */
long l;	/* long to output */
int siz;
int procfmt; /* 0 when called from _wrtbyte() */
{
char *c,
	str[MAX_FMTLEN+1];
int len;
int i=l;	/* used if default int specified w/char hiding */
short si=l;	/* used if short specified w/char hiding */

	c = str;
	if( procfmt ){
		io_fmtuntl(o_up,NO,1); /* write until field descr */
		getunitfd( o_up, &out ); /*get field descr*/
		}
	if( out.fd=='I' || out.fd=='O' || out.fd=='Z' 
#if   defined(DG_F77)
	   || out.fd=='B'
#endif
	   ){
		/* NOTE: 'O' & 'Z' are VAX F77 extensions for octal & hex edits */
		/*       'B' is an extension for binary edits */
		/* convert to ASCII str */
		if( out.w > MAX_FMTLEN )
			fmterr( "I edit width is too big", o_up->ff->cf );
		intedit(l,out.fd,out.w,out.d,(o_up->ff->signc==PLUSSIGN),str);
		for( c=str; *c; c++ )
			WRTCH( *c, o_up )
		}
	else if( out.fd == 'A' ){	/* treat integers as chars */
#if defined(ONE_CHAR_PER_INT)
		/* assign long int to char */
		str[0] = (char)l;	str[1] = '\0';
		len = 1;
#else	/* presume FORTRAN-IV char hiding - default model */ 
		if( siz == sizeof(short) ){
			strcpy( str, ntstr(&si,(long)sizeof(short)) );
			len = sizeof(short);
			}
		else if( siz == sizeof(long) ){
			strcpy( str, ntstr(&l,(long)sizeof(long)) );
			len = sizeof(long);
			}
		else{	/* unspecified default int */
			strcpy( str, ntstr(&i,(long)sizeof(int)) );
			len = sizeof(int);
			}
		len = (len > out.w ? out.w : len);	/* insure '*'s aren't output */
#endif
		while( len-- )	/* output the int */
			WRTCH( *c++, o_up )
		}
	else if( out.fd == 'Q' ) /* VAX Q edit extension */
		;	/* ignored on output, output object skipped */
#if defined(PRIME_F77)
	else if( out.fd == 'B' ){ /* PRIME Business edit extension */
		int_busedit( l, out.s, str );
		for( c=str; *c; c++ )
			WRTCH( *c, o_up );
		}
#endif
	else{
		char tmp[80];
		sprintf( tmp,
		  "expecting 'I', 'B', 'O', 'Z', 'Q', or 'A' edit field - not %c",
		  out.fd );
		fmterr( tmp, o_up->ff->cf );
		}

return;
}

	/*	*	*	*	*/

#ifdef OLD_WRTDBL
void _wrtdbl( f )	/* write formatted floating pt no. to unit */
double f;	/* no. to output */
{
char *c, *fmtstr,
	*s, str[MAX_DBLLEN+1];
int len,	/* int length to output */
	mw;	/* minimum printing width */
int sf;	/* positive scale factor */

	io_fmtuntl(o_up,NO,1); /* write until field descr */
	getunitfd( o_up, &out ); /*get field descr*/

	mw = (out.w>MAX_DBLLEN) ? MAX_DBLLEN: out.w;
	/* convert float to ASCII str */
	if( out.fd == 'F' ){
		/* apply any scale factor */
		for( sf=abs(o_up->ff->scalef); sf > 0 ; sf-- )
			f = ( o_up->ff->scalef > 0 ? f*10 : f/10 );
		/* NOTE: scale factor logic for E, D, and G fields on output is
				not presently supported.
		 */
		fmtstr = (o_up->ff->signc == PLUSSIGN) ? "%+*.*f" : "%*.*f";
		}
	else if( out.fd == 'E' || out.fd == 'D' )
		fmtstr = (o_up->ff->signc == PLUSSIGN) ? "%+*.*e" : "%*.*e";
	else if( out.fd == 'G' ){
		/* C approx to G field */
		fmtstr = (o_up->ff->signc == PLUSSIGN) ? "%+*.*g  " : "%*.*g  ";
		mw -= 2;	/* since 2 spaces added above */
		}
	else if( out.fd != 'A' )
		fmterr("expecting 'D', 'E', 'F', or 'G' edit field",o_up->ff->cf);
	s = str; /*3/31/92*/
	if( out.fd == 'A' )	/* char hiding extension */
		strcpy( str, ntstr(&f,8L) );
	else{	/* regular floating point */
		sprintf( str, fmtstr, mw, out.d, f );
		if( out.fd=='F' && strncmp(s,"0.",2)==0 && strlen(s)>2 ) /*3/31/92*/
			s = s+1; /* skip lead '0' to match F77 output */
			/*(also insures the field width isn't exceeded in this case)*/
#ifdef LIT_D_FIELDDES
		/* NOTE: to be literal 'D', find the 'e' in str and convert 
		 *       it to a 'd' for F77 double precision.
		 */
		if( out.fd=='D' && (c=strchr(str,'e')) != NULL )
			*c = 'd';
#endif
		}

	len = strlen( s ); /*str->s 3/31/92*/
	if( len > out.w ){
		while( out.w-- )	/* number exceeds field width, output *'s */
			WRTCH( '*', o_up )
		}
	else{
		while( (out.w--) > len )	/* pad w/' ' */
			WRTCH( ' ', o_up )
		while( len-- )	/* output the no. */
			WRTCH( *s++, o_up )
		}

return;
}

#else /* exact/standard F77 form */

void _wrtdbl( f )	/* write formatted floating pt no. to unit */
double f;	/* no. to output */
{
	char str[MAX_FMTLEN+1];
	int i;

	io_fmtuntl(o_up,NO,1); /* write until field descr */
	getunitfd( o_up, &out ); /*get field descr*/

	out.w =
	   ((out.w>MAX_FMTLEN) ? MAX_FMTLEN: out.w); /*max print width*/

	/* convert float to ASCII str */
	if(out.fd == 'F' || out.fd == 'E' || out.fd == 'D' || out.fd == 'G'){
		fpedit( f, &out, o_up->ff->scalef, 
		          o_up->ff->signc==PLUSSIGN, str );
		for( i=0; i < out.w  && str[i]; i++ )
			WRTCH( str[i], o_up )
		}
	else if( out.fd == 'A' ){  /* char hiding extension (dp only) */
		/*NOTE: only Hollerith hiding in 8 byte dbls is supported,
		 *    since floats are promoted to doubles, thus messing
		 *    up any hidden chars.
		 */
		if( out.w < 8 ){ /* ck for overflow */
			memset( str, OVERFLOW_CHAR, out.w );
			str[out.w] = '\0';
			}
		else
			strcpy( str, ntstr(&f,8L) );
		while( (out.w--) > 8 )	/* pad w/' ' */
			WRTCH( ' ', o_up )
		for( i=0; i < 8; i++ ) /* output the no. */
			WRTCH( str[i], o_up )
		}
	else if( out.fd == 'B' ){ /* PRIME Business edit extension */
		char *c;
		fp_busedit( f, out.s, str );
		for( c=str; *c; c++ )
			WRTCH( *c, o_up );
		}
	else
		fmterr("expecting 'D', 'E', 'F', or 'G' edit field",o_up->ff->cf);

	return;
}


void fpedit(num, fe, sf, plus, str) /* gen. F77 fp edits */
double num; /* number to format */
FEDIT *fe; /* field edit descriptor */
int sf, /* scale factor */
	plus; /* plus sign flag */
char str[]; /* space for formatted number */
{
	double efactor = 1.0, tmp, pos_num = num;
	double int_part, frac_part, tmpint, tmpfrac1, tmpfrac2, pos_exp = 0.;
	int i, frac_len = 0, int_len = 0, numsign=0;
	int exp_factor = 0, exp_len = 0, exp_size, int_max_sz;
	int lead_blanks,
	     trail_blanks,
	     wd = fe->w; /* working field width */
	char expstr[MAX_EXPLEN+1],
		tmpstr[MAX_DBLLEN+1];
	char *ts = tmpstr;
	const int BASE = 10;

	/* insurance */
	fe->fd = TOUPPER( fe->fd );
	fe->x = ( fe->x > 0 ? fe->x : 2 );
	fe->d = ( fe->d >= 0 ? fe->d : 3 ); /* can't be -, why not 3! */
	/* NOTE: Officially, the F77 std considers these errors if
	 *     any of the above quantities are out of range.
	 */

	if( fe->d > 0 && !(sf > -fe->d && sf <= (fe->d + 2)) ) {
		f_warn("scale factor out of range");
		goto over_flow;
		}

	trail_blanks = fe->x + 2; /* normally 4, since fe->x defaults to 2 */

	/*---  get the positive value of the number  --- */
	if( num < 0.0 ){
		*ts++ = '-';
		pos_num = -num;
		numsign = -1;
		}
	else if( num > 0.0 )
		numsign = 1;

	/* special format for 'G'; change fe->d if pos_num within specific ranges */
	if( fe->fd == 'G' ){
		int g_flag = FALSE;
		for( i=0; i <= fe->d; i++ ){
			if( powi(10.0,(long)fe->d -i-1) <= pos_num 
			        && pos_num < powi(10.0,(long)fe->d-i) ){
				fe->d = i;
				wd -= trail_blanks;
				g_flag = TRUE;
				break;
				}
			}
		if ( !g_flag )
			fe->fd = 'E';
		}
	int_max_sz = wd - (fe->d + 1);

	/*-------  apply scale factor  ------- */
	if( sf != 0 && fe->fd == 'F' ) /*not applied for G; E & D done below*/
		pos_num *= powi(10.0,(long)sf); /* pos_num = pos_num * 10**sf */

	/*-----  determine length of integer part  ----- */
	frac_part = modf( pos_num, &int_part );
	tmpint = int_part;
	while ( int_part >= 1.0 ) {
		int_part /= BASE;
		++int_len;
		++exp_factor;
		}

	for ( i=1; i<int_len; i++ )
		efactor *= BASE;

	if( fe->fd=='E' || fe->fd=='D' ){

		if( sf > 0 && sf < fe->d + 2 )
			fe->d += 1 - sf;
		
		exp_size = fe->x + 2;
		if( exp_size > MAX_EXPLEN )
			f_err( "fpedit: str size" ); /* ins. */
		memset(expstr,'0',exp_size); /* fill exp field w/0's */
		expstr[exp_size] = '\0';
		expstr[0] = fe->fd; /* 'D' or 'E' */

		/*- determine exponent factor for 1.0 > pos_num > 0.0 -*/
		if( pos_num < 1.0 && pos_num > 0.0 ) {
			tmpfrac1 = modf( pos_num*BASE, &tmpint );
			while( !(tmpint > 0.0) ) {
				--exp_factor;
				tmpfrac1 = modf( tmpfrac1*BASE, &tmpint );
				}
			}

		/*-------  apply (opposite) exponent factor  ------- */
		pos_num *= powi(10.0,(long)(-exp_factor));

		/* round the last printed digit */
		if( !IS_FPZERO(pos_num) )
			pos_num += 0.5*powi(10.0,(long)-(fe->d + sf));

		if( pos_num >= 1.0 ) {
			pos_num *= 0.1;
			exp_factor++;
		}

		/*-------  apply sf factor  ------- */
		pos_num *= powi(10.0,(long)sf);

		int_len = sf > 0 ? sf : 0;

		/*----  efactor changes when int_len changes  ----*/
		efactor = 1.0;
		for ( i=1; i<int_len; i++ )
			efactor *= BASE;

		exp_factor -= sf; /* scale factor subtracted from exponent */

		/*----  determine number of leading blanks  ----*/
		lead_blanks = wd - (int_len + 1 + fe->d + exp_size);
		if( numsign < 0 )
			lead_blanks--; /* '-' added above */
		else if( plus && lead_blanks > 0 ){ /*room for leading +?*/
			lead_blanks--;
			*ts++ = '+';
			}
		if( lead_blanks > 0 && !int_len ){ /*when >0, room for lead 0*/
			lead_blanks--;
			int_len = 1; /* forces leading zero */
			}

		/*--- get the positive value of the exponent factor ---*/
		pos_exp = exp_factor;
		if ( pos_exp < 0 ) {
			pos_exp = -pos_exp;
			expstr[1] = '-';
			}
		else
			expstr[1] = '+';
		i = exp_size-1; /*start at last position (before '\0')*/
		if( pos_exp != 0 ){
			int k;
			while( pos_exp >= 1 ){
				k = (int)fmod(pos_exp,10.);
				expstr[i--] = edigit[k];
				pos_exp /= BASE;
				++exp_len;
				if( exp_len > fe->x ) /* exponent too big */
					goto over_flow;
				}
			}
		}
	else{ /* F or G edit */
		/* round the last printed digit */
		if( !IS_FPZERO(pos_num) ){
			double tmp, itmp;
			pos_num += 0.5*powi(10.0,(long) - fe->d);
			/* recalc int_len & efactor after rounding, in
			 * case "9.95+" is rounded to "10.0" with F5.1, etc.
			 * ... modified version of J Sandelin's fix, cwl 9/5/95
			 */
			int_len = 0;  efactor = 1.0;
			tmp = modf( pos_num, &itmp );
			while( itmp >= 1.0 ){
				itmp /= BASE;
				++int_len;
				}
			for( i=1; i < int_len; i++ )
				efactor *= BASE;
			}

		/*----  determine number of leading blanks  ----*/
		lead_blanks = int_max_sz - int_len;
		if( numsign < 0 )
			lead_blanks--; /* '-' added above */
		else if( plus && lead_blanks > 0 ){ /* room for lead +? */
			lead_blanks--;
			*ts++ = '+';
			}
		if( lead_blanks > 0  && int_len==0 ){
			/* space for leading 0 */
			lead_blanks--;
			int_len = 1;
			}
		}

	/*-----  write int part to hold string  ----- */
	frac_part = modf( pos_num, &int_part );
	for ( i=int_len; i>0; i-- ) {
		tmp = int_part/efactor;
		*ts++ = edigit[(int)tmp];
		int_part = int_part - ((int)tmp*efactor);
		efactor /= BASE;
		}
	*ts++ = '.';	/* write '.' to hold string */

	/*-----  write fraction part to hold string  ----- */
	tmpfrac1 = frac_part;
	tmpfrac2 = modf( tmpfrac1*BASE, &tmpint );
	for ( i=0; i < fe->d; i++ ) {
		*ts++ = edigit[(int)tmpint];
		++frac_len;
		tmpfrac1 = tmpfrac2;
		tmpfrac2 = modf( tmpfrac1*BASE, &tmpint );
		}
	*ts = '\0';

	/*-----  write return string for specific format  ----- */
	if( lead_blanks < 0 )
		goto over_flow;
	for ( i=0; i < lead_blanks; i++ )
		str[i] = ' ';

	switch ( fe->fd ) {
		case 'F' :
		case 'G' :
			{ int j = 0;
			while ( i < wd )
				str[i++] = tmpstr[j++];
			if( fe->fd == 'G' ) {
				for ( j=0; j < trail_blanks; j++ )
					str[i++] = ' ';	/*(unique to G format)*/
				}
			str[i] = '\0';
			}
			break;

		case 'E' :
		case 'D' :
			{ int k = 0;
			for( k=0; tmpstr[k]; k++ )  /* output fractional part */
				str[i++] = tmpstr[k];
			for( k=0; expstr[k]; k++ ) /* output exponent */
				str[i++] = expstr[k];
			str[i] = '\0';
			}
			break;

		}

	return;

over_flow:
	/* field overflow - exception condition */
	memset( str, OVERFLOW_CHAR, fe->w );
	str[fe->w] = '\0';
	return;
}

#endif

					/*	*	*	*	*/


int wrtubuf( us )	/* write out the contents of the unit buffer */
UNITSTRU *us;
{
char *startprt;	/* position to start print from */
int err=NO,	/* error flag */
	nl=NO, /* if true, add \n at end */
	n,  /* no. of chars to write out for UNformatted writes */
	nw; /* no. of chars actually written out */

startprt = us->ubuf;	/* NOTE: formatted output may differ */

if( us->access == SEQUENTIAL ){	/* (write out to us->nxtu) */
	if( us->ateof && !milspec )
		return( OK );	/* well so-so */
		/* NOTE: The Mil Spec (MIL-STD-1753) extensions require that it 
			be possible to write past the ENDFILE record (i.e., basically 
			ignore it) */
	/* evaluate the carriage control char to finish off LAST line */
		/* NOTE: the logic of finishing OFF the last line before 
			printing the current is required in order to handle the '+'
			carriage control char, (if a \n is output at the end of
			the current line to a physical device, then it is normally
			impossible to back up 1 line to do the over-print operation);
		 additionally,
			list directed & unformatted I/O don't used carriage control
		 */

	/* formatted, sequential output */
	nl = 1;	/* add \n at end of line */
	if( us->carrgc ){	/* stdout or stdprn UNIT */
		switch( (int) *us->ubuf ){ /* test the carriage control character! */
			case '0':	/* advance 2 lines */
				err = (fputc('\n',us->fp)==EOF) ? YES : NO;
				break;
			case '1':	/* advance to the top of the form */
				err = (fputc('\f',us->fp)==EOF) ? YES : NO;
				break;
			case '+':	/* don't advance, but rather print over next record */
				/* err = (fputc('\r',us->fp)==EOF) ? YES : NO; */
				*us->nxtu++ = '\r';
				nl = 0;
				break;
			case '$':	/* suppress \n at end -- common extension */
				nl = 0;
				break;
			case ' ':	/* space & anything else in col 1, adv 1 line */
			default:
				break;
			}	/* end of switch */
		if( err ){
			us->status = WRITE_ERR;	/* error writing to phys. device */
			return( _ERR_ );
			}
		/* std F77, although it prevents using the same format for a write
			followed by a read (since the read is always 1 char shy & an
			insufficient record error occurs) */
		startprt++;	/* 1st col isn't printed */
		}
	if( nl ){
		if( us->nxtu == us->ubuf )	/* blank line (from // type format) */
			*us->nxtu++ = ' ';	/* add ' ' for carriage cntl char */
		*us->nxtu++ = '\n';
		}
	*us->nxtu = '\0';	/* terminate the buffer */
	}
else{	/* DIRECT, (write out the entire record) */
	*us->ubufend = '\0'; /* insurance */
	/* NOTE: direct writes don't have a \n appended to them,
		even if formatted writes, since access is made through
		the record no. and NOT via a line */
	}

if( startprt > us->nxtu )
	startprt = us->nxtu;	/* insurance */
n = strlen(startprt);

nw = fwrite(startprt,sizeof(char),n,us->fp);
#ifdef DG_F77
 us->chars_proc = nw;
 if( nl )
	us->chars_proc--; /*don't count the \n appended to the record*/
#endif
if( nw < n ){
	us->status = WRITE_ERR;	/*error writing to phys. device*/
	return( _ERR_ );
	}

us->nxtu = us->ubuf;	/* reset */
return( OK );
}


	/*	*	*	*	*/


void intedit( num, edit, wd, md, plus, str) /* format for F77 Im.n edit */
long num; 
int edit, /* I, O, or Z edit */
	wd,  /* field (i.e., string) width */
	md,  /* width of leading zeros */
	plus; /* plus sign flag (if true, add + sign) */
char *str; /* ptr to string space */
{
	long efactor = 1, digit, pos_num, tmpint;
	int i, 
		int_len = 0,
		lead_blanks = 0, 
		zero_fill = 0,
		numsign=0;
	char tmpstr[MAX_LONGLEN+1],
		 *ts = tmpstr;
	int base=10;

	switch( edit ){
		case 'I':
			base = 10; break;
		case 'B':
			base = 2; break;
		case 'O':
			base = 8; break;
		case 'Z':
			base = 16; break;
		default:
			f_err("unknown integer edit (intedit)");
		}

	/*---  get the positive value of the number  --- */
	pos_num = num;
	if ( num < 0 ) {
		pos_num = -num;
		numsign = -1;
	}
	else if( num > 0 )
		numsign = 1;

	/*-----  determine length of integer  ----- */
	if( pos_num == 0 )
		int_len = 1;  /*always print at least 1 zero (except w/wd==0)*/
	else{
		tmpint = pos_num;
		while ( tmpint >= 1 ) {
			tmpint /= base;
			++int_len;
			}
		}
#ifdef DEBUG
	if( int_len > MAX_LONGLEN )
		f_err( "intedit: int_len > MAX_LONGLEN" ); /* insurance */
#endif

	/*-----  determine magnitude of number  ----- */
	for ( i=1; i<int_len; i++ )
		efactor *= base;

	/*-----  write (integer) number to hold string  ----- */
	tmpint = pos_num;
	while ( efactor >= 1 ) {
		digit = tmpint/efactor;
		*ts++ = edigit[(int)digit];
		tmpint = tmpint - ((int)digit*efactor);
		efactor /= base;
	}
	*ts = '\0';

	/*-----  write hold string to return string  ----- */
	if ( md > 0 ) {
		md = ( md > int_len ) ? md : int_len;
		zero_fill = md - int_len;
		lead_blanks = wd - (md + (numsign<0 || (plus && numsign>0)?1:0));
		}
	else
		lead_blanks = wd - (int_len + (numsign<0 || (plus && numsign>0)?1:0));

	/*----  check for format overflow  ---- */
	if ( lead_blanks < 0 ) {
		for ( i=0; i<wd; i++ )
			str[i] = OVERFLOW_CHAR;
		str[wd] = '\0';
		}
	else if( md == 0 && pos_num == 0 ){
		for ( i=0; i < wd; i++ )
			str[i] = ' ';
		str[wd] = '\0';
		}
	else {
		int j = 0, k = 0;
		for ( i=0; i<lead_blanks; i++ )	 /* output leading blanks */
			str[i] = ' ';
		if ( numsign < 0 )    /* output sign */
			str[i++] = '-';
		else if( plus && numsign > 0 )
			str[i++] = '+';
		for ( k=0; k < zero_fill; k++ )  /* output zero fill */
			str[i++] = '0';
		for ( k=0; k < int_len; k++ )  /* output integer number */
			str[i++] = tmpstr[j++];
		str[wd] = '\0';
	}

	return;
}

	/*   *   *   *   */

/* PRIME Business edit support */

void fp_busedit( v, bstr, to )
double v;
char *bstr;
char *to;
{
	/*stub*/
	strcpy( to, bstr );
	return;
}

void int_busedit( v, bstr, to )
long v;
char *bstr;
char *to;
{
	/*stub*/
	strcpy( to, bstr );
	return;
}
