/* detab.c - convert tabs to equivalent number of blanks
	by C Lightfoot, logic from "Software Tools", by K&P, pg 20) */

/*	usage:
			detab   <input_file   >output_file  [tab_siz]

	NOTE:	detab is simply a filter, reading from stdin & writing to stdout
			tab_siz is optional

	compilation of detab.c w/MS-C v4.0:

			cl detab.c
 */

#include <stdio.h>

#define DEF_TABSIZ	4
#define MXLINE	136
#define MAXLINE	MXLINE-1
#define TABPOS(col,tabs)	( col > MAXLINE ? 1 : tabs[col] )


main(argc, argv)
int argc;
char *argv[];
{
int c, col, i,
	tabsiz=DEF_TABSIZ;
char tabs[MXLINE];
	/* NOTE: tabs[0] is ignored, detab uses tabs[1] -> tabs[MAXLINE]
		to refer to columns 1 through MAXLINE respectively */

if( argc > 1 ){	/* ck for optional tabsiz */
	tabsiz = atoi( argv[1] );
	if( tabsiz < 1  ||  tabsiz > 9){
		puts( "DETAB ERROR: tabsiz must be between 1 and 9" );
		exit( 1 );
		}
	}

	/* set initial tab stops */
for( i=1; i <= MAXLINE; i++ )
	tabs[i] = ( i%tabsiz == 1 ) ? 1 : 0;

for( col=1; (c=getchar()) != EOF; ){
	if( c == '\t' ){
		do{
			putchar( ' ' );
			col++;
			} while( !TABPOS(col,tabs) );
		}
	else if( c == '\n' ){
		putchar( '\n' );
		col = 1;
		}
	else{
		putchar( c );
		col++;
		}
	}

exit( 0 );
}
