/*
* CLISRV.C                                                   Version 3.1
*
* Copyright 1994 Etc Bin Systems. All rights reserved
*
* Author: Peter Van Oudenaren
*
***********************************************************************/

// ********************************************************************
// This program demonstrates the use of PING, TCP ECHO, UDP ECHO, FTP,
// TFPT, TELNET and WEB over RT-IP.
// 
// Except for web server, all these tests can be run with 2 machines 
// networked together with clisrv.exe running on both machines or with
// clisrv.exe running against another network program which supports
// these protocols.  The web server can be run against a web browser.
// ********************************************************************

#include "socket.h"
#include "os_app.h"
#include "os_echo.h"
#if (defined(INCLUDE_WEB))
	#include "web.h"
#endif
#if (defined(INCLUDE_FTP))
	#include "ftpsufn.h"
	#include "ftpsrv.h"
	#include "ftpcli.h"
	#include  <string.h>
#endif

// ********************************************************************
#define FTP_BUF_SIZE 2048

#define FTP_TASK 	0
#define TFTP_TASK 	1
#define TELNET_TASK 2
#define UDP_TASK 	3
#define TCP_TASK 	4
#define WEB_TASK	5
#define TOTAL_TASKS 6

// ********************************************************************
// internal
void process_special_key(unsigned int key);
void write_box(BOOLEAN write_server);
void print_statistics(void);
BOOLEAN wait_reply(int socket_no, int seconds);
void echo_client(int socket_type);
void ping_client(void);
void start_udp_server(void);
void start_tcp_server(void);
void start_ftp_server(void);
void start_tftp_server(void);
void start_telnet_server(void);
void start_web_server(void);
SPAWN_DECLARE(tcp_server_main);
SPAWN_DECLARE(udp_server_main);
SPAWN_DECLARE(ftp_server_main);
SPAWN_DECLARE(tftp_server_main);
SPAWN_DECLARE(telnet_server_main);
SPAWN_DECLARE(web_server_main);

// external
void os_start_tasks(void);
void os_start_console_tasks(void);
#if (defined(INCLUDE_FTP))
void ftp_demo(int test);
void ftpsrv_main(void);
int tftpd_main(void);
int ftp_server_daemon(char *buffer, int bufsize);
void do_ftp_client(void);
void do_tftp_client(void);
int telnet_server_daemon(void);
#endif

// ********************************************************************
// GLOBAL DATA

#if (PLUS || AMXNET)
// stack for FTP task
STACK demo_task_stack[TOTAL_TASKS][DEMO_SPAWN_STACK_SIZE];
#endif

// task handles for spawned tasks
TASK_HANDLE demo_spawn_task[TOTAL_TASKS];

// flags used ensure each task is only spawned once
BOOLEAN FAR tcp_spawned = FALSE;
BOOLEAN FAR udp_spawned = FALSE;
BOOLEAN FAR ftp_spawned = FALSE;
BOOLEAN FAR tftp_spawned = FALSE;
BOOLEAN FAR telnet_spawned = FALSE;
BOOLEAN FAR web_spawned = FALSE;

char FAR ftp_buf[FTP_BUF_SIZE];
char FAR echo_buf[255];
char FAR ping_status[80];
int FAR tot_pkt_succ;
int FAR tot_pkt_fail;
int FAR tot_pkt;
int FAR test_type; 		// current test running
BOOLEAN do_quit;
#if (SUPPORTS_XNWIN)
	/* Windows */
	XMWIN_PTR tcd_appl_win   = 0;
	/* Buffers for holding what is underneath a window when we use 
   	   save under  */
	unsigned char FAR app_screen[80*25*2];   /* Holds the initial screen */
#endif


// ********************************************************************
// test types
#define TCP_SERVER  	1
#define TCP_CLIENT  	2
#define UDP_SERVER  	3
#define UDP_CLIENT  	4
#define FTP_SERVER  	5
#define FTP_CLIENT  	6
#define TFTP_SERVER 	7
#define TFTP_CLIENT 	8
#define TELNET_SERVER 	9
#define PING_CLIENT 	10
#define WEB_SERVER	    11

// ********************************************************************
// EXTERNAL DATA

#if (!USE_DB_L0)
	extern BOOLEAN do_return;
#endif
extern int curr_line;

// ********************************************************************
// exit_echo() - exit the program
//
//   This routine is called to perform exit processing and to terminate
//   this program.
//
//   This routine never returns.
//

void exit_echo(void)
{
AS_PUSH_PROLOG()

    AS_PUSH_INTERRUPTS()
    AS_DISABLE_INTERRUPTS()

	if (test_type == PING_CLIENT)
		print_statistics();

	app_interface_close();
	#if (SUPPORTS_XNWIN)
		if (tcd_appl_win)
            xnwpub_destroy(tcd_appl_win);
        xnwpub_wr_ustring(1, 22, FGCOLOR, BGCOLOR, 
			" Thank You..... Call 1-508-448-9340 to order                                 ");
	#endif
    os_exit();
}

// ********************************************************************
// tcd_error() - process a fatal error
//
//   This routine is called to exit the program when a fatal error has
//   been detected.
//
//   This routine never returns.
//

void tcd_error(char *p)
{
char b[10];

    SHOW_ERROR(p)
    os_promptstring("Press return to exit program", (unsigned char FAR *) b);
	exit_echo();
}


// ********************************************************************
// process_key() - process an input character
//
//    This routine is called to process input characters by the console
//    task.    
//    NOTE: for POLLOS this is called from the RTIP state machine stepping
//          routine
//
//    Returns nothing.
//

void process_key(unsigned int key)                             /* __fn__ */
{
extern word packet_delay;

    if (key == 27 /* KEY_ESC */)
    {
		exit_echo();
	}
    if ( (key == 'q') || (key == 'Q') )
	{
		do_quit = TRUE;
	}

	// slow down or speed up the packet driver
	else if (key == '+')
    {
        packet_delay += 100;
    }
    else if (key == '-')
    {
        packet_delay -= 100;
    }

	else
		process_special_key(key);
}

// ********************************************************************
// process_key() - process an input character
//
//    This routine is called to process input characters.
//
//    Returns nothing.
//

void process_special_key(unsigned int key)
{
#if (!USE_DB_L0)
	
	// turn debug logging on and off
	if (key == 's')
	{
        if (do_return == TRUE)
		{
			do_return = FALSE;
		}
		else
		{
			do_return = TRUE;
		}
	}
#endif
	if (key == 't')
	{
		// display track info
		track_packet_info(TRUE);
	}
}

// ********************************************************************
// print_statistics() - print ping statistics
//
//   Prints ping statistics, i.e. success versus failures for all ping
//   messages sent by the current request.
//
//   Returns nothing.

void print_statistics(void)
{
	strcpy(ping_status, "total success packets = ");
 	_itoa(tot_pkt_succ, ping_status + strlen(ping_status), 10);
	strcpy(ping_status + strlen(ping_status), ", total failures = ");
 	_itoa(tot_pkt_fail, ping_status + strlen(ping_status), 10);
   	SHOW_PROGRESS(ping_status)
   	SHOW_PROGRESS(" ")
}


// ********************************************************************
// APPLICATION INITIALIZATION
// ********************************************************************
void  application_init()                             /* __fn__ */
{
	// start the tasks, i.e. app_entry
    os_start_tasks();
}


// ********************************************************************
// THIS TASK IS THE MAIN PROGRAM
// ********************************************************************
void  app_entry(void)                             /* __fn__ */
{
byte io_buf[10];
BOOLEAN exit_loop;

	// query user as to IP addresses, devices etc etc
	if (!app_init_parms())
		exit_echo();

	#if (SUPPORTS_XNWIN)
		/* Create the video windows for the demo */
	    tcd_appl_win = xnwpub_create( 0,  0, 80, 25, FGCOLOR, BGCOLOR,  XWS_BORDER_DOUBLE|XWS_SAVE_UNDER);
		write_box(TRUE);
	#endif

	// Initialize the tcpip subsystem
    SHOW_PROGRESS("Starting Network Services")
    if (xn_init() != 0)
		tcd_error("Network Initialize Failed");

	// Open the interface - i.e. do xn_interface_open or xn_attach
	// depending upon the interface; possibly do xn_rarp or xn_bootp()
	// then do xn_set_ip()
	// NOTE: this should not be done until after app_init_parms() is called
	app_interface_open();

	// for PPP interface, open the PPP connection
	#if (INCLUDE_PPP)
		app_ppp_open();
	#endif

	// start the background console task
	os_start_console_tasks();

	exit_loop = FALSE;
	while (!exit_loop)		// loop until q or ESC
	{
		// stop the background console task in order for os_promptstring
		// to beable to capture all the input
		os_stop_console_tasks();

		////////
		os_out_string("Select test to run where:");
		os_out_string("  1) SPAWN TCP ECHO Server  2) TCP ECHO Client");
		os_out_string("  3) SPAWN UDP ECHO Server  4) UDP ECHO Client");
		os_out_string("  5) SPAWN FTP Server       6) FTP Client");
		os_out_string("  7) SPAWN TFTP Server      8) TFTP Client");
		os_out_string("  9) SPAWN TELNET Server   10) PING Client");
		os_out_string(" 11) SPAWN WEB Server");
		os_out_string("  Q) Quit");
		os_promptstring("Enter choice: ", io_buf);

		////////
		// start the background console task again
		os_resume_console_tasks();

		////////
		// modify test to a number, setting Q to 99
		if (io_buf[0] == 'Q')
			test_type = 99;
		else
			test_type = myatoi((PFCHAR)io_buf);

		switch (test_type)
		{
		case TCP_SERVER: 
			start_tcp_server();
			break;
		case TCP_CLIENT: 
			echo_client(SOCK_STREAM);
			break;
		case UDP_SERVER: 
			start_udp_server();
			break;
		case UDP_CLIENT: 
			echo_client(SOCK_DGRAM);
			break;
		case PING_CLIENT:
			ping_client();
			break;
#if (defined(INCLUDE_WEB))
		case WEB_SERVER: 
			start_web_server();
			break;
#endif
		case 99:
			exit_loop = TRUE;
			break;
#if (defined(INCLUDE_FTP))
		case FTP_CLIENT:
		case FTP_SERVER:  
		case TELNET_SERVER:
		case TFTP_SERVER:
		case TFTP_CLIENT:
			ftp_demo(test_type);
			break;
#endif
		default:
			process_special_key(io_buf[0]);
		}
	}

	exit_echo();
}
	
// ********************************************************************
// cat_server() - add the server to the list of servers 
//
//   Writes the server (server) to the list (io_buf) of servers which
//   which servers are running.  Decrements the count of the number of
//   servers which is used to determine if a "and" or a "," should
//   be written before the current server.
//
//   Returns nothing.
//

void cat_server(char *io_buf, char *server, int *num_servers)
{
	if (*num_servers == 1)
		strcat(io_buf, " and ");
	else
		strcat(io_buf, ", ");
 
	strcat(io_buf, server);

	(*num_servers)--;
}

// ********************************************************************
// write_box() -  write the border around the demo
//
//   Writes the border around the demo and the message at the bottom of
//   the screen which lists which servers are running.
//
//   Returns nothing.
//

void write_box(BOOLEAN write_servers)
{
char io_buf[85];
int num_servers;

	// write the box around the screen
    xnwpub_render(tcd_appl_win, app_screen);

	if (write_servers)
	{
		num_servers = 0;
		if (tcp_spawned)
			num_servers++;
		if (udp_spawned)
			num_servers++;
		if (ftp_spawned)
			num_servers++;
		if (tftp_spawned)
			num_servers++;
		if (telnet_spawned)
			num_servers++;
		if (web_spawned)
			num_servers++;

		strcpy(io_buf, " Send PING");
		if (num_servers == 0)
			strcat(io_buf, " and SNMP");
		else
			strcat(io_buf, ", SNMP");
		if (tcp_spawned && udp_spawned)
		{
			num_servers--;
			cat_server(io_buf, "ECHO", &num_servers);
		}
		else
		{
			if (tcp_spawned)
				cat_server(io_buf, "TCP ECHO", &num_servers);
			if (udp_spawned)
				cat_server(io_buf, "UDP ECHO", &num_servers);
		}
		if (ftp_spawned)
			cat_server(io_buf, "FTP", &num_servers);
		if (tftp_spawned)
			cat_server(io_buf, "TFTP", &num_servers);
		if (telnet_spawned)
			cat_server(io_buf, "TELNET", &num_servers);
		if (web_spawned)
			cat_server(io_buf, "WEB", &num_servers);
		strcat(io_buf, " requests at any time ");

		/* Put up a prompt at the bottom of the screen */
		xnwpub_wr_ustring(1, 24, FGCOLOR, BGCOLOR, io_buf);
	}
}

// ********************************************************************
// echo client() - echo client for TCP or UDP
//   Performs the following:
//     Connects to a tcp or udp server. 
//     Requests input from the keyboard.
//     Sends it to the server
//     Waits for an echo
//     prints the echo 
//	 This routine returns when "QUIT" is entered for the string to be
//   echoed.
//
//   Returns nothing.
//

void echo_client(int socket_type)
{
int socket_no;
struct sockaddr_in sin;
LINGER  ling;
int n_sent, n_received;
char ebuf[10];

	os_out_string("Welcome to the echo client program");
	os_out_string(" ");
	os_out_string("This progam:");
	if (socket_type == SOCK_STREAM)
		os_out_string("Connects to a tcp server. ");
	else
		os_out_string("Connects to a udp server. ");
	os_out_string("Requests input from the keyboard.");
	os_out_string("Sends it to the server");
	os_out_string("Waits for an echo");
	os_out_string("Prints the echo ");

	////////
	// Allocate a socket
	if ( (socket_no = socket(AF_INET, socket_type, 0)) < 0 )
	{
		os_out_string("Can\'t Allocate Socket");
		return;
	}

	// connect to the echo server at port 7
	sin.sin_family = AF_INET;
	tc_movebytes((PFBYTE)&sin.sin_addr, ap_p[0].his_ip_address, 4);
	sin.sin_port = hs2net(7);
    if (connect(socket_no, (PSOCKADDR)&sin, sizeof(sin)) < 0)
	{
        os_out_string("Connect Failed");
		return;
	}
	if (socket_type == SOCK_STREAM)
	{
		// Select hard close so we can close down even if the other
		// side doesn't respond 
		// set socket option to linger with 0 timeout
		ling.l_onoff = 1;
		ling.l_linger = 0;
		setsockopt(socket_no, SOL_SOCKET, SO_LINGER, (PFCCHAR)&ling, 
		               sizeof(struct linger));
	 }

	while (1)
	{
		echo_buf[0] = '\0';
		while (strlen(echo_buf) == 0)
		{
			// stop the background console task 
			os_stop_console_tasks();

			os_promptstring("Enter string to echo, QUIT to quit: ", 
					(unsigned char FAR *)echo_buf);

			// start the background console task again
			os_resume_console_tasks();
		}

		// Send the data
		do
		{
			n_sent = send(socket_no, (PFCCHAR)echo_buf, strlen(echo_buf), 0);
			if (n_sent < 0)
			{
				os_out_string("Send Failed");
				ebuf[0] = 0;
				os_promptstring("Type Q to quit or RETURN to try again",
					(unsigned char FAR *)ebuf);
				if (ebuf[0] == 'Q' || ebuf[0] == 'q')
				{
					closesocket(socket_no);
			        return;
				}
			}
		 } while (n_sent <= 0);

		if (echo_buf[0] == 'Q' &&	
			echo_buf[1] == 'U' &&
			echo_buf[2] == 'I' &&
			echo_buf[3] == 'T')
		{
			closesocket(socket_no);
		 	return;
		}

		// Wait for a response. Be flexible in case they are single stepping on
		// the other side.
		while (!wait_reply(socket_no, 10))
		{
			os_out_string("Target not responding after 10 seconds, trying 30");
			if (!wait_reply(socket_no, 30))
			{
				os_out_string("Target not responding after 30 seconds");
				ebuf[0] = 0;
				os_promptstring("Type C to re-send Q  to Quit just return to keep trying",
				(unsigned char FAR *)ebuf);
				if (ebuf[0] == 'C' || ebuf[0] == 'c')
					break;
				if (ebuf[0] == 'Q' || ebuf[0] == 'q')
				{
					closesocket(socket_no);
			        return;
				}
			}		
		}

		// Receive the echo
		if (wait_reply(socket_no, 1))
			n_received = recv(socket_no, (char FAR *)echo_buf, n_sent, 0);
		else
			n_received = 0;

		/* display it on the console */
		echo_buf[n_received] = '\0';	   // make sure it's null term */
		echo_buf[60] = '\0';
		os_out_two(echo_buf, "Echoed: ");
	}
}

// ********************************************************************
// wait_reply() - calls select to wait for data to be available
//
//   Calls select to wait upto wait seconds for input data to be available.
//
//   Returns TRUE if data available or FALSE if timeout
//

BOOLEAN wait_reply(int socket_no, int wait)
{
fd_set fd_read;
timeval tmeout; 

	tmeout.tv_sec =  wait; 
    tmeout.tv_usec = 0;  

	FD_ZERO(&fd_read);
	FD_SET(socket_no, &fd_read);

	if (select(1, (PFDSET)&fd_read, (PFDSET)0, (PFDSET)0, 
	           (PTIMEVAL)&tmeout) == 1)
		return(TRUE);
	else
		return(FALSE);
}

// ********************************************************************
// ping_client() - sends ping messages and waits for response
//
//    Queries users as to number of ping request, delay between sending
//    them, size of request.  Then sends the requests and updates the
//    statistics as to the success or failure.
//
//    Returns nothing.
//

void ping_client(void)
{
int sequence;
int len;
int delay;
int num_req;
int decr;
int looping;
char io_buf[10];

	looping = 1;

	SHOW_PROGRESS("PING CLIENT RUNNING")

	do
	{
		// stop the background console task in order for os_promptstring
		// to beable to capture all the input
		os_stop_console_tasks();

		os_promptstring("Enter number of ping requests to send (-1=Return Main Menu, 0=forever): ",
			(PFBYTE)io_buf);
		num_req = myatoi(io_buf);
	 	if (num_req < 0)
			break;

		// get test information
		os_promptstring("Enter delay in seconds between requests (0=flood mode): ",
			(PFBYTE)io_buf);
		delay = myatoi(io_buf);

		os_promptstring("Enter number of data bytes in ping request: ",
			(PFBYTE)io_buf);
		len = myatoi(io_buf);

		decr = 1;
		if (!num_req)	// if forever
		{
			num_req = 1;
			decr = 0;		// will cause loop forever
		}

		////////
		// start the background console task again
		os_resume_console_tasks();

		////////
		tot_pkt_succ = 0;
		tot_pkt_fail = 0;
		tot_pkt = 0;
		sequence = 0x34;

		while (num_req)
		{
			if ( xn_ping(sequence, len, ap_p[0].his_ip_address, 10*os_ticks_p_sec()) )
			{
				tot_pkt_fail++;
				strcpy(ping_status, "FAILURE: ");
			 	_itoa(len, ping_status + strlen(ping_status), 10);
				strcpy(ping_status + strlen(ping_status), " bytes, seq = ");
			 	_itoa(tot_pkt, ping_status + strlen(ping_status), 10);
	    		SHOW_PROGRESS(ping_status)
			}
			else
			{
				tot_pkt_succ++;
				strcpy(ping_status, "SUCCESS: ");
			 	_itoa(len, ping_status + strlen(ping_status), 10);
				strcpy(ping_status + strlen(ping_status), " bytes, seq = ");
			 	_itoa(tot_pkt, ping_status + strlen(ping_status), 10);
	    		SHOW_PROGRESS(ping_status)
			}
			tot_pkt++;

			if (delay > 0)
				os_sleep((word)(delay*os_ticks_p_sec()));

			num_req -= decr;
		}
		print_statistics();

	} while (looping);
}



// ********************************************************************
// str_replace() - replace template with another string.
// 
//   Replaces one occurance of xxx_str with task_str in the string msg.
//
//   Returns nothing.
void str_replace(char *msg, char *xxx_str, char *task_str)
{
char *ptr_str;
int xxx_len;
int task_len;

	xxx_len = strlen(xxx_str);
	task_len = strlen(task_str);
	ptr_str = strstr(msg, xxx_str);
	if (xxx_len < task_len)
	{
		DEBUG_ERROR("str_replace - warning - xxx_len < task_len = ",
			INT2, xxx_len, task_len);
	}

	// if length of string replace is < length of string replacing it
	// with, shift the part of the string to the right of the XXX over
	// to the left
	else if (xxx_len > task_len)
		strcpy(ptr_str+task_len, ptr_str+xxx_len);
	if (ptr_str != 0)
		strncpy(ptr_str, task_str, strlen(task_str));
}

// ********************************************************************
// display_server_screen() - displays screen when a server task is started.
//
//    Displays screen when server task, task, is started.  Uses a template
//    which it replaces XXXXXX with task in order to print out the correct
//    information about the task started.
//
//    Returns nothing.
//

void display_server_screen(char *task, BOOLEAN *spawned)
{
byte io_buf[10];
char msg1[80];
char msg2[80];
char msg3[80];
char msg4[80];
char msg5[80];

	strcpy(msg1, "[======= A XXXXXXXXX SERVER TASK HAS BEEN SPAWNED ========]");
    strcpy(msg2, "[======= A XXXXXXXXX SERVER TASK HAS ALREADY BEEN SPAWNED ========]");
	strcpy(msg3, "The XXXXXXXXX task is running in the background.");
	strcpy(msg4, "You can send XXXXXXXXX requests from a remote machine");
	strcpy(msg5, "even after returning to the main menu.");

	str_replace(msg1, "XXXXXXXXX", task);
	str_replace(msg2, "XXXXXXXXX", task);
	str_replace(msg3, "XXXXXXXXX", task);
	str_replace(msg4, "XXXXXXXXX", task);

	clear_screen(FALSE, FALSE);
	write_box(FALSE);
	if (!(*spawned))
	   	xnwpub_wr_string(tcd_appl_win, 12, 1, FGCOLOR, BGCOLOR, msg1);
	else
	   	xnwpub_wr_string(tcd_appl_win, 12, 1, FGCOLOR, BGCOLOR, msg2);
    xnwpub_wr_string(tcd_appl_win, 13, 5, FGCOLOR, BGCOLOR, msg3);
    xnwpub_wr_string(tcd_appl_win, 13, 7, FGCOLOR, BGCOLOR, msg4);
    xnwpub_wr_string(tcd_appl_win, 13, 8, FGCOLOR, BGCOLOR, msg5);

	*spawned = TRUE;

	/* Put up a prompt at the bottom of the screen */
	xnwpub_wr_ustring(1, 24, FGCOLOR, BGCOLOR,
		"Press ENTER to return to main menu");
	curr_line = 10;  // set curr line for scrolling routines
	os_stop_console_tasks();
	os_promptstring("", io_buf);
	os_resume_console_tasks();
	clear_screen(FALSE, FALSE);
	write_box(TRUE);
}

// ********************************************************************
// ftp_demo() - process requests for FTP, TFTP and TELNET
//
//   Called in response to user requesting to run FTP, TFTP or TELNET.
//   Based upon requests, calls appropriate routine to handle the request.
//
//   Returns nothing.
//

#if (defined(INCLUDE_FTP))
void  ftp_demo(int test)                             /* __fn__ */
{
AS_PUSH_PROLOG()

	// Initialize RTFS if needed 
	switch (test)
	{
		case FTP_SERVER:  
		case FTP_CLIENT:
		case TFTP_SERVER:
		case TFTP_CLIENT:
    	case 1:   /* Ftp Client */
    	case 2:   /* Ftp server */
    	case 4:   /* tftp server */
    	case 3:   /* tftp client */
			SHOW_PROGRESS("Initializing RTFS")
        	if (ufn_init("A:") != 0)
               	os_out_string("ftp_demo - RTFS Initialize failed ");
	}

	switch (test)
	{
	case FTP_CLIENT:
		clear_screen(FALSE, FALSE);
		write_box(FALSE);
    	/* Put up a prompt at the bottom of the screen */
		#if (SUPPORTS_XNWIN)
			xnwpub_wr_ustring(1, 24, FGCOLOR, BGCOLOR, "Type QUIT to exit ");
		#endif

		// stop the background console task in order for os_promptstring
		// to beable to capture all the input
		os_stop_console_tasks();

    	do_ftp_client();

		os_resume_console_tasks();

		clear_screen(FALSE, FALSE);
    	/* Clear prompt at the bottom of the screen */
		write_box(TRUE);
    	break;

	case FTP_SERVER:  
		start_ftp_server();
    	break;

	case TELNET_SERVER:
		start_telnet_server();
    	break;

	case TFTP_SERVER:
		start_tftp_server();
    	break;

	case TFTP_CLIENT:
		clear_screen(FALSE, FALSE);

		// stop the background console task in order for os_promptstring
		// to beable to capture all the input
		os_stop_console_tasks();

    	do_tftp_client();

    	/* Clear prompt at the bottom of the screen */
		#if (SUPPORTS_XNWIN)
			clear_line(24, TRUE, TRUE);
		#endif

		os_resume_console_tasks();
		clear_screen(FALSE, FALSE);
		write_box(TRUE);
    	break;
	}
}
#endif


// ********************************************************************
// ROUTINES TO SPAWN SERVER TASKS 
// ********************************************************************

// ********************************************************************
// start_tcp_server() - spawn tcp server task
//
//    If the tcp server task has not previouly been spawned, this function
//    spawns the task tcp_server_main().  Then it displays server window 
//    which the user exits by hitting any key.
//
//    Returns nothing.
//

void start_tcp_server(void)
{
	if (!tcp_spawned)
	{
		// spawn an tcp server task
		if ( !os_spawn_task(TCP_TASK, tcp_server_main, DEMO_SPAWN_STACK(TCP_TASK), 
	                		DEMO_SPAWN_STACK_SIZE, DEMO_SPAWN_PRIO, 
							demo_spawn_task, "TCPE") )
		{
			os_out_string("spawn of TCP task failed");
			return;
		}
	}
	#if (SUPPORTS_XNWIN)
		display_server_screen("TCP ECHO", &tcp_spawned);
	#endif
}

// ********************************************************************
// start_udp_server() - spawn udp server task
//
//    If the udp server task has not previouly been spawned, this function
//    spawns the task udp_server_main().  Then it displays server window 
//    which the user exits by hitting any key.
//
//    Returns nothing.
//

void start_udp_server(void)
{
	if (!udp_spawned)
	{
		// spawn an udp server task
		if ( !os_spawn_task(UDP_TASK, udp_server_main, DEMO_SPAWN_STACK(UDP_TASK), 
	                		DEMO_SPAWN_STACK_SIZE, DEMO_SPAWN_PRIO, 
							demo_spawn_task, "UDPE") )
		{
			os_out_string("spawn of UDP task failed");
			return;
		}
	}
	#if (SUPPORTS_XNWIN)
		display_server_screen("UDP ECHO", &udp_spawned);
	#endif
}

// ********************************************************************
// start_ftp_server() - spawn ftp server task
//
//    If the ftp server task has not previouly been spawned, this function
//    spawns the task ftp_server_main().  Then it displays server window 
//    which the user exits by hitting any key.
//
//    Returns nothing.
//

#if (defined(INCLUDE_FTP))
void start_ftp_server(void)
{
	if (!ftp_spawned)
	{
		// spawn an ftp server task
		if ( !os_spawn_task(FTP_TASK, ftp_server_main, DEMO_SPAWN_STACK(FTP_TASK), 
		                	DEMO_SPAWN_STACK_SIZE, DEMO_SPAWN_PRIO, 
							demo_spawn_task, "FTP ") )
		{
			os_out_string("spawn of FTP task failed");
			return;
		}
	}
	#if (SUPPORTS_XNWIN)
		display_server_screen("FTP", &ftp_spawned);
	#endif
}

// ********************************************************************
// start_tftp_server() - spawn tftp server task
//
//    If the tftp server task has not previouly been spawned, this function
//    spawns the task tftp_server_main().  Then it displays server window 
//    which the user exits by hitting any key.
//
//    Returns nothing.
//

void start_tftp_server(void)
{
	if (!tftp_spawned)
	{
		// spawn an tftp server task
		if ( !os_spawn_task(TFTP_TASK, tftp_server_main, DEMO_SPAWN_STACK(TFTP_TASK), 
		                	DEMO_SPAWN_STACK_SIZE, DEMO_SPAWN_PRIO, 
							demo_spawn_task, "TFTP") )
		{
			os_out_string("spawn of TFTP task failed");
			return;
		}
	}
	#if (SUPPORTS_XNWIN)
		display_server_screen("TFTP", &tftp_spawned);
	#endif
}

// ********************************************************************
// start_telnet_server() - spawn telnet server task
//
//    If the telnet server task has not previouly been spawned, this function
//    spawns the task telnet_server_main().  Then it displays server window 
//    which the user exits by hitting any key.
//
//    Returns nothing.
//

void start_telnet_server(void)
{
	if (!telnet_spawned)
	{
		// spawn an telnet server task
		if ( !os_spawn_task(TELNET_TASK, telnet_server_main, DEMO_SPAWN_STACK(TELNET_TASK), 
		                	DEMO_SPAWN_STACK_SIZE, DEMO_SPAWN_PRIO, 
							demo_spawn_task, "TELN") )
		{
			os_out_string("spawn of TELNET task failed");
			return;
		}
	}
	#if (SUPPORTS_XNWIN)
		display_server_screen("TELNET", &telnet_spawned);
	#endif
}
#endif

// ********************************************************************
// start_web_server() - spawn web server task
//
//    If the web server task has not previouly been spawned, this function
//    spawns the task web_server_main().  Then it displays server window 
//    which the user exits by hitting any key.
//
//    Returns nothing.
//

#if (defined(INCLUDE_WEB))
void start_web_server(void)
{
	if (!web_spawned)
	{
		// spawn an web server task
		if ( !os_spawn_task(WEB_TASK, web_server_main, DEMO_SPAWN_STACK(WEB_TASK), 
	                		DEMO_SPAWN_STACK_SIZE, DEMO_SPAWN_PRIO, 
							demo_spawn_task, "WEB ") )
		{
			os_out_string("spawn of WEB task failed");
			return;
		}
	}
	#if (SUPPORTS_XNWIN)
		display_server_screen("WEB", &web_spawned);
	#endif
}

#endif

// ********************************************************************
// SERVER TASKS WHICH ARE SPAWNED
// ********************************************************************

// ********************************************************************
// tcp_server_main() - tcp server task
//
//    The tcp server task is spawned via user request.  It processes
//    echo requests sent from a remote host to the well-known echo
//    port by echoing them back to the remote host.
//
//    This task never returns.
//

SPAWN_DECLARE(tcp_server_main)
{
int socket_m;
int socket_no;
struct sockaddr_in sin;
int sin_len;
int n_received;
	
	// Allocate a socket
    if ( (socket_m = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
        os_out_string("tcp_server_main - Can\'t Allocate Socket");

    // Bind my ip address port 7
	sin.sin_family = AF_INET;
	tc_movebytes((PFBYTE)&sin.sin_addr, ap_p[0].my_ip_address, 4);

	sin.sin_port = hs2net(7);
	if (bind(socket_m, (PSOCKADDR)&sin, sizeof(sin)) < 0)
		os_out_string("tcp_server_main - Bind Failed");

    if (listen(socket_m, 2) == -1)   // 2 slave ports
		os_out_string("tcp_server_main - Listen Failed");

	while (1)
	{
		sin_len = sizeof(sin);
    	if ( (socket_no	= accept(socket_m, (PSOCKADDR)&sin, (int FAR *) &sin_len)) < 0)
        	os_out_string("tcp_server_main - Accept Failed");
	
	 	while (1)
	 	{
			/* Get a packet */
	    	n_received = recv(socket_no, (char FAR *)echo_buf, 255, 0);
			if (n_received < 0)
				break;

			/* Echo it to the other side */
			DEBUG_LOG("echo server - send back response", LEVEL_2, NOVAR, 0, 0);
			send(socket_no, (const char FAR *)echo_buf, n_received, 0);
		
			/* display it on the console */
			echo_buf[n_received] = '\0';	   // make sure it's null term */
			echo_buf[60] = '\0';
			os_out_string(echo_buf);
			if (echo_buf[0] == 'Q' &&	
				echo_buf[1] == 'U' &&
				echo_buf[2] == 'I' &&
				echo_buf[3] == 'T')
				break;
		}	// end of while current session
		closesocket(socket_no);
	}		// end of while - loops forever
}

// ********************************************************************
// udp_server_main() - udp server task
//
//    The udp server task is spawned via user request.  It processes
//    echo requests sent from a remote host to the well-known echo
//    port by echoing them back to the remote host.
//
//    This task never returns.
//

SPAWN_DECLARE(udp_server_main)
{
int socket_no;
struct sockaddr_in sin;
int sin_len;
int n_received;
	
	// Allocate a socket
    if ( (socket_no = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
        os_out_string("udp_server_main - Can\'t Allocate Socket");
	
    // Bind my ip address port 7
	sin.sin_family = AF_INET;

	tc_movebytes((PFBYTE)&sin.sin_addr, ap_p[0].my_ip_address, 4);
	sin.sin_port = hs2net(7);
	if (bind(socket_no, (PSOCKADDR)&sin, sizeof(sin)) < 0)
		os_out_string("udp_server_main - Bind Failed");

	 while (1)
	 {
		 sin_len = sizeof(sin);
		/* Get a packet */
    	n_received = recvfrom(socket_no, (char FAR *)echo_buf, 255, 0, 
									  (PSOCKADDR) &sin, 
									  (int FAR *) &sin_len);
		/* echo it */
		sendto(socket_no, (PFCCHAR)echo_buf, n_received, 0,
			   (PSOCKADDR) &sin, sin_len);
		/* display it on the console */
		echo_buf[n_received] = '\0';	   // make sure it's null term */
		echo_buf[60] = '\0';
		os_out_string(echo_buf);

		// ignore QUIT, i.e. nothing to do since server needs to run
		// forever
	}
}

// ********************************************************************
// ftp_server_main() - ftp server task
//
//    The ftp server task is spawned via user request.  It processes
//    ftp requests sent from a remote host.
//
//    This task never returns.
//

#if (defined(INCLUDE_FTP))
SPAWN_DECLARE(ftp_server_main)
{
	while (1)
	{
    	if (ftp_server_daemon(ftp_buf, FTP_BUF_SIZE) == -1)
		{
        	DEBUG_ERROR("ftp server failed", NOVAR, 0, 0);
		}
    	else
		{
        	DEBUG_ERROR("ftp server finished", NOVAR, 0, 0);
		}
	}
}

// ********************************************************************
// tftp_server_main() - tftp server task
//
//    The tftp server task is spawned via user request.  It processes
//    tftp requests sent from a remote host.
//
//    This task never returns.
//

SPAWN_DECLARE(tftp_server_main)
{
	while (1)
	{
    	if (tftpd_main() == -1)
		{
        	DEBUG_ERROR("tftp server failed", NOVAR, 0, 0);
		}
    	else
		{
        	DEBUG_ERROR("tftp server finished", NOVAR, 0, 0);
		}
	}
}


// ********************************************************************
// telnet_server_main() - telnet server task
//
//    The telnet server task is spawned via user request.  It processes
//    telnet requests sent from a remote host by echoing them back.  A
//    command interpreter would need to be added for the application.
//
//    This task never returns.
//

SPAWN_DECLARE(telnet_server_main)
{
	while (1)
	{
    	if (telnet_server_daemon() == -1)
		{
        	DEBUG_ERROR("telnet server failed", NOVAR, 0, 0);
		}
    	else
		{
        	DEBUG_ERROR("telnet server finished", NOVAR, 0, 0);
		}
	}
}


#endif		// end of defined(INCLUDE_FTP)

// ********************************************************************
// web_server_main() - web server task
//
//    The web server task is spawned via user request.  It processes
//    web requests sent from a remote host.
//
//    This task never returns.
//

#if (defined(INCLUDE_WEB))
SPAWN_DECLARE(web_server_main)
{
	while (1) 
	{
		if (http_server_daemon() == -1) 
		{
		   	DEBUG_ERROR("HTTP server failed!", NOVAR, 0, 0);
			os_sleep(os_ticks_p_sec());
		}
		else
		   	DEBUG_ERROR("HTTP server finished", NOVAR, 0, 0);
	}
}

#endif

